#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                        Andrew Seawright
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are not implemented in this version.
// This was done to trade-off efficiency vs. ease of implemetation and ease
// of modification.
//
//	Revised (7/19/93) F. Brewer
//	* changed the inconsistent node creation procedure to prevent
//	some bugs leading to unknown numbers of inputs by the manager.
//
////////////////////////////////////////////////////////////////////////////
//
// manager constructor
//

BDDmgr::BDDmgr( integer bits, integer cofbits) {
  
  one        = new BDDnode(-1,(BDDnode*)1,NULL); // make them different
  zero       = new BDDnode(-1,NULL,NULL);	// Create zero and one

  tempnode     = new BDDnode;			// used as temp BDDnode in Makenode
  tempcomputed = new BDDCacheKey;		// used as temp BDDcomp. in ite

						// allocate the caches
  computed = new BDDCache( (1<<bits)-1, bits );

  if( cofbits > 0 )				// then one cache for unique & cof
    cofactor_hc  = new BDDCache( (1<<cofbits)-1, cofbits);
  else
    cofactor_hc = computed;
  restrict_hc  = cofactor_hc;

  one->SetOverflow();				// Since refs are needed for
  zero->SetOverflow();				// the 1st call to Makenode

  num_ite_calls = 0;				// stats to get perf.
  num_nodes     = 2;				// zero & one already allocated
  num_inputs    = 0;				// keeps track of indices

  unique        = new BDDHasher;		// BDD hashes

  defaultIA	= new IndexArray(MAX_BDD_VAR);	// all vars

  num_keeps	= 0;
  keptNodes	= new HasherTree(TRUE,FALSE);	// All nodes which are kept

  timestamp     = 0;
  SRANDOM();                                    // to keep Ivan happy
}

BDDmgr::BDDmgr(BDDmgr* bdd2, integer bits, integer cofbits)
  {
  one          = new BDDnode(-1,(BDDnode*)1,NULL);
  zero         = new BDDnode(-1,NULL,NULL);  // Create zero and one

  tempnode     = new BDDnode;                   // used as temp BDDnode in Makenode
  tempcomputed = new BDDCacheKey;               // used as temp BDDcomp. in ite

						// allocate the caches
  computed = new BDDCache( (1<<bits)-1, bits );
  if( cofbits > 0 )				// then one cache for unique & cof
    cofactor_hc  = new BDDCache( (1<<cofbits)-1, cofbits);
  else
    cofactor_hc = computed;
  restrict_hc  = cofactor_hc;

  one->SetOverflow();                             // Since refs are needed for
  zero->SetOverflow();                            // the 1st call to Makenode

  num_ite_calls = 0;                            // stats to get perf.
  num_nodes     = 2;                            // zero & one already allocated
  num_inputs    = 0;                            // keeps track of indices
  unique        = new BDDHasher(
		   bdd2->unique->lookup_index,
		   bdd2->unique->lookup_level); // BDD hashes
  defaultIA = new IndexArray( bdd2->defaultIA );
  num_keeps	= 0;
  keptNodes	= new HasherTree( bdd2->keptNodes, TRUE );

  timestamp     = 0;
  }

BDDmgr::~BDDmgr()
  {
  delete tempnode;
  delete tempcomputed;
  unique->Clear();
  delete unique;
  delete defaultIA;
  delete keptNodes;
  computed->Clear();
  cofactor_hc->Clear();
  restrict_hc->Clear();
  delete computed;
  if( cofactor_hc != computed )
    delete cofactor_hc;
  if( restrict_hc != computed && restrict_hc != cofactor_hc)
    delete restrict_hc;
  delete one;
  delete zero;
  }

integer
BDDmgr::GetLevel(BDDnode*f)
  { return unique->LookupLevel(f->GetIndex()); }

////////////////////////////////////////////////////////////////////////
//
// The following copies a function from the named bdd manager (bdd2)
// into the current bddmanager.
//

BDDnode*
BDDmgr::Copy2Manager( BDDnode* f2, BDDmgr *bdd2, HashBase* ht )
  {
  boolean new_ht = !ht;
  if( new_ht )
    ht = new HasherTree;

  int i,top = MAX( bdd2->NumInputs(), NumInputs() );

  // check to see if the bdd's have the same ordering!
  for( i=0 ; i<top ; i++ )
    if( unique->LookupIndex(i) != bdd2->unique->LookupIndex(i) )
      break;

  BDDnode *f;

  if( i<top )
    f = Copy2ManagerR( f2, bdd2, ht );
  else
    f = FastCopy2ManagerR( f2, bdd2, ht );

  if( new_ht )
    delete ht;

  return f;
  }


BDDnode*
BDDmgr::Copy2ManagerR( BDDnode* f2, BDDmgr *bdd2, HashBase* ht)
  {
  if( f2 == bdd2->One() )
    return One();

  if( f2 == bdd2->Zero() )
    return Zero();

  Object *o;
  if( ht->Get( &o, f2 ) )
    return (BDDnode*) o;

  BDDnode* T = Copy2ManagerR(f2->Then(), bdd2, ht);
  BDDnode* E = Copy2ManagerR(f2->Else(), bdd2, ht);

  BDDnode* f = ite( Input(f2->GetIndex()), T, E);
  if( f2->NumReferences() > 1 )
    ht->Put( f, f2 );
  return f;
  }

BDDnode*
BDDmgr::FastCopy2ManagerR( BDDnode* f2, BDDmgr *bdd2, HashBase* ht)
  {
  if( f2 == bdd2->One() )
    return One();

  if( f2 == bdd2->Zero() )
    return Zero();

  Object *o;
  if( ht->Get( &o, f2 ) )
    return (BDDnode*) o;

  BDDnode* T = FastCopy2ManagerR(f2->Then(), bdd2, ht);
  BDDnode* E = FastCopy2ManagerR(f2->Else(), bdd2, ht);

  BDDnode* f = MakeNode( f2->GetIndex(), T, E);
  if( f2->NumReferences() > 1 )
    ht->Put( f, f2 );
  return f;
  }

///////////////////////////////////////////////////////////////////////////
//
// This copies ALL the BDDnodes from one manager to the other, keeping
// reference counts the same, etc.
//

void
BDDmgr::Dump2Manager( BDDmgr *bdd2 ) {
  HasherTree done;

  for( int i=0 ; i<bdd2->NumInputs() ; i++ )
    unique->AddHasher();

  BDDHasherArray_ForEach_Type(bdd2->unique, node) {

    CopyNode(node, node, bdd2->One(), bdd2->Zero(), done);
    }BDDHasherArray_End_ForEach

  // anything left was "kept" by the user
  Hasher_ForEach_Type( &done, BDDnode, oldnode, newnode) {
    while( newnode->NumReferences() < ((BDDnode*)oldnode)->NumReferences() )
      newnode->Reference();
    }Hasher_End_ForEach

  }

/////////////////////////////////////////////////////////////////////////
//
// CopyNode is like MakeNode, but it assume the BDD structure passed
// to it is already canonical, and nice. For example, it doesn't check
// to see if a node already exists before Put'ing it.
// CopyNode is a private function, which can be used to copy one
// bdd structure to a new manager with minimal overhead. It assumes
// that all the nodes below the current "node" exist inside HasherTree
// "done" and removes them from the HasherTree when the reference
// count indicates that no more BDDnodes make use of them.
//

BDDnode*
BDDmgr::CopyNode( BDDnode* node, BDDnode *address,
		      BDDnode *bdd2_one, BDDnode *bdd2_zero, HashBase& done )
{
  if( node == bdd2_one  ) return One();
  if( node == bdd2_zero ) return Zero();

  BDDnode *l2 = node->Then();
  BDDnode *r2 = node->Else();
  
  BDDnode *l, *r;
  if( l2 == bdd2_one ) l = One();
  else if( l2 == bdd2_zero ) l = Zero();
  else
    l = (BDDnode*) done.Get(l2);
  
  if( r2 == bdd2_one ) r = One();
  else if( r2 == bdd2_zero ) r = Zero();
  else
    r = (BDDnode*) done.Get(r2);
  
  tempnode->index = (short) node->GetIndex();
  tempnode->right = r;
  tempnode->left  = l;
  BDDnode *n;
  if ( tempnode == (n=unique->ForceGet(tempnode)) ) {   // node exist in BDD?
    tempnode = new BDDnode;
    l->Reference();                                     // mark children for
    r->Reference();                                     // GC
    ++num_nodes;
  }
  
  done.Put(n,address);
  return n;
}

////////////////////////////////////////////////////////////////////////////
//
// Unmark a bdd (the whole tree). Assumes one and zero do not need to be
// unmarked
//

void
BDDmgr::UnMark( BDDnode *f ) {
  if( f==Zero() || f==One() )
    return;
  if( f->Marked() ) {
    UnMark( f->Then() );
    UnMark( f->Else() );
    f->UnMark();
    }
  return;
  }

////////////////////////////////////////////////////////////////////////////
//
// BDDnode member functions

integer
BDDnode::Compare(Comparable *p) {

  ASSERT( p->isa() == BDDNODE, "not a BDDnode");

  BDDnode *n = (BDDnode *)p;

  if ( (left == n->left) && (right == n->right) && (GetIndex() == n->GetIndex()) )
    return 0;
  return ((integer)this - (integer)n);
}

