#ifndef BDD2_H
#define BDD2_H

#include "unate_info.h"
#include "bv.h"
#include "bddm.h"
#include "floatobj.h"
#include "odlist.h"

// computation types 100-199 are reserved for BDDmanager
#define XRESTRICT_OP 104
#define XRESTRICT2_OP 105
#define CORRELATION_OP 106
#define DISTINGUISH_OP 107
#define ANDR_OP 108
#define SUPERAND_OP 109
#define SPECIAL_OP 199


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                    Andrew Seawright
//                                                      Forrest Brewer
//                                                      Chuck Monahan
//                                                      Ivan R.
//                                                      Andy Crews
//                                                      Tony Stornetta
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are implemented in this version, with the
// intent to minimize the amount of necessary code rewriting in existing
// software.
//
// The basic BDD package was originally designed at UCSB, and called HomeBrew.
//
////////////////////////////////////////////////////////////////////////////


class BDDmanager : public BDDmgr {

 public:
  BDDmanager( integer bits=19, integer cofbits=17 ) : BDDmgr( bits, cofbits) {;}
  BDDmanager( BDDmanager* bdd2, integer bits=19, integer cofbits=17 ) :
    BDDmgr( bdd2, bits, cofbits ) {;}
 ~BDDmanager() {;}

  BDDnode* Inputs(integer n)    { return Input(n);}
  BDDnode* ReadCubes(istream& in=cin) { return BDDmgr::ReadEq( in);}
  void PrintStructure(BDDnode *, ostream& out = cout);
  void PrintLabeledCubes(BDDnode *, char **var_names, int name_len=1,
			 char eol='\n', ostream& out = cout);

  void GCfast() { doGCfast(TRUE); }
  void ForceGC()	{ BDDmgr::GC(); }

  // bdd core2 functions

  BDDnode* Constrain (BDDnode *f, BDDnode *c) { return Cofactor (f,c); }
  BDDnode* SupersetAnd( BDDnode *, BDDnode *, IndexArray *ia );
  BDDnode* VariableShift(BDDnode* f, integer shift) {
    return BDDmgr::VariableShift(f,shift); }
  BDDnode* VariableShift(BDDnode* f, integer shift, IndexArray *ia) {
    return BDDmgr::VariableShift(f,shift,ia); }

  BDDnode* VariableShift(BDDnode* f, integer shift, integer max, integer min=0) {
    IndexArray ia(min,max);
    return BDDmgr::VariableShift(f, shift, &ia);
    }
  
  BDDnode* Smooth(BDDnode* f, IndexArray*ia, HashBase* ht=NULL, boolean InList=TRUE)
    { return BDDmgr::Smooth(f,ia,ht,InList); }

  BDDnode* Smooth(BDDnode*f, integer a, integer b, integer m=1, integer o=0, HashBase* ht=NULL) {
    IndexArray ia(a,b,m,o);
    return BDDmgr::Smooth(f, &ia, ht, TRUE);
    }
  BDDnode* Smooth(BDDnode*f, ODList*list, HashBase* ht=NULL, boolean InList=TRUE ) {
    if (list->Count() == 0) {
      IndexArray ia(list);
      f=BDDmgr::Smooth(f,&ia,ht,InList);
      }
    return f;
    }

  BDDnode* CondConstraint (BDDnode *f, BDDnode *v, BDDnode *g) { return and( f, ite( v, g, One()));}
#define VECTOR_AND_DEBUG (int) 0x1
#define VECTOR_AND_DO_COF (int) 0x2
  BDDnode* VectorAnd( List *f, int max_nodes=INT_MAX, int flags=0 );
  BDDnode* VectorOr( List *f, int max_nodes=INT_MAX, int flags=0 );
  BDDnode* BuildHamming(BDDnode* minterm, BDDnode* constraint,integer distance);
  BDDnode *DC2Zero(BDDnode* f, IndexArray* ia=NULL, HashBase* ht=NULL) {
    return BDDmgr::DC2Zero(f,ia,ht); }
  BDDnode *DC2One(BDDnode* f, IndexArray* ia=NULL, HashBase* ht=NULL) {
    return BDDmgr::DC2One(f,ia,ht); }
  BDDnode* DC2Zero(BDDnode*, integer index, HashBase* ht=NULL);  
  BDDnode* DC2One (BDDnode*, integer index, HashBase* ht=NULL);  
  IndexArray* Minterm2IA( BDDnode* set );
  BDDnode* MakeMin2N (int n, IndexArray* ia=NULL); 
  BDDnode* Int2BDD( int n, IndexArray* ia=NULL );

  BDDnode* ExactlyN( integer n, IndexArray* ia=NULL) { return Tuple( n, ia); }
  BDDnode* SetUnion( BDDnode*, IndexArray*ia=NULL, HashBase* ht=NULL);
  BDDnode* ComplementBits( BDDnode *, IndexArray* ia=NULL);
  BDDnode* Distinguish(BDDnode *s1, BDDnode *s2, int max, IndexArray *va=NULL, HashBase *ht=NULL);
  BDDnode* Distinguish(BDDnode *s, int max, IndexArray *va, HashBase *ht=NULL);
  BDDnode* DistinguishCubes( BDDnode *c1, BDDnode *c2, IndexArray *v );

  int      MinCube( BDDnode *s, HashBase *ht=NULL );
  int      MaxCube( BDDnode *s, IndexArray *vars, HashBase *ht=NULL );


  // experimental core functions
  BDDnode* XRestrict (BDDnode *, BDDnode *, boolean greedy=FALSE); // experimental restrict
  BDDnode* AndRestrict(BDDnode* f, BDDnode* g, BDDnode* c);
  BDDnode* AndSmooth  (BDDnode* f, BDDnode* g, IndexArray* ia, HashBase* hc);
  BDDnode* AndSmoothRestrict(BDDnode* f, BDDnode* g, BDDnode* c, IndexArray* ia, HashBase* hc, HashBase* sm);

  // bdd statistics functions
  integer  NodesInFunction( BDDnode *f )   { return NumNodes(f); }
  integer  PathsInFunction(BDDnode *f)     { return NumCubes(f); }
  integer  PathsToOne(BDDnode *f)          { return NumCubes(f); }
  integer  OnSetSize(BDDnode *f, integer n) { return NumMinterms(f,n); }
  double   NumMintermsD(BDDnode *, integer, HasherTree* = NULL);
  float    Correlation(BDDnode* f1, BDDnode* f2, HashBase* corr_ht = 0);

  boolean    Evaluate(BDDnode*, BitVector *);
  BDDnode*   Code2Cube(List* functions, integer code);
  BitVector* DCterms ( BDDnode* f, integer bot_index=0, integer top_index=-1);

  UnateInfo** FormUnateArray(BDDnode*, integer, UnateInfo**);
  BitVector   FindOnePath(BDDnode*);
  BitVector   FindOnePath1(BDDnode*, integer);
  int         AllButOneSameOrDiff( BDDnode *, BDDnode * );
  BDDnode *   CommonZeros( BDDnode *, BDDnode *, int &n1, int &n0 );

  BDDnode* AtLeastN( integer n, IndexArray* ia, BDDnode* pattern);
  BDDnode* AtLeastN( integer n, IndexArray* ia=NULL) { return BDDmgr::AtLeastN(n,ia); }
  BDDnode* UpToN( integer n, IndexArray* ia, BDDnode* pattern);
  BDDnode* UpToN( integer n, IndexArray* ia=NULL) { return BDDmgr::UpToN(n,ia); }
  BDDnode* Tuple( integer n, IndexArray* ia, BDDnode* pattern);
  BDDnode* Tuple( integer n, IndexArray* ia=NULL) { return BDDmgr::Tuple(n,ia); }

  BDDnode* Reach1( BDDnode *ss, BDDnode *rel, IndexArray *ps, IndexArray *ns );
  BDDnode* SteadyState1( BDDnode *ss, BDDnode *rel, IndexArray *ps, IndexArray *ns );

friend class ZBDDmanager;

protected:
  BDDnode* SupersetAndR( BDDnode *, BDDnode *, IndexArray *ia, HasherTree *ht );
  BDDnode* AndSmoothR(BDDnode* f, BDDnode* g, IndexArray* ia, HashBase* hc, HashBase* sm);
  void     NumMintermsDR(BDDnode *, HasherTree*, double& p, integer& h);
  BDDnode *DC2ZeroR(BDDnode* f, IndexArray* ia, HashBase* ht)
    { return BDDmgr::DC2ZeroR(f,ia,ht); }
  BDDnode *DC2OneR (BDDnode* f, IndexArray* ia, HashBase* ht)
    { return BDDmgr::DC2OneR(f,ia,ht); }
  BDDnode *DC2OneR (BDDnode*, integer level, HashBase* ht);
  BDDnode *DC2ZeroR(BDDnode*, integer level, HashBase* ht);
  BDDnode *SetUnionR( BDDnode*, IndexArray*, HashBase* ht);
  BDDnode *MakeMin2NR( int n, int place, IndexArray* ia);
  BDDnode *DistinguishR(BDDnode *s1, BDDnode *s2, int max, HashBase *ht, IndexArray *ia, BDDnode **t);
  BDDnode *DistinguishR(BDDnode *s1, int max, IndexArray *ia, HashBase *ht, BDDnode **t);
  int      MinCubeR( BDDnode *s, HashBase *ht );
  int      MaxCubeR( BDDnode *s, IndexArray *vars, HashBase *ht );
  boolean  XRestrictIgnoreVar( BDDnode *fchild, BDDnode *f, BDDnode *c, BDDnode *cchild );
  BDDnode *XRestrictIgnoreVarR( BDDnode *f, BDDnode *g, integer &count );
  boolean  XRestrictRedCheck( BDDnode *f, BDDnode *c, BDDnode* f1, BDDnode *c1,
				integer &count1, integer &count2 );
  BitVector* DCtermsR( BDDnode* f, BitVector* bv, integer bot, integer top, HashBase* ht);

  void     FormUnateArrayR(BDDnode*, UnateInfo**, HashBase*);
  boolean  FindOnePathR(BDDnode*, BitVector&);

  void     PrintStructureR(BDDnode *, HasherTree *ht, ostream& out = cout);
  void     PrintLabeledCubesR(BDDnode *, string, ostream& out = cout, char eol='\n',
			      char **var_names=NULL, int pos=0);

  void	   RemoveNodeR( BDDnode* node );

  float    CorrelationR(BDDnode* f1, BDDnode* f2, HashBase* corr_ht);
  void     doGCfast(boolean clear_caches);
  void     RGCfast(BDDnode *);

  BDDnode *MakePatternR( IndexArray* vars, integer ipos, BDDnode* pattern, BDDnode* troot, BDDnode* froot);

};

#endif

