#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                    Andrew Seawright
//                                                      Forrest Brewer
//                                                      Chuck Monahan
//                                                      Ivan R.
//                                                      Andy Crews
//                                                      Tony Stornetta
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are implemented in this version, with the
// intent to minimize the amount of necessary code rewriting in existing
// software.
//
// The basic BDD package was originally designed at UCSB, and called HomeBrew.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
//
// The purpose of MakeNode is to ensure that the nodes which are
// created are unique and consistent. So it checks
// to see if the new node is unique, and if it is a literal, it
// also checks to see if it is less than the max literal. If not,
// it generates new literals up to that point to ensure that the
// addresses returned by Input and iterators are always valid. If
// not a literal, it make a new node, and updates the references.
//
BDDnode *
BDDmgr::MakeNode(integer i, BDDnode *g, BDDnode *h) {

  BDDnode *r;

  ASSERT(i>=0, "negative variable index");
  ASSERT( g != h, "MakeNode called with identical children" );

  //Check if a out of range variable is requested
  for (; num_inputs <= i; num_inputs++) {
    unique->AddHasher( );
    defaultIA->Put( i );
    }
  
  //in-bounds variable or function
  tempnode->index = (short) i;
  tempnode->right = h;
  tempnode->left = g;					// make one...
  if ( tempnode == (r=unique->ForceGet(tempnode)) ) {	// node exist in BDD?
    tempnode = new BDDnode;
    g->Reference();					// mark children for
    h->Reference();					// GC
    ++num_nodes;
  }
  return r;
} // end MakeNode()

////////////////////////////////////////////////////////////////////////////
//
// This ite() function is essentially the core of all operations on the ROBDD
// structure.
//

BDDnode*
BDDmgr::ite(BDDnode *f, BDDnode *g, BDDnode *h) {

  ++num_ite_calls;

    //
   //  check for terminal cases
  //
  
  if (f == one)
    return g;
  if (f == zero)
    return h;

  if (f==g) 
    g = one;
  if (f==h)
    h = zero;

  if (g == one && h == zero)
    return f;
  if (g == h)  // independent of f
    return g;

     //
    // do as much "standard triple" stuff as can be done without
   //  complement edges
  //
  
  if (g==one) {
    if ( (integer)f > (integer)h ) {
      SWAP( BDDnode*, f, h);
    }
  } else if (h==zero) {
    if ( (integer)f > (integer)g ) {
      SWAP( BDDnode*, f, g);
    }
  }

    //
   //  have we already computed this function ?
  //

  BDDnode     *o;

  tempcomputed->Set(ITE_OP,f,g,h);

  if (computed->Get(&o, *tempcomputed)) {
    return o;
  }

     //
    // recursively compute ite()
   //

  integer top;

  if( GetLevel(f) > GetLevel(g) )
    top = (GetLevel(f) > GetLevel(h))? f->GetIndex() : h->GetIndex();
  else
    top = (GetLevel(g) > GetLevel(h)) ? g->GetIndex() : h->GetIndex();

  BDDnode *t, *e, *r;

  if ( h == Zero()) {
    t = and(Shannon1(f, top), Shannon1(g, top));
    e = and(Shannon0(f, top), Shannon0(g, top));
  } else {
    t = ite(Shannon1(f, top), Shannon1(g, top), Shannon1(h, top));
    e = ite(Shannon0(f, top), Shannon0(g, top), Shannon0(h, top));
  }

  r = (t == e)  ? t : MakeNode(top, t, e);

  tempcomputed->Set(ITE_OP,f,g,h);
  computed->Replace(r, *tempcomputed); // store computed ite() in cache

  return r;
}


////////////////////////////////////////////////////////////////////////////
//
// This and() function is essentially the ite function with some fine tuning
//

BDDnode*
BDDmgr::and(BDDnode* f, BDDnode* g) { 

  ++num_ite_calls;

    //
   //  check for terminal cases
  //
  
  if (f == one || f == g)
    return g;
  if (g == one)
    return f;
  if (f == zero || g == zero)
    return zero;

    //
   //  have we already computed this function ?
  //

  if ( (integer)f > (integer)g ) {         //  order for computed
    SWAP( BDDnode*, f, g);
  }

  BDDnode* o;
  tempcomputed->Set(ITE_OP,f,g,Zero());

  if (computed->Get(&o, *tempcomputed)) {
    return o;
  }

     //
    // recursively compute ite()
   //

  integer top = (GetLevel(f) > GetLevel(g))? f->GetIndex() : g->GetIndex();

  BDDnode *t, *e, *r;

  t = and(Shannon1(f, top), Shannon1(g, top));
  e = and(Shannon0(f, top), Shannon0(g, top));

  /* small change made to improve cache performance */

  r =  (t == e)  ? t : MakeNode(top, t, e);

  tempcomputed->Set(ITE_OP,f,g,Zero());
  computed->Replace(r, *tempcomputed); // store computed ite() in cache

  return r;
}
