#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <limits.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                    Andrew Seawright
//                                                      Forrest Brewer
//                                                      Chuck Monahan
//                                                      Ivan R.
//                                                      Andy Crews
//                                                      Tony Stornetta
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are implemented in this version, with the
// intent to minimize the amount of necessary code rewriting in existing
// software.
//
// The basic BDD package was originally designed at UCSB, and called HomeBrew.
// Inverted edges were implemented with the help of the C BDD package from CMU.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
//
// Safe versions of shannon, which call cofactor if the index is below the
// current top node.
//

BDDnode*
BDDmgr::ShannonT(BDDnode* n, integer v)
  {
    if( unique->LookupLevel(v) == GetLevel(n) )
      return n->Then();
    else if(unique->LookupLevel(v) > GetLevel(n) )
      return n;
    else
      return Cofactor(n, Input(v));
  }

BDDnode*
BDDmgr::ShannonF(BDDnode* n, integer v)
  {
    if( unique->LookupLevel(v) == GetLevel(n) )
      return n->Else();
    else if(unique->LookupLevel(v) > GetLevel(n) )
      return n;
    else
      return Cofactor(n, not(Input(v)));
  }

////////////////////////////////////////////////////////////////////////////
//
// Intersection report whether the two functions overlap.
//

boolean
BDDmgr::Intersect( BDDnode* f, BDDnode* g, HashBase* ht) {
  boolean new_ht = (ht==NULL);
  if ( new_ht)
    ht = new HasherTree(TRUE);
  boolean     r  = IntersectR( f, g, ht);

  if ( new_ht)
    delete ht;
  return r;
}

boolean
BDDmgr::IntersectR( BDDnode* f, BDDnode* g, HashBase* ht) {

  ++num_ite_calls;

    //
   //  check for terminal cases
  //
  
  if (f == Zero() || g == Zero())  // nothing intersects zero
    return FALSE;
  if (f == One()  || g == One())   // since Zeros are gone.. always True
    return TRUE;
  if (f == g)                      // ditto!!
    return TRUE;

     //
    // do as much "standard triple" stuff as can be done without
   //  complement edges
  //

  if ( (integer)f > (integer)g ) {
    SWAP( BDDnode*, f, g);
  }

    //
   //  have we already computed this function ?
  //

  tempcomputed->Set( ITE_OP, f, g, zero);

  // precomputed globally?
  BDDnode* prv_func;
  if (computed->Get(&prv_func, *tempcomputed)) {
    return prv_func != Zero();
  }
  // precomputed locally?
  if ( f->NumReferences() > 1 && g->NumReferences() > 1) {
    Object*      o;
    if (ht->Get( &o, tempcomputed)) 
      return FALSE;
  }

     //
    // recursively compute ite()
   //

  integer top = (GetLevel(f) > GetLevel(g))? f->GetIndex() : g->GetIndex();

  if ( IntersectR(Shannon1(f, top), Shannon1(g, top), ht)) {
    return TRUE;
  }
  if ( IntersectR(Shannon0(f, top), Shannon0(g, top), ht)) {
    return TRUE;
  }

  if ( f->NumReferences() > 1 && g->NumReferences() > 1) {
    tempcomputed->Set( ITE_OP, f, g, zero);
    ht->Put( new BDDCacheKey( *tempcomputed));
  }

  return FALSE;
}

////////////////////////////////////////////////////////////////////////////
//
// Subset report whether the function g lies entirely within f
//
////////////////////////////////////////////////////////////////////////////

boolean
BDDmgr::Subset( BDDnode* f, BDDnode* g, HashBase* ht) {
  boolean new_ht = (ht==NULL);
  if ( new_ht)
    ht = new HasherTree(TRUE);
  boolean r  = SubsetR( f, g, ht);

  if ( new_ht)
    delete ht;
  return r;
}

boolean
BDDmgr::SubsetR( BDDnode* f, BDDnode* g, HashBase* ht) {

//  ++num_ite_calls;

    //
   //  check for terminal cases
  //
  
  if (f == g) 
    return TRUE;
  if (f == One())
    return TRUE;
  if (f == Zero())
    return FALSE;

  if (g == One())
    return FALSE;
  if (g == Zero())
    return TRUE;

    //
   //  have we already computed this function ?
  //

  // is it stored globally? 
  if ( (integer)f > (integer)g ) {
    tempcomputed->Set(ITE_OP,g,f,Zero());
  } else {
    tempcomputed->Set(ITE_OP,f,g,Zero());
  }

  BDDnode* prv_func;
  if (computed->Get(&prv_func, *tempcomputed)) {
    return (prv_func == g);
  }

  // is it stored locally?
  if ( f->NumReferences() > 1 && g->NumReferences() > 1) {
    tempcomputed->Set(ITE_OP,f,g,Zero());
    Object* o;
    if ( ht->Get( &o, tempcomputed))
      return TRUE;
  }
     //
    // recursively compute ite()
   //

  integer top = (GetLevel(f) > GetLevel(g))? f->GetIndex() : g->GetIndex();

  if ( !SubsetR(Shannon1(f, top), Shannon1(g, top), ht))
    return FALSE;
  if ( !SubsetR(Shannon0(f, top), Shannon0(g, top), ht))
    return FALSE;

  if ( f->NumReferences() > 1 && g->NumReferences() > 1) {
    tempcomputed->Set( ITE_OP, f, g, zero);
    ht->Put( new BDDCacheKey( *tempcomputed)); // store local hash
  }

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//
// This routine builds a new function that is identical to 'f' except
// that the variable indexes it is built on are equal to 'f''s shifted by the
// parameter 'shift'. Care must be taken that this shifting will not imply
// negative variable indexes - if this should occur a fatal error will
// be generated. This routine has time complexity |f|.
//

void
BDDmgr::OrderIndexes( IndexArray* ia ) { 
  ia->Order( unique->lookup_level, timestamp );
}


BDDnode *
BDDmgr::VariableShift(BDDnode *f, integer shift) {

  HasherTree *ht = new HasherTree;
  f = VariableShift(f, shift, ht);
  delete ht;
  return f;
}

BDDnode *
BDDmgr::VariableShift(BDDnode *f, integer shift, HashBase *ht) {

  if (f==one || f==zero)
    return f;

  Object *o;
  if (ht->Get(&o, f))		//Have we been here before?
    return (BDDnode *) o;	// yup

  BDDnode *g, *h, *r;

  g = VariableShift(f->Then(),  shift, ht);
  h = VariableShift(f->Else(), shift, ht);

  integer v = f->GetIndex() + shift;
  ASSERT (v>=0, "variable shift to negative index");

// this change is necessary for dynamic reordering of bdd's
  integer l = unique->LookupLevel(v);
  if( l > GetLevel(g) && l > GetLevel(h) )	// save ite calls, if user
    r = MakeNode(v,g,h);			// was smart about shifting
  else
    r = ite(Input(v),g,h);
  ht->Put(*r, f);
  return r;
}

BDDnode*
BDDmgr::VariableShift(BDDnode* f, integer shift, IndexArray* ia) {
  if (ia->Count() == 0)
    return f;

  HasherTree *ht = new HasherTree;
  OrderIndexes( ia);
  f = VariableShift(f, shift, ht, ia);
  delete ht;
  return f;
}

// private function... do not check for empty ia, or order ia

BDDnode *
BDDmgr::VariableShift(BDDnode *f, integer shift, HashBase *ht, IndexArray* ia) {
  if (f==one || f==zero || GetLevel(f) < ia->LowestLevel())
    return f;

  Object *o;
  if (ht->Get(&o, f))		//Have we been here before?
    return (BDDnode *) o;	// yup

  BDDnode *g, *h, *r;

  g = VariableShift(f->Then(), shift, ht, ia);
  h = VariableShift(f->Else(), shift, ht, ia);

  integer v = ( ia->Find(GetLevel(f)) ) ? f->GetIndex()+shift :f->GetIndex();
  ASSERT (v>=0, "variable shift to negative index");

  integer l = unique->LookupLevel(v);
  if( l > GetLevel(g) && l > GetLevel(h) )	// save ite calls, if user
    r = MakeNode(v,g,h);			// was smart about shifting
  else
    r = ite( Input(v),g,h);
  ht->Put(*r, f);
  return r;
  }


////////////////////////////////////////////////////////////////////////////
//
// Range computation algorithm - this routine computes the range of a
// multiple output boolean function represented by a list F of single output
// boolean functions. 
//

BDDnode *
BDDmgr::Range(List *F) {
  if (F->Count() == 0)  // terminating case
    return one;

  BDDnode *f;
  BDDsubstitution *p;
  int i;
  for ( i=0; i< F->Count(); ++i) {
    p = (BDDsubstitution *) F->Pop();
    if (p->function == one) {
      f = and( p->variable, Range(F));
      F->Append(*p);
      return f;
    }
    if (p->function == zero) {
      f = and_not( Range(F), p->variable );
      F->Append(*p);
      return f;
    }
    F->Append(*p);
  }
    
  BDDsubstitution* x;
  BDDsubstitution* y;
  List* G = new List;
  List* H = new List;
  
  p = (BDDsubstitution *) F->Pop();
  List_ForEach_Type(*F, BDDsubstitution, q) {
    
    x = new BDDsubstitution;
    x->function = Cofactor(q->function, p->function);    
    x->variable = q->variable;
    
    y = new BDDsubstitution;
    y->function = Cofactor(q->function, not( p->function));    
    y->variable = q->variable;   
    
    G->Add(*x);
    H->Add(*y);
    
  } List_End_ForEach;
  
  f = ite( p->variable, Range(G), Range(H));
  
  H->Clear(); delete H;
  G->Clear(); delete G;
  
  F->Append(*p);
  return f;
}

////////////////////////////////////////////////////////////////////////////
//
// Generalized Cofactoring algorithm, should be the same as Coudert-Madre
// constraint operator, however in this routine the top variable is
// assumed to be the MAX or the top vars of 'f' and 'c'.
//

BDDnode *
BDDmgr::Cofactor(BDDnode *f, BDDnode *c) {

  ASSERT(c != zero, "Cofactor(): assertion that c!=0 failed");

  if (f == c)
    return One();
  if (c == one || f == one || f == zero)
                                { return f; }
  integer          v;
  BDDnode        *o, *c0, *c1;
  tempcomputed->Set( COFACTOR_OP, f, c );

  v = (GetLevel(f) > GetLevel(c)) ? f->GetIndex() : c->GetIndex();

  if (cofactor_hc->Get(&o,*tempcomputed)==0) {
    c0 = Shannon0(c,v);
    c1 = Shannon1(c,v);
    if (c0 == zero) {
      o = Cofactor(Shannon1(f, v), c1);

    } else  if (c1 == zero) {
      o = Cofactor(Shannon0(f, v), c0);

    } else {
      c1 = Cofactor(Shannon1(f,v), c1);
      c0 = Cofactor(Shannon0(f,v), c0);
      o = (c0 == c1 ) ? c0 : MakeNode( v, c1, c0);
    if( f->NumReferences() > 1 || c->NumReferences() > 1)
      {
      tempcomputed->Set( COFACTOR_OP, f, c );
      cofactor_hc->Replace(o,*tempcomputed); 
      }
    }
  }
  return o;
}


////////////////////////////////////////////////////////////////////////
// Restrict is a slightly different version of cofactor that can
// simplify the result quite a bit. ref: Coudert and Madre--
// 	A Unified Framework for the Formal Verification of Sequential Cir.
//		1990 IEEE


BDDnode *
BDDmgr::Restrict(BDDnode *f, BDDnode *c) {

  ASSERT(c != zero, "Restrict(): assertion that c!=0 failed");

  if (f == c)
    return One();
  if (c == one || f == one || f == zero)
                                { return f; }
  integer v = ( GetLevel(f) > GetLevel(c) ) ? f->GetIndex() : c->GetIndex();

  BDDnode* o, *c0, *c1;

  tempcomputed->Set(RESTRICT_OP, f, c);
  if (restrict_hc->Get(&o, *tempcomputed)==0)
    {
    c0 = Shannon0(c,v);
    c1 = Shannon1(c,v);
    if (c0 == zero)
      o = Restrict(Shannon1(f, v), c1);

    else if (c1 == zero)
      o = Restrict(Shannon0(f, v), c0);

    else
      {

      ///////////////////////////////////////////////////////////////
      // If index(c) > index(v) then we will walk down c with
      // Shannon expansions, and not f, so smooth c's node and
      // continue to cofactor.

      if( GetLevel(c) > GetLevel(f) )
        o = Restrict( f,or( c0, c1 ));

      ///////////////////////////////////////////////////////////////
      // Walk down f (and maybe c too), so just do normal cofactor.
      else {
	c1 = Restrict(Shannon1(f,v), c1);
	c0 = Restrict(Shannon0(f,v), c0);
	o = (c0 == c1 ) ? c0 : MakeNode(v, c1, c0);
	}
      if( f->NumReferences() > 1 && c->NumReferences() > 1)
	{
	tempcomputed->Set( RESTRICT_OP, f, c );
	restrict_hc->Replace(o,*tempcomputed);
	}
      }
  }
  return o;
}


////////////////////////////////////////////////////////////////////////////
//
// Routine to compose Boolean functions. Operates on function f by
// substituting, for subfunction v of f, the function g.
//

BDDnode *
BDDmgr::Compose(BDDnode* f, BDDnode* v, BDDnode* g) {

  BDDnode *f1, *f2;
  f1 = Cofactor(f, v);
  f2 = Cofactor(f, not(v));
  return (f1!=f2) ? ite(g, f1, f2) : f;
}


////////////////////////////////////////////////////////////////////////////
//
// Existential Abstraction or Smoothing operator
//
// For the usual use the assumption is that List X holds pointers to functions
// that are simple single variable functions.
//       __
//         |
//        -|  f  =   f   +   f
//       __|x         x       ~x
//            
//
//

BDDnode *
BDDmgr::ExistentialAbstraction (BDDnode* f, List *X) {

  List_ForEach_Type(*X, BDDnode, n) {
    f = or(Cofactor(f,n), Cofactor(f, not(n)));
  } List_End_ForEach;
  return f;
}

BDDnode*
BDDmgr::ExistentialAbstraction (BDDnode* f, BDDnode* n) {

  return or(Cofactor(f,n), Cofactor(f, not(n)));
}

////////////////////////////////////////////////////////////////////////////
//
// Universal Abstraction or Consensus operator
//
// For the usual use the assumption is that List X holds pointers to functions
// that are simple single variable functions.
//
//     \    /
//      \--/  f  =   f  . f
//       \/x          x    ~x 
//       
//

BDDnode *
BDDmgr::UniversalAbstraction (BDDnode* f, List *X) {

  List_ForEach_Type(*X, BDDnode, n) {
    f = and(Cofactor(f,n), Cofactor(f, not(n)));
  } List_End_ForEach;
  return f;
}


/////////////////////////////////////////////////////////////////////////
//
// Smooth out all variables v in the closed interval [a, b] that satisfy
// the relation (v%m) == i.
//


BDDnode*
BDDmgr::Smooth( BDDnode *f, IndexArray *ia, HashBase *hc, boolean InList)
  {
  if (ia==NULL) ia=defaultIA;
  if (ia->Count() == 0)
    return f;

  boolean test = (hc==NULL);
  OrderIndexes( ia);

  if( test )
    hc = new HasherTree;
  
  BDDnode* t = SmoothR(f, ia, hc, InList);

  if( test )
    delete hc;
  return t;
  }


BDDnode *
BDDmgr::SmoothR(BDDnode *f, IndexArray *ia, HashBase* hc, boolean InList ) {

  if (f == one || f == zero)
    return f;

  Object* o;

  integer  v       = GetLevel(f);
  if ( v < ia->LowestLevel() )
    return (InList)? f: One();       // Either leave or smooth the list
  
  boolean s_level = ia->Find( v);     // Is this a level that we want to smooth?
  if( !InList )
    s_level = !s_level;

  if( f->NumReferences() > 1 && hc->Get(&o, f) )
    return (BDDnode*)o;

  BDDnode *g, *h, *g1, *h1;
  g = f->Then();
  h = f->Else();

  g1 = SmoothR(g, ia, hc, InList);
  h1 = SmoothR(h, ia, hc, InList);

  if (g1 != h1){
    if ( s_level)
      g1 = or( g1, h1 );
    else if (g == g1 && h == h1)
      g1 = f;
    else
      g1 = MakeNode( f->GetIndex(), g1, h1);
    }
  if( f->NumReferences() > 1 )
    hc->Put((Object*)g1, f);
  return g1;
}

BDDnode*
BDDmgr::USmooth( BDDnode *f, IndexArray *ia, HashBase *hc, boolean InList)
  {
  if (ia==NULL) ia=defaultIA;
  boolean test = (hc==NULL);
  OrderIndexes( ia);

  if( test )
    hc = new HasherTree;
  
  BDDnode* t = USmoothR(f, ia, hc, InList);

  if( test )
    delete hc;
  return t;
  }

BDDnode *
BDDmgr::USmoothR(BDDnode *f, IndexArray *ia, HashBase* hc, boolean InList ) {

  if (f == one || f == zero)
    return f;

  Object* o;

  integer  v       = GetLevel(f);
  if ( v < ia->LowestLevel() )
    return (InList)? f: Zero();     // Either leave or Usmooth the list
  
  boolean s_level = ia->Find( v);    //Is this a level that we want to smooth?
  if( !InList )
    s_level = !s_level;

  if( f->NumReferences() > 1 && hc->Get(&o, f) )
    return (BDDnode*)o;

  BDDnode *g, *h, *g1, *h1;
  g = f->Then();
  h = f->Else();

  g1 = USmoothR(g, ia, hc, InList);
  h1 = USmoothR(h, ia, hc, InList);

  if (g1 != h1){
    if ( s_level)
      g1 = and( g1, h1 );
    else if (g == g1 && h == h1)
      g1 = f;
    else
      g1 = MakeNode( f->GetIndex(), g1, h1);
    }
  if( f->NumReferences() > 1 )
    hc->Put((Object*)g1, f);
  return g1;
}

//////////////////////////////////////
//
// The function returns a list of cubes which exist in function f.
// Use with caution!!  There could be an exponential number of
// cube.  Use NumCubes, to see what sort of trouble you are getting 
// yourself into ;-)
//

List*
BDDmgr::ListCubes(BDDnode* f, IndexArray* ia) {
  OrderIndexes(ia);

  List* cubes = new List();
  char* cube  = new char[NumInputs()+1];
  for (int i=0; i<NumInputs(); i++) {cube[i] = '-';cube[i+1] = '\0';}

  ListCubesR( f, cubes, cube, ia);
  delete cube;
  return cubes;
}


void
BDDmgr::ListCubesR(BDDnode* f, List* cubes, char* cube, IndexArray* ia) {
  if ( f == Zero()) return;
  
  if ( f == One() || GetLevel(f) < ia->LowestLevel()) {
    cubes->AppendUnique( ReadCube(cube));    // ia may cause cubes to look the same
    return;
  }
  
  if ( ia->Find( GetLevel(f))) {
    cube[f->GetIndex()] = '1';
    ListCubesR( f->Then(), cubes, cube, ia);
    cube[f->GetIndex()] = '0';
    ListCubesR( f->Else(), cubes, cube, ia);
    cube[f->GetIndex()] = '-';
  } else {
    ListCubesR( f->Then(), cubes, cube, ia);
    ListCubesR( f->Else(), cubes, cube, ia);
  }
}

///////////////////////////////////////////
// This routine takes a function and replaces don't care elements
// with zero terms.  It replaces only those elements whose indexes 
// are listred in the IndexArray

BDDnode*
BDDmgr::DC2Zero(BDDnode* f, IndexArray *ia, HashBase* ht) {
  if (ia==NULL) ia=defaultIA;
  boolean new_ht = (ht == NULL);
  if (new_ht)
    ht = new HasherTree;

  OrderIndexes( ia);
  BDDnode* res = DC2ZeroR( f, ia, ht);  

  int i;
  for ( i=GetLevel(res)+1 ; i<=ia->HighestLevel(); i++) {
    if  (ia->Find(i) && res != Zero()) {
      integer index = unique->LookupIndex(i);
      res = MakeNode( index, Zero(), res);
    }
  }

  if (new_ht)
    delete ht;
  return res;
}

BDDnode*
BDDmgr::DC2ZeroR(BDDnode* f, IndexArray *ia, HashBase* ht) {
  integer level = GetLevel(f);

  if( level <= ia->LowestLevel() )
    return f;

  Object* res = ht->Get( f);
  if ( res!=NULL)
    return (BDDnode*) res;

  integer i;
  BDDnode* tru = DC2ZeroR( f->Then(), ia, ht);
  for ( i=GetLevel(tru)+1 ; i<=MIN( ia->HighestLevel(), level-1) ; i++) {
    if  (ia->Find(i) && tru != Zero()) {
      integer index = unique->LookupIndex(i);
      tru = MakeNode( index, Zero(), tru);
    }
  }
  
  BDDnode* fal = DC2ZeroR( f->Else(), ia, ht);
  for ( i=GetLevel(fal)+1 ; i<=MIN( ia->HighestLevel(), level-1) ; i++) {
    if  (ia->Find(i) && fal != Zero()) {
      integer index = unique->LookupIndex(i);
      fal = MakeNode( index, Zero(), fal);
    }
  }
  
  // tru can not equal fal.  When in doubt, try to build an example
  BDDnode* r = MakeNode( f->GetIndex(), tru, fal);
  if( f->NumReferences() > 1)
    ht->Put( r, f);
  return r;
}

//////////////////////////////////////

BDDnode*
BDDmgr::DC2One(BDDnode* f, IndexArray* ia, HashBase* ht) {
  if (ia==NULL) ia=defaultIA;
  boolean new_ht = (ht == NULL);
  if (new_ht)
    ht = new HasherTree;

  OrderIndexes( ia);
  BDDnode* res = DC2OneR( f, ia, ht);  

  int i;
  for ( i=GetLevel(res)+1 ; i<=ia->HighestLevel(); i++) {
    if  (ia->Find(i) && res != Zero()) {
      integer index = unique->LookupIndex(i);
      res = MakeNode( index, res, Zero());
    }
  }
  
  if (new_ht)
    delete ht;
  return res;
}

BDDnode* 
BDDmgr::DC2OneR(BDDnode* f, IndexArray* ia, HashBase* ht) {
  integer level = GetLevel(f);

  if( level <= ia->LowestLevel() )
    return f;

  Object* res = ht->Get( f);
  if ( res!=NULL)
    return (BDDnode*) res;

  integer i;
  BDDnode* tru = DC2OneR( f->Then(), ia, ht);
  for ( i=GetLevel(tru)+1 ; i<=MIN( ia->HighestLevel(), level-1) ; i++) {
    if  (ia->Find(i) && tru != Zero()) {
      integer index = unique->LookupIndex(i);
      tru = MakeNode( index, tru, Zero());
    }
  }
  
  BDDnode* fal = DC2OneR( f->Else(), ia, ht);
  for ( i=GetLevel(fal)+1 ; i<=MIN( ia->HighestLevel(), level-1) ; i++) {
    if  (ia->Find(i) && fal != Zero()) {
      integer index = unique->LookupIndex(i);
      fal = MakeNode( index, fal, Zero());
    }
  }
  
  // tru can not equal fal.  When in doubt, try to build an example
  BDDnode* r = MakeNode( f->GetIndex(), tru, fal);
  if( f->NumReferences() > 1)
    ht->Put( r, f);
  return r;
}

BDDnode*
BDDmgr::One2DC(BDDnode* f, IndexArray* ia, HashBase* ht) {
  if (ia==NULL) ia=defaultIA;
  boolean new_ht = (ht == NULL);
  if (new_ht)
    ht = new HasherTree;

  OrderIndexes( ia);
  BDDnode* res = One2DCR( f, ia, ht);

  if (new_ht)
    delete ht;
  return res;
}

BDDnode*
BDDmgr::One2DCR(BDDnode* f, IndexArray* ia, HashBase* ht) {
  if( f==One() || f==Zero() ) return f;


  integer level = GetLevel(f);
  if( level < ia->LowestLevel() )
    return f;

  BDDnode* r;
  if( f->NumReferences() > 1 && ht->Get((Object**)&r, f) )
    return r;

  BDDnode *f1 = One2DCR(f->Then(), ia, ht );
  BDDnode *f0 = One2DCR(f->Else(), ia, ht );

  if( !(ia->Find(level)) )
    return ((f0==f1) ? f0 : MakeNode(f->GetIndex(), f1, f0));


  // it is a variable we care about, so do it:
  if( f0==Zero() ) return f1; // make sure we can use makenode below
  if( f1==Zero() ) return MakeNode(f->GetIndex(), f1, f0);
  
  r = or(f1, MakeNode(f->GetIndex(), Zero(), f0));
  if( f->NumReferences() > 1)
    ht->Put( r, f);
  return r;
}

BDDnode*
BDDmgr::Zero2DC(BDDnode* f, IndexArray* ia, HashBase* ht) {
  if (ia==NULL) ia=defaultIA;
  boolean new_ht = (ht == NULL);
  if (new_ht)
    ht = new HasherTree;

  OrderIndexes( ia);
  BDDnode* res = Zero2DCR( f, ia, ht);

  if (new_ht)
    delete ht;
  return res;
}

BDDnode*
BDDmgr::Zero2DCR(BDDnode* f, IndexArray* ia, HashBase* ht) {
  if( f==One() || f==Zero() ) return f;


  integer level = GetLevel(f);
  if( level < ia->LowestLevel() )
    return f;

  BDDnode* r;
  if( f->NumReferences() > 1 && ht->Get((Object**)&r, f) )
    return r;

  BDDnode *f1 = Zero2DCR(f->Then(), ia, ht );
  BDDnode *f0 = Zero2DCR(f->Else(), ia, ht );

  if( !(ia->Find(level)) )
    return ((f0==f1) ? f0 : MakeNode(f->GetIndex(), f1, f0));


  // it is a variable we care about, so do it:
  if( f1==Zero() ) return f0; // make sure we can use makenode below
  if( f0==Zero() ) return MakeNode(f->GetIndex(), f1, f0);
  
  r = or(f0, MakeNode(f->GetIndex(), f1, Zero()));
  if( f->NumReferences() > 1)
    ht->Put( r, f);
  return r;
}

////////////////////////////////////////////////////////////////////////////
//
// Returns a single cube from the BDD function
// Set var_pref to FALSE if you want the cube to contain many Zeroes
//

BDDnode*
BDDmgr::SampleCube( BDDnode* f, boolean var_pref) {
  
  if ( f == One())         return f;
  if ( f == Zero())        return NULL;
  
  if ( var_pref) {
    if ( f->Then() == One()) 
      return MakeNode( f->GetIndex(), One(), Zero());  
    else if ( f->Then() == Zero()) 
      return MakeNode( f->GetIndex(), Zero(), SampleCube( f->Else(), var_pref));
    else
      return MakeNode( f->GetIndex(), SampleCube( f->Then(), var_pref), Zero());
  } else {
    if ( f->Else() == One())  
      return MakeNode( f->GetIndex(), Zero(), One());  
    else if ( f->Else() == Zero()) 
      return MakeNode( f->GetIndex(), SampleCube( f->Then(), var_pref), Zero());
    else
      return MakeNode( f->GetIndex(), Zero(), SampleCube( f->Else(), var_pref));
  }
}

//////////////////////////////////////////////////////////////////
//
//
//
//

BDDnode*
BDDmgr::Tuple( integer n, IndexArray* vars) {
  if (vars==NULL) vars=defaultIA;
  // initial checks
  integer max_width = vars->Count()-(--n);
  if ( max_width <= 0 ) return Zero();

  OrderIndexes(vars);

  BDDnode** temp = new BDDnode*[max_width];
  temp[0] = One();
  int i, j;
  for ( i=1; i< max_width; i++)
    temp[i] = ite( Input( vars->Index(vars->Count()-i)), Zero(), temp[i-1]);

  for (i=0; i<=n; i++) {
    temp[0] = MakeNode( vars->Index(vars->Count()-i-1), temp[0], Zero());
    for( j=1; j < max_width; j++)
      temp[j] = MakeNode( vars->Index(vars->Count()-i-j-1), temp[j], temp[j-1]);
  }
  BDDnode* result = temp[max_width-1];
  delete temp;

  return result;
}

BDDnode*
BDDmgr::AtLeastN( integer n, IndexArray* vars) {
  if (vars==NULL) vars=defaultIA;
  integer max_width = vars->Count() - (--n);
  if ( max_width <= 0 ) return Zero();

  OrderIndexes(vars);

  BDDnode** temp = new BDDnode*[max_width];
  int i,j;
  for ( i=0; i< max_width; i++)
    temp[i] = One();

  for (i=0; i<=n; i++) {
    temp[0] = MakeNode( vars->Index(vars->Count()-i-1), temp[0], Zero());
    for( j=1; j < max_width; j++)
      temp[j] = MakeNode( vars->Index(vars->Count()-i-j-1), temp[j], temp[j-1]);
  }
  BDDnode* result = temp[max_width-1];
  delete temp;

  return result;
}

/////////////////////////////////////////////////////////////////////////////////
//
// up to n variables true of vars->Count(), max_width (=mw) = vars->Count()-n
//
//                     /
//		  temp[0]
//		     / \ /
//		    N temp[1]
//		   / \ / \ /       -
//                   ...
//		  N   N temp[mw]  /
//		 / \ / \ / \    ...
//    -		0   N   N   1   /
//     \	   / \ / \    n+1
//     mw         0   N   1   /
//	 \	     / \     /
//	  -	    0   1   -

BDDnode*
BDDmgr::UpToN( integer n, IndexArray* vars) {
  if (vars==NULL) vars=defaultIA;
  int max_width = vars->Count()-n;
  if ( max_width <= 0) return One();

  OrderIndexes(vars);

  BDDnode** temp = new BDDnode*[max_width+1];
  int i,j;
  for ( i=0; i< max_width; i++)
    temp[i] = Zero();
  temp[max_width] = One();

  for ( i=n; i>=0; i--)
    for( j=max_width-1 ; j >= 0; j--)
      temp[j] = MakeNode( vars->Index(i+j), temp[j], temp[j+1]);

  BDDnode* result = temp[0];
  delete temp;

  return result;
}

