#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"
#include "strobj.h"
#include "hash.h"
#include "list.h"
#include "blifdata.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                        Andrew Seawright
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are not implemented in this version.
// This was done to trade-off efficiency vs. ease of implemetation and ease
// of modification.
//
//	Revised (7/19/93) F. Brewer
//	* changed the inconsistent node creation procedure to prevent
//	some bugs leading to unknown numbers of inputs by the manager.
//
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Function to read in an equation in algebraic form and produce a BDD
// Equation example: ( a + b ) * ( c' ^ d ) '
// The spaces in the above equation are necessary.
// If the variable is complemented then the prime must not be preceded by a space

BDDnode *BDDmgr::ReadEqn( HashTree *htname, HashTree *htindex, char *eqn )
{
  int i;
  static int ind;
  char *tok;
  StringObj *item;
  IntObj *index;
  BDDnode *op1, *op2;         // operands 1 and 2
  List *rpn, *res;

  rpn = new List;
  res = new List;
  //  ind = 0;

  tok = strtok(eqn," \t;");  // get the first token
  // convert algebraic notation into reverse polish notation
  ReadEqnR( tok, rpn, htname, htindex, ind );
  rpn->Reverse();

    List_ForEach_Type(rpn,StringObj,o){
      cout << o->String() << " ";
    }List_End_ForEach;
    cout << endl;

  // Time to do the actual bdd operations
  ind = rpn->Count();
  for( i=1; i <= ind; i++ ){
    item = (StringObj *)rpn->Pop();
    if( strstr( item->String(),"OPERATION_" )==NULL ){ // must be variable
      index = (IntObj *)htindex->Get(item->String());
      op1 = Input( index->val );
      res->Add( op1 );
    }else{  // must be operation
      if( strcmp( item->String(),"OPERATION_INVERT" )==0 ){  // unary operation
	op1 = (BDDnode *)res->Pop();
	res->Add( not(op1) );
      }else{  // binary operation
	op2 = (BDDnode *) res->Pop();
	op1 = (BDDnode *) res->Pop();
	if( strcmp(item->String(),"OPERATION_AND")==0 )
	  op1 = and( op1, op2 );
	else if( strcmp(item->String(),"OPERATION_OR")==0 )
	  op1 = or( op1, op2 );
	else if( strcmp(item->String(),"OPERATION_XOR")==0 )
	  op1 = xor( op1, op2 );
	res->Add( op1 );
      }
    }// endif on variable or operation
  }// endfor

  ind--;
  return( (BDDnode *) res->Pop() );
}

// Recursive function for ReadEqn so we can handle parentheses
List *BDDmgr::ReadEqnR( char *tok,
			    List *rpn,
			    HashTree *htname,
			    HashTree *htindex,
			    int& ind )
{
  StringObj *var;
  StringObj *lastop, *thisop; // last 2 operations
  IntObj *index;
  int invflag;
  char *varkey;

  invflag = 0;
  lastop = thisop = NULL;

  do{  // Loop to convert algebraic notation to rpn
    //    cout << "tok: " << tok << "  length: " << strlen(tok) << endl;
    if( strcmp(tok,"+")==0 || strcmp(tok,"^")==0 || strcmp(tok,"*")==0){
      // must be binary operator
      if( lastop==NULL ){
	if( strcmp(tok,"+")==0 )
	  lastop = new StringObj("OPERATION_OR");
	else if( strcmp(tok,"^")==0 )
	  lastop = new StringObj("OPERATION_XOR");
	else if( strcmp(tok,"*")==0 )
	  lastop = new StringObj("OPERATION_AND");
      }else{
	if( strcmp(tok,"+")==0 )
	  thisop = new StringObj("OPERATION_OR");
	else if( strcmp(tok,"^")==0 )
	  thisop = new StringObj("OPERATION_XOR");
	else if( strcmp(tok,"*")==0 )
	  thisop = new StringObj("OPERATION_AND");
      }
    }else{  // must be variable or paren
      if( strcmp(tok,"(")!=0 ){      // not a left paren
	if( strcmp(tok,")")==0 ){    // found a right parenthesis
	  if( lastop!=NULL )
	    rpn->Add(lastop);
	  return(rpn);
	}
	if( strstr(tok,"'")!=NULL ){ // complement operation
	  if( strlen(tok)>1 ){  // complement of variable
	    // assume the prime character is the last character
	    tok[strlen(tok)-1] = '\0';
	    if( !(htname->Get((Object **)&var,tok)) ){  // Not in hash tree
	      var = new StringObj(tok);
	      varkey = new char[strlen(tok)];
	      strcpy(varkey,tok);
	      htname->Put(var,varkey);
	      index = new IntObj(ind++);
	      htindex->Put(index,varkey);
	    }else{
	      index = (IntObj *) htindex->Get(tok);  // for debugging purposes
	    }
	    rpn->Add(var);
	  }
	  invflag++;
	}else{  // variable is not negated
	  if( !(htname->Get((Object **)&var,tok)) ){  // Not in hash tree
	    var = new StringObj(tok);
	    varkey = new char[strlen(tok)];
	    strcpy(varkey,tok);
	    htname->Put(var,varkey);
	    index = new IntObj(ind++);
	    htindex->Put(index,varkey);
	  }else{
	    index = (IntObj *) htindex->Get(tok);  // for debugging purposes
	  }
	  rpn->Add(var);
	}
	cout << "ind = " << ind << endl;
	cout << "variable: " << tok << "  index: " << index->val << endl;
	if( invflag ){
	  var = new StringObj("OPERATION_INVERT");
	  rpn->Add(var);
	  invflag--;
	}
	if( lastop!=NULL ){
	  if( strcmp( lastop->String(),"OPERATION_OR" )!=0 ){
	    rpn->Add(lastop);
	    lastop = thisop;
	    thisop = NULL;
	  }
	  if( thisop!=NULL ){
	    if( strcmp( thisop->String(),"OPERATION_OR" )==0 ){
	      rpn->Add(lastop);
	      lastop = thisop;
	    }else{
	      rpn->Add(thisop);
	    }
	    thisop = NULL;
	  }
	}
      }else{  // found a left parenthesis
	if( (tok = strtok( NULL, " \t;" ))==NULL )
	  break;
	ReadEqnR( tok, rpn, htname, htindex, ind );
	if( lastop!=NULL ){
	  rpn->Add(lastop);
	  lastop = thisop;
	}
      }

    }// endif on variable or operation 
  }while((tok=strtok(NULL," \t;")) != NULL);
  if( lastop!=NULL )
    rpn->Add(lastop);

  return(rpn);
}


/////////////////////////////////////////////////////////////////////////////
//
// Read a pla file into bdd. Generate an on set, an off set, and a dc set
//
//

void
BDDmgr::ReadPLA( List *onset, List *offset, List *dcset, istream &in ) {
  
  char line[400];
  char *token;
  BDDnode **on=NULL;
  BDDnode **off=NULL;
  BDDnode **dc=NULL;
  int numi=0, numo=0, nump=0, numP=0;
  int dcspec=0,onspec=1,offspec=0;
  char *incube;
  int linenum=0, i;

  while( in.getline( line, 400 )  ){
    linenum++;
    if( line[0] == '#' ) continue; // comment
    token = strtok( line, " \t" );
    if( token == NULL ) continue; // blank line
    if( token[0] == '.' ) {
      if( strcmp( token, ".i" ) == 0 ) {
	token = strtok( NULL, " \t" );
        numi=atoi(token);
        if( numi==0 ) {
	  cerr << "Bad number of inputs in PLA file.\n";
	  return;
	  }
	incube = new char[numi+1];
	continue;
	}
      if( strcmp( token, ".o" ) == 0 ) {
	token = strtok( NULL, " \t" );
        numo=atoi(token);
        if( numo==0 ) {
	  cerr << "Bad number of outputs in PLA file.\n";
	  return;
	  }
	on = new BDDnode *[numo];
	off= new BDDnode *[numo];
	dc = new BDDnode *[numo];
	for( i=0 ; i<numo ; i++ )
	  on[i] = off[i] = dc[i] = Zero();
	continue;
	}
      if( strcmp( token, ".p" ) == 0 ) {
	token = strtok( NULL, " \t" );
        nump=atoi(token);
        if( nump==0 ) {
	  cerr << "Bad number of outputs in PLA file.\n";
	  return;
	  }
	continue;
	}
      if( strcmp( token, ".type" ) == 0 ) {
	token = strtok( NULL, " \t" );
	int l;
	onspec=0;
	for( l=strlen(token)-1 ; l>=0 ; l-- ) {
	  if( token[l] == 'f' ) onspec=1;
	  if( token[l] == 'r' ) offspec=1;
	  if( token[l] == 'd' ) dcspec=1;
	  }
	continue;
	}
      if( strcmp( token, ".kiss" )==0 ) {
	cerr << "Kiss style pla files are not readable with this function.\n";
	return;
	}
      if( strcmp( token, ".e" )==0 || strcmp( token, ".end" )==0 )
	break;
      cerr << "Ignoring specifier: " << token << " at line " << linenum << endl;
      continue;
      }
    // product term
    int p;
    // try to handle cubes with spaces in the middle...
    if( numi == 0 ) {
      cerr << "PLA file has no inputs, or no .i # specifier\n";
      return;
      }
    if( numo == 0 ) {
      cerr << "PLA file has no outputs, or no .o # specifier\n";
      return;
      }
    for( i=0,p=0 ; i< numi ; i++,p++ ) {
      if(token[p] == '\0') {
	token = strtok( NULL, " \t" );
	p=0;
	}
      if( token == NULL ) {
	cerr << "Warning, product terms span multiple lines\n";
	if( !(in.getline( line, 400 )) ) {
	  cerr << "Bad input while reading PLA file (last line)\n";
	  return;
	  }
	linenum++;
	token = strtok( line, " \t" );
	p=0;
	}
      incube[i]=token[p];
      }
    incube[i] = '\0';
    BDDnode *cube = ReadCube( incube );
    // try to handle cubes with spaces in the middle...
    for( i=0 ; i< numo ; i++,p++ ) {
      if(token[p] == '\0') {
	token = strtok( NULL, " \t" );
	p=0;
	}
      if( token == NULL ) {
	cerr << "Warning, product terms span multiple lines\n";
	if( !(in.getline( line, 400 )) ) {
	  cerr << "Bad input while reading PLA file (last line)\n";
	  return;
	  }
	token = strtok( line, " \t" );
	p=0;
	}
      if( onspec && (token[p]=='1' || token[p] == '4') )
	on[i] = or(on[i] , cube);
      else if( offspec&& token[p]=='0' )
	off[i]= or(off[i], cube);
      else if( dcspec && (token[p]=='-' || token[p] == '2') )
	dc[i] = or(dc[i] , cube);
      }
    numP++;
    
    
    }
  if( nump!=numP ) {
    cerr << "Warning: number of product terms specified: " << nump;
    cerr << ", encountered: " << numP << endl;
    }
  for( i=numo-1 ; i>=0 ; i-- ) {
    onset-> Add( on[i] );
    dcset-> Add( dc[i] );
    offset->Add( off[i]);
    }
  }


/////////////////////////////////////////////////////////////////////////////
//
// This function prints a list of cubes representing the function,
// where each cube is a unique path from the top node to the terminal 
// one node. In the output, the least significant variable (index 0) is on 
// the left.
//

void
BDDmgr::PrintCubes( BDDnode *f, char eol, boolean braces, ostream& out) {
  int i;
  if(braces)
    if ( eol == '\n') out << "{\n";
    else              out << "{";
  
  if (f==one)
    out << "Tautology" << eol;
  else if (f==zero)
    out << "False" << eol;
  else {
    string cube = new char[(NumInputs() + 1)];
    for (i=0; i<NumInputs(); ++i)
      cube[i] = '-';
    cube[i] = '\0';
    PrintCubesR(f, cube, out, eol);
    delete cube;
  }
  if(braces)
    if ( eol == '\n') out << "}\n";
    else              out << "}";
}


void
BDDmgr::PrintCubesR(BDDnode *f, string cube, ostream& out, char eol ) {

  if (f == one) {
    out << cube << eol;
    return;
  } else if (f == zero) {
    return;
  } else {

    cube[f->GetIndex()] = '1';
    PrintCubesR(f->Then(),  cube, out, eol);
    cube[f->GetIndex()] = '0';
    PrintCubesR(f->Else(), cube, out, eol);
    cube[f->GetIndex()] = '-';
  }
}


static void
BDDmgr_FindHighIndex( BDDnode *f, int &h ) {
  if( f->Else()==NULL ) return;

  if( f->GetIndex() > h ) h = f->GetIndex();
  BDDmgr_FindHighIndex( f->Then(), h );
  BDDmgr_FindHighIndex( f->Else(), h );
  }

// %c is the cube in lineformat, or set of cubes in the paraformat
// %C is the alternate in lineformat, but doesn't print "Tautology"
// add more format commands as desired
//
// the IndexArray must be ordered by the user, and holds the order
// the variables will appear (in reverse)

void
BDDmgr::PrintCubesF( BDDnode *f, ostream& out, char *lineformat,
		char *paraformat, IndexArray *ia ) {
  int i;
  // parse paraformat
  char *pformStart = strcpy( new char[strlen(paraformat)+1], paraformat );
  char *pformEnd = pformStart;
  char *lformStart = strcpy( new char[strlen(lineformat)+1], lineformat );
  char *lformEnd = lformStart;
  boolean taut=TRUE;

  while( *pformEnd != '%' ) pformEnd++;
  *(pformEnd++) = '\0'; *(pformEnd++) = '\0';

  while( *lformEnd != '%' ) lformEnd++;
  *(lformEnd++) = '\0';
  if( *lformEnd == 'C' )
    taut = FALSE;
  *(lformEnd++) = '\0';

  cout << pformStart;
  if (f==one) {
    out << lformStart;
    if( taut ) out << "Tautology";
    out << lformEnd;
    }
  else if (f==zero) {
    if( taut )
      out << lformStart << "False" << lformEnd;
    }
  else {
    int h=0;
    if( ia )
      h = ia->Count()-1;
    else
      BDDmgr_FindHighIndex( f, h );
    char *cube = new char[h + 2];
    for (i=0; i<=h ; ++i)
      cube[i] = '-';
    cube[i] = '\0';
    PrintCubesFR(f, cube, out, lformStart, lformEnd, ia);
    delete cube;
    }

  cout << pformEnd;

  delete pformStart;
  delete lformStart;
  }
  

void
BDDmgr::PrintCubesFR(BDDnode *f, string cube, ostream& out, char *start,
		char *end, IndexArray *ia ) {

  if (f == one) {
    out << start << cube << end;
    return;
  } else if (f == zero) {
    return;
  } else {

    if( ia ) {
      ASSERT( ia->Find( GetLevel(f) ), "PrintCubes, indexarray not complete" );
      cube[ ia->Position() ] = '1';
      }
    else cube[f->GetIndex()] = '1';

    PrintCubesFR(f->Then(),  cube, out, start, end, ia);

    if( ia ) {
      ASSERT( ia->Find( GetLevel(f) ), "PrintCubes, indexarray not complete" );
      cube[ ia->Position() ] = '0';
      }
    else cube[f->GetIndex()] = '0';

    PrintCubesFR(f->Else(), cube, out, start, end, ia);

    if( ia ) {
      ASSERT( ia->Find( GetLevel(f) ), "PrintCubes, indexarray not complete" );
      cube[ ia->Position() ] = '-';
      }
    else cube[f->GetIndex()] = '-';
  }
}

////////////////////////////////////////////////////////////////////////////
//
//      routine for dumpping the entire ROBDD data structure
//

void
BDDmgr::Dump(ostream& out) {

  cout << "\nBDD node dump\n";
  unique->Print(out);
  out << "terminal zero node = " << (void *) zero << "\n";
  out << "terminal one node = " << (void *) one << "\n";
  computed->Print(out);

}


/////////////////////////////////////////////////////////////////////////////
//
//	Write BDDnodes (for example to a file)
//

void
BDDmgr::WriteBinary(List *funcs, ostream &out) {

  // first write the number of levels, and the index of each level
  int index, level, outoforder=FALSE;

  for( index=0 ; index<num_inputs ; index++ )
    if( unique->LookupLevel(index) != index ) {
      outoforder=TRUE;
      break;
      }
  if( outoforder == FALSE ) {
    IntObj::WriteInt( out, 0 );
    }
  else {
    IntObj::WriteInt( out, num_inputs );
    for( index = 0 ; index<num_inputs ; index++ ) {
      level = unique->LookupLevel( index );
      IntObj::WriteInt( out, unique->LookupLevel( index ) );
      }
    }
  // second save all the inportant functions you want to keep:
  IntObj::WriteInt( out, funcs->Count() );

  List_ForEach_Object( funcs, f ) {
    out.write((unsigned char*)&f, sizeof(BDDnode*));
    }List_End_ForEach

  // mark end of important nodes
  out.write((unsigned char*)&zero, sizeof(BDDnode*));
  out.write((unsigned char*)&one,  sizeof(BDDnode*));

  List_ForEach_Type(funcs, BDDnode, node) {
    WriteBinaryR( node, out );
    }List_End_ForEach
  List_ForEach_Type(funcs, BDDnode, node) {
    UnMark( node );
    }List_End_ForEach
  
  out.write("\0\0\0\0\0\0\0\0", sizeof(BDDnode*));
  out << flush;
  }

void
BDDmgr::WriteBinaryR(BDDnode *f, ostream &out ) {
  if( f==Zero() || f==One() ) return;

  if( f->Marked() ) return;
  f->Mark();

  WriteBinaryR(f->Then(),out);
  WriteBinaryR(f->Else(),out);

  out.write((unsigned char*)&f, sizeof(BDDnode*));
  out.write((unsigned char*)&(f->left),  sizeof(BDDnode*));
  out.write((unsigned char*)&(f->right), sizeof(BDDnode*));
  IntObj::WriteShort( out, f->index );
  }


////////////////////////////////////////////////////////////////////////////
//
//	Read BDDnodes in Binary format.
//
//	WARNING: reading in a bdd will force the manager to have the
//	order that the bdd was saved in (which affects other
//	bdd's already inside the manager).
//	As a consequence, garbage collection must be done, and the cache
//	cleared, etc.
//

List*
BDDmgr::ReadBinary(istream &in) {


  // first, set the order to the order of the saved bdd
  integer levels, *level;
  levels = IntObj::ReadInt( in );
  // make sure we have created enough variables
  if( levels<NumInputs() )
    Input( levels );
  level = new integer[levels];

  int i,j;
  for( i=0 ; i<levels ; i++ )
    level[i] = IntObj::ReadInt( in );

  boolean ordering_done;
  boolean change = FALSE;
  // BubbleSort, because all we can do is swap elements j and j-1
  for( i=0 ; i<levels-1 ; i++ ) {
    ordering_done = TRUE;
    for( j=levels-1 ; j>i ; j-- )
      if( level[unique->LookupIndex(j)] < level[unique->LookupIndex(j-1)] )
	{
	VariableSwap( j );
	ordering_done = FALSE;
	if( change == FALSE ) {
	  GC();
	  change = TRUE;
	  }
	}
    if( ordering_done ) break;
    }
  
  if( change ) timestamp++;
  List funcs;

  IntObj address;
  int count;
  for( count = IntObj::ReadInt( in ) ; count>0 ; count-- ) {
    in.read((unsigned char*)&address.val, sizeof(BDDnode*));
    funcs.Add(new IntObj(address.val));
    }

  HasherTree done(FALSE,TRUE);

  in.read((unsigned char*)&address.val, sizeof(BDDnode*));
  done.Put( Zero(), new IntObj(address.val) );
  in.read((unsigned char*)&address.val, sizeof(BDDnode*));
  done.Put( One(), new IntObj(address.val) );


  BDDnode *T, *F;
  IntObj taddress,faddress;
  short index;
  while( !in.eof() )
    {
    in.read((unsigned char*)&address.val, sizeof(BDDnode*));
    if( address == 0 )
      break;
    in.read((unsigned char*)&taddress.val, sizeof(BDDnode*));
    in.read((unsigned char*)&faddress.val, sizeof(BDDnode*));
    index = IntObj::ReadShort( in );
    if( !(done.Get( (Object**)&T, &taddress )) ||
	!(done.Get( (Object**)&F, &faddress ))  ) {
      cerr << "Nodes out of order during ReadBinary command\n";
      }
    BDDnode *local = MakeNode( index, T, F );
    IntObj *I = new IntObj( address.val);
    if( !(done.Put( local, I )) ) {
      cerr << "BDDnode encountered twice during read: "<<I<< endl;
      cerr << local;
      delete I;
      }
    }

  // now put the corresponding nodes in the List..
  List *out = new List;
  List_ForEach_Type( funcs, IntObj, a ) {
    BDDnode *local;
    if( !(done.Get( (Object**)&local, a )) ) {
      cerr << "Error reading bdd from stream\n";
      }
    out->Add(local);
    }List_End_ForEach
  funcs.CleanOut();
  delete level;
  return out;
  }

//////////////////////////////////////////////////////////////////////
//
// BLIF format reader. Generates only the intermediate functions,
// until you ask the BLIFdata manager for functions of primary inputs
//

BLIFdata*
BDDmgr::ReadBLIF( istream &in, BLIFdata* bd ) {
  static char buf[2000];
  char *token;
  char *slashmarker;
  int tlen;
  boolean done=FALSE;

  if( bd==NULL) bd = new BLIFdata(this);

  in.getline(buf,2000);
  token = strtok( buf, " \t" );

  while( in && !done ) {
    // still need to check if buffer size exceeded


    if( token == NULL )
      // blank line
      token=token;

    else if( token[0] == '.' ) {
      if( strcmp(token, ".model")==0 ) {
        token = strtok( NULL, " \t" );
        bd->SetModel( token );
        }

      else if( strcmp(token, ".outputs")==0 ) {

        while( token = strtok( NULL, " \t" ) ) {
	  if( token[0] == '#' ) break;
	  while( (slashmarker=strchr(token, '\\'))!=NULL ) {
	    tlen = (int)slashmarker-(int)token;
	    sprintf(buf, "%s", token);
	    in.getline(buf+tlen,2000-tlen);
	    token = strtok(buf, " \t");
	    }
          bd->AddOutput( token );
          }
        }

      else if( strcmp(token, ".inputs")==0 ) {

        while( token = strtok( NULL, " \t" ) ) {
	  if( token[0] == '#' ) break;
	  while( (slashmarker=strchr(token, '\\'))!=NULL ) {
	    tlen = (int)slashmarker-(int)token;
	    sprintf(buf, "%s", token);
	    in.getline(buf+tlen,2000-tlen);
	    token = strtok(buf, " \t");
	    }
          bd->AddInput( token );
          }
        }

      else if( strcmp(token, ".latch")==0 ) {
        char *latchin, *latchout, *latchreset;
        latchin = strtok(NULL, " \t");
        latchout = strtok(NULL, " \t");
        latchreset = strtok(NULL, " \t");
        bd->AddLatch( latchin, latchout, latchreset );
        }

      else if( strcmp(token, ".names")==0 ) {
        token = ReadBLIFFunction( in, bd );
        continue;
        }
      else if( strcmp(token, ".end")==0 ) {
        done=TRUE;
        }
      else if( strcmp(token, ".start_kiss")==0 ) {
	do {
	  in.getline(buf,2000);
	  token = strtok( buf, " \t" );
	  }while( in && strcmp(token, ".end_kiss")!=0 );
	continue;
	}

      }
    else if( token[0] != '#' ) {
      cerr << "Warning, unrecognized identifier in BLIF file:\n";
      cerr << token << endl;
      }
    in.getline( buf, 2000 );
    token = strtok( buf, " \t" );
    }

  if( done==FALSE )
    cerr << "Warning, BLIF file missing .end statement\n";
  return bd;
  }

char*
BDDmgr::ReadBLIFFunction( istream &in, BLIFdata *bd ) {

  static char buf[2000], *token, *slashmarker;
  StringObj *output;

  int i=0, tlen;
  List inputs;

  while( token = strtok( NULL, " \t" ) ) {
    if( token[0] == '#' ) break;
    // turn "backslash-eol" combo into a single line in buf
    if( (slashmarker=strchr(token, '\\'))!=NULL ) {
      tlen = (int)slashmarker-(int)token;
      sprintf(buf, "%s", token);
      in.getline(buf+tlen,2000-tlen);
      token = strtok(buf, " \t");
      }
    inputs.Add( new StringObj( token ) );
    }
  output = (StringObj*)inputs.Pop();
  inputs.Reverse();
  ASSERT( inputs.Count() < 1000, "Way too many inputs to function in BLIF file!" );

  BDDnode *function = Zero();
  in.getline( buf,2000 );
  token = strtok( buf, " \t" );
  while( (slashmarker=strchr(token, '\\'))!=NULL ) {
    tlen = (int)slashmarker-(int)token;
    sprintf(buf, "%s", token);
    in.getline(buf+tlen,2000-tlen);
    token = strtok( buf, " \t" );
    }
  while( in && (!token || token[0] != '.') ) {
    
    // check if blank line
    if( token && (token[0] == '1' || token[0] == '0' || token[0] == '-') )
      {
      BDDnode * cube;
      if( inputs.Count()>0 ) {
        cube = ReadCube(token);
        token = strtok( NULL, " \t" );
        }
      else
        cube = One();
      if( token[0] == '1' )
        function = or(function, cube);
      else
        function = or_not(function, cube);
      }
    in.getline(buf,2000);
    token = strtok( buf, " \t" );

    }
  bd->AddFunctionLocalVars( &inputs, function, output->String() );
  inputs.CleanOut();
  delete output;

  return token;
  }

////////////////////////////////////////////////////////////////////////////
//
// Print info and stats about the current state of the ROBDD
//

void 
BDDmgr::Print(ostream& out) {

  out << "\nBDD data structure info\n";

  out << "number of inputs    = " << num_inputs << "\n";
  out << "number of nodes     = " << num_nodes << "\n";
  out << "number of ite calls = " << num_ite_calls << "\n";
  computed->Print(out);
}

////////////////////////////////////////////////////////////////////////////
//
// A routine for reading in a function in the form of a list of cubes, from
// a stream. 
//

BDDnode *
BDDmgr::Read(istream& in) {

  BDDnode *f, *g;
  string s;
  integer i, j, k, n;
  
  in >> n;

  f = g = zero; 	// avoid a warning message

  string cube = new char[n+1];

  k=0;
  do {

    s=cube;
    *s = '\0';
    in >> s;
    if (strlen(s) != n)
      FATAL("reading boolean function with different support");

    for (j=i=0; i<n; ++i) {

      if (j) {
	switch (cube[i]) {
	case '1':
	  g = and(g, Input(i));
	  break;
	case '0':
	  g = and(g, not(Input(i)));
	  break;
        case '-':
        case ' ':
        case '*':
          break;
        default:
          FATAL("syntax error while reading function from stream");        
	}
      } else {
	if (cube[i] == '1') {
	  g = Input(i);
	  ++j;
	} else if (cube[i] == '0') {
	  g = not(Input(i));
	  ++j;
	}
      }
    }
    
    if (j) 
      if (k) {
        f = or (f,g);
      } else {
	f = g;
	++k;
      }

  } while (cube[0] != '.');

  delete cube;
  return f;
}

/////////////////////////////////////////////////////////////
//
// Read Equation - Chuck Monahan
// Reads-in format of PrintCubes function
//

BDDnode *
BDDmgr::ReadEq(istream& in, integer base, char eol) {
  BDDnode* eq = Zero();
  char c;

  for ( in >> c; c != '{'; in >> c);
  for ( in >> c; c != '}'; in >> c) {
    in.putback( c);
    eq = or( eq, ReadCube( in, base, eol));
  }
  return eq;
}

/////////////////////////////////////////////////////////////
//
// Read Cube - Chuck Monahan
// Reads-in cube from a single line
//    User can specify when to stop parsing (eol) and
//    how much of an offset (if variable shift is desired)
//

BDDnode *
BDDmgr::ReadCube(istream &in, integer base, char eol) {
  char cube[512];
  in.getline( cube, 512, eol );
  ASSERT( in.gcount()<511, "ReadCube can handle strings of 512 symbols");
  return ReadCube( cube, base);
}

BDDnode*
BDDmgr::ReadCube(char* cube, integer base) {
  int      len   = strlen(cube);
  BDDnode* eq    = One();
  int      match = 0;

  for (integer lev = 0; TRUE; lev++) {
    int offset = unique->LookupIndex( lev) - base;
    if (offset >= 0 && offset < len) {
      switch (cube[offset]) {
      case '0':	
	eq = MakeNode( offset+base, Zero(), eq);
	break;
      case '1':
	eq = MakeNode( offset+base, eq, Zero());
	break;
	
	// don't care
      case '2':
      case '-':
      case '*':
      case 'x':
      case 'X':
	break;
	
	// white spaces is ignored
      case ' ':
      case '\n':
      case '\t':
	base--;
	break;
	
	// don't know what it is
      default:
	FATAL("syntax error while reading function from stream");        
      }
      if (++match == len) break;
    }
  }
  return eq;
}

BDDnode *
BDDmgr::ReadCube(char* in, IndexArray* ia, char eol) {
  ASSERT( ia->Count() >= strlen(in), "Attempted to read cube over insufficient indexes");
  BDDnode* eq = One();
  OrderIndexes(ia);

  ia->Find( ia->LowestLevel());
  for ( integer i = 0; in[i] != eol && in[i] != '\0'; i++) {
    switch (in[i]) {
    case '0':	
      eq = MakeNode( ia->CurrIndex(), Zero(), eq);
      break;
    case '1':
      eq = MakeNode( ia->CurrIndex(), eq, Zero());
      break;

    // don't care
    case '2':
    case '-':
    case '*':
    case 'x':
    case 'X':
      break;

    // white spaces is ignored
    case ' ':
    case '\n':
    case '\t':
      ia->Find( ia->NextLevel());
      break;

    // don't know what it is
    default:
      FATAL("syntax error while reading function from stream");        
    }
    if (!ia->AtHighestLevel()) {         // adding this to for loop is regretable
      ia->Find( ia->PrevLevel());
    }
  }
  return eq;
}


////////////////////////////////////////////////////////////////////////////
//
// BDDnode member functions

void
BDDnode::Print( ostream& out) {
  out << "BDDn " << (void*) this
      << "\t	ref=" << NumReferences() << "\tv=" << GetIndex();
  out << " Then=" << (void*)left;
  out << " Else=" << (void*)right << endl;
}

void
BDDnode::PrintWithManager(BDDmgr *bdd, HasherTree* ht, ostream& out) {

  out << "BDDn " << (void *) this
      << "\t     r=" << NumReferences() << "\t  v=" << GetIndex() << "\n";
  out << "T=";
  if( bdd->Zero() == left )		out << "ZERO     ";

  else if( bdd->One() == left )		out << "ONE      ";

  else					out << (void *) left;


  out << " F=";
  if( right == bdd->One() )		out << "ONE     ";

  else if( right == bdd->Zero() )	out << "ZERO    ";

  else					out << (void *) right;

  Object* o;
  if( ht->Get(&o, Then()) )
    out << "\t     t=" << Then()->GetIndex();
  else
    out << "\t     T=" << Then()->GetIndex();
  

  if( ht->Get(&o, Else()) )
    out << "\tf=" <<  Else()->GetIndex() << endl;
  else
    out << "\tF=" <<  Else()->GetIndex() << endl;
}

