#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                        Andrew Seawright
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are not implemented in this version.
// This was done to trade-off efficiency vs. ease of implemetation and ease
// of modification.
//
//	Revised (7/19/93) F. Brewer
//	* changed the inconsistent node creation procedure to prevent
//	some bugs leading to unknown numbers of inputs by the manager.
//
////////////////////////////////////////////////////////////////////////////

void
BDDmgr::Keep( List *l ) {
  List_ForEach_Type( l, BDDnode, f ) {
    Keep( f );
    }List_End_ForEach
  }

void
BDDmgr::UnKeep( List *l ) {
  List_ForEach_Type( l, BDDnode, f ) {
    UnKeep( f );
    }List_End_ForEach
  }

BDDnode*
BDDmgr::Keep( BDDnode *f ) {
#ifdef DEBUG
  IntObj *num;

  if( keptNodes->Get((Object**)&num, f) )
    (num->val)++;
  else
    keptNodes->Put( new IntObj(1), f );

  num_keeps++;
#endif
  f->Reference();
  return f;
  }

void
BDDmgr::UnKeep( BDDnode *f ) {
#ifdef DEBUG
  IntObj *num;

  if( keptNodes->Get((Object**)&num, f) ) {
    if( --(num->val) == 0 )
      keptNodes->Remove( f );
    }
  else
    if( f!= Zero() && f != One() )
      cerr << "Warning: UnKeeping a node which is not currently Kept!\n";

  num_keeps--;
#endif
  f->UnReference();
  }



////////////////////////////////////////////////////////////////////////////
//
// This routine translates the BDDnode f to an equivalent function with
// a different variable ordering

BDDnode *
BDDmgr::Translate( BDDnode* f, HasherTree* variable_order, HashBase* ht )
  {
  boolean new_ht;
  if( variable_order->Count() == 0 ) return f;

  if( (new_ht = ht==NULL) )
    ht = new HasherTree;
  
  BDDnode* ret = TranslateR( f, variable_order, ht );

  if( new_ht )
    delete ht;
  
  return ret;
  }

BDDnode*
BDDmgr::TranslateR( BDDnode* from, HasherTree* new_index, HashBase* translated )
  {
  if( from == one || from == zero )
    return from;

  Object* o;
  if( translated->Get(&o, from) )
    return (BDDnode*) o;

  BDDnode *to, *g, *h;
  g = TranslateR( from->Then(), new_index, translated );
  h = TranslateR( from->Else(), new_index, translated );

  BDDnode* from_var = Input( from->GetIndex() );

  // if the index is in the hashertree, then translate it
  if( new_index->Get( (Object**) &to, from_var) )
    to = ite( to, g, h);

  else			// otherwise leave it alone!
    to = ite( from_var, g, h);

  translated->Put( to, from );

  return to;
  }

///////////////////////////////////////////////////////////////////
//
// These routines add a variable at an arbitrary location in the bdd
// also increment time stamp?

BDDnode *
BDDmgr::CreateVarBefore( int index ) {

  ASSERT( index < num_inputs, "No such variable exists" );

  int level = unique->LookupLevel(index);
  unique->InsertLevel( level );
  defaultIA->Put(NumInputs());

  BDDnode *r, *s;

  r = new BDDnode(NumInputs(), one, zero);
  s = new BDDnode(NumInputs(), zero, one);

  num_inputs++;
  unique->Put(r);
  unique->Put(s);
  num_nodes += 2;
  return r;
  }

BDDnode *
BDDmgr::CreateVarAfter( int index ) {

  ASSERT( index < num_inputs, "No such variable exists" );

  int level = unique->LookupLevel(index) + 1;
  unique->InsertLevel( level );
  defaultIA->Put(NumInputs());

  BDDnode *r, *s;

  r = new BDDnode(NumInputs(), one, zero);
  s = new BDDnode(NumInputs(), zero, one);

  num_inputs++;
  unique->Put(r);
  unique->Put(s);
  num_nodes += 2;
  return r;
  }

////////////////////////////////////////////////////////////////////
//
// Dynamic variable reordering routines:
// 
// VariableSwap() is the base for all reordering, and just swaps
// the positions of two variables in the unique table. This
// reordering is transparent to the user--i.e. indexes remain
// constant, and all BDD pointers are still valid, even though
// they may point to different (equivalent) structure now.

void
BDDmgr::VariableSwap( int level )
  {
  ASSERT( level>0, "Cannot variable swap level 0 and level -1.");
  if( level>=num_inputs)
    Input(level);		// create extra inputs up to level

  int upper = unique->LookupIndex(level);
  int lower = unique->LookupIndex(level - 1);

  unique->SwapIndex( level );
  unique->SwapLevel( upper, lower );

  if( unique->hashTables[upper]->Count() == 0 ||
      unique->hashTables[lower]->Count() == 0    )
    return;

  List moveum;
  BDDHasherArray_ForEach_in_Index( unique, upper, f )
  {
    if( f->Then()->GetIndex() != lower &&
	f->Else()->GetIndex() != lower   ) //independent of lower index!
      continue;
    else
      {
      unique->Remove( f );
      moveum.Add(f);
      }
    }BDDHasherArray_End_ForEach

  BDDnode *f11, *f10, *f01, *f00, *f1, *f0;
  List_ForEach_Type_Remove( moveum, BDDnode, f ) {

    // do one if statement for each pair, and skip the shannon1(0) stuff
    // to increase speed here
    f1 = f->Then();
    f0 = f->Else();
    f11 = Shannon1(f1,lower);
    f10 = Shannon0(f1,lower);
    f01 = Shannon1(f0,lower);
    f00 = Shannon0(f0,lower);


    // dereferencing gives us the chance to remove nodes
    f1->UnReference();
    f0->UnReference();

    f->index = lower; // but same memory location, so same level!
    f->left  = ( f11 == f01 ) ? f11 : MakeNode(upper,f11,f01);
    f->right = ( f10 == f00 ) ? f00 : MakeNode(upper,f10,f00);

    f->Then()->Reference();
    f->Else()->Reference();

    unique->hashTables[lower]->Add( f );
    }List_End_ForEach
  
  BDDHasherArray_ForEach_in_Index( unique, lower, g )
    {
    if( g->NumReferences() == 0 )
      {
      g->Then()->UnReference();
      g->Else()->UnReference();

      unique->Remove( g);
      delete g;
      num_nodes--;
      }

//VSGSR( g );
    }BDDHasherArray_End_ForEach
  }

void
BDDmgr::VSGSR( BDDnode *g ) {
  BDDnode* g1 = g->Then(), *g0 = g->Else();
  g0->UnReference();
  g1->UnReference();

  unique->Remove( g);
  delete g;
  num_nodes--;
  if( g0->NumReferences() == 0 )
    VSGSR( g0 );
  if( g1->NumReferences() == 0 )
    VSGSR( g1 );
  }

  

//////////////////////////////////////////////////////////////////////
//
// This calls VariableSwap() to move a group of nodes (size mod) at
// level.

void
BDDmgr::SwapMVar( int level, int mod)
  {
  for( int bing_cherries=0 ; bing_cherries<mod ; bing_cherries++ )
    for( int chaquita_bananas=level ; chaquita_bananas<level+mod ; chaquita_bananas++ )
      VariableSwap( chaquita_bananas-bing_cherries );
  }


///////////////////////////////////////////////////////////////////////
//
// Most general reordering algorithm. Works in
// 		O( nodes * levels ) time.
//

void
BDDmgr::Reorder_Sift(int start, int end, int mod)
{
  int max_nodes = 2 * num_nodes;
  int best_nodes = num_nodes;
  int best_level;
  HasherTree done(TRUE,FALSE);


  int index;
  int level;
  IntObj* Index = new IntObj;
  if (end==-1)
    end = NumInputs();
  for( index=0 ; index<num_inputs ; index++ )
    {
    best_level = level = unique->LookupLevel(index);
    Index->val = index;

    if( (level%mod != 0) || (done.Get( Index )) || level > end || level < start)
      continue;
    
    done.Put( Index );
    Index = new IntObj;
    int down=level;
    if( down > 0) do 			// move down in groups of size "mod"
      {
      SwapMVar( down, mod );

      if( num_nodes < best_nodes )
	{
	best_level = down;
	best_nodes = num_nodes;
	}
      down-=mod;
      }while ( down>=start+mod && num_nodes < max_nodes);
    
    int up=down+mod;
    do			// move up in groups of size "mod"
      {
      SwapMVar( up, mod );

      if( num_nodes <= best_nodes )	// = picks best level which is closest to top
	{				// (requires less work)
	best_level = up;
	best_nodes = num_nodes;
	}
      up+=mod;
      } while ( up<end && num_nodes < max_nodes );
    
    // move to optimal level

    down=up-mod;
    while( down>best_level )
      {
      SwapMVar( down, mod );
      down-=mod;
      }
    }
    delete Index;
  keptNodes->Rehash();
  timestamp++;
  }

///////////////////////////////////////////////////////////////////////
//
// Reordering algorithm where the starting order is generally good. Works in
// 			O( nodes ) time.
//

void
BDDmgr::Reorder_Window3( int start, int end, int mod )
  {
  if (end==-1)
    end = NumInputs();

  for( start+=mod; start<end-2*mod ; start+=mod )
    {
    // 1 2 3
    int best_nodes = num_nodes;
    int permutation = 0;

    SwapMVar( start+mod, mod );

    // 1 3 2
    if( num_nodes<best_nodes )
      {
      best_nodes=num_nodes;
      permutation = 1;
      }

    SwapMVar( start, mod );

    // 3 1 2
    if( num_nodes<best_nodes )
      {
      best_nodes=num_nodes;
      permutation = 2;
      }
    

    SwapMVar( start+mod, mod );

    // 3 2 1
    if( num_nodes<best_nodes )
      {
      best_nodes=num_nodes;
      permutation = 3;
      }

    SwapMVar( start, mod );

    // 2 3 1
    if( num_nodes<best_nodes )
      {
      best_nodes=num_nodes;
      permutation = 4;
      }

    SwapMVar( start+mod, mod );

    // 2 1 3
    if( num_nodes<best_nodes )
      {
      best_nodes=num_nodes;
      permutation = 5;
      }

    switch(permutation)
      {
      case 0: 			// 2 1 3 => 1 2 3
	SwapMVar( start, mod );

	break;
      case 1:			// 2 1 3 => 1 3 2
	SwapMVar( start, mod );
	SwapMVar( start+mod, mod );

	break;
      case 2:			// 2 1 3 => 3 1 2
	SwapMVar( start, mod );
	SwapMVar( start+mod, mod );
	SwapMVar( start, mod );

	break;
      case 3:			// 2 1 3 => 3 2 1
	SwapMVar( start+mod, mod );
	SwapMVar( start+mod, mod );

	break;
      case 4:			// 2 1 3 => 2 3 1
	SwapMVar( start+mod, mod );

	break;

      case 5:			// 2 1 3 => 2 1 3
      default:
	break;
      }
    }
  timestamp++;
  keptNodes->Rehash();
  }

////////////////////////////////////////////////////////////////////////////
//
// Swap2Var calls VariableSwap to swap the positions of two variables
//
// v1 and v2 are levels
//

void
BDDmgr::Swap2Var( int l1, int l2 ) {

  if( l1 > l2 )
    SWAP( int, l1, l2);
  int level = l1+1;
  
  while( level <= l2 )
    VariableSwap( level++ );
  
  level = l2-1;
  while( level > l1 )
    VariableSwap( level-- );
  }
  

////////////////////////////////////////////////////////////////////////////
//
//  routines to perform garbage collection - BDDnodes supporting only 
//  'unneeded' functions are deleted. 
//
//

void
BDDmgr::GC_Window3(int start, int end, int mod) {

  doGC(FALSE);
  Reorder_Window3(start, end, mod);
  computed->Clear();
  cofactor_hc->Clear();
  restrict_hc->Clear();
}

void
BDDmgr::GC_Sift(int start, int end, int mod) {

  doGC(FALSE);
  Reorder_Sift(start, end, mod);
  computed->Clear();
  cofactor_hc->Clear();
  restrict_hc->Clear();
}

void
BDDmgr::doGC(boolean clear_hcs) {

  BDDHasherArray_ForEach_Type(unique, key) {
    if (key->NumReferences() == 0 && ! key->Marked())
      RGC(key);
  } BDDHasherArray_End_ForEach

  if( clear_hcs ) {
//  computed->ConditionalClear();
    computed->Clear();
    cofactor_hc->Clear();
    restrict_hc->Clear();
    }

  BDDHasherArray_ForEach_Type(unique, node) {
    if (node->NumReferences() > 0) {
      node->UnMark();
    } else {
      ASSERT(node->Marked(), "deleting an unmarked node");
      unique->Remove(node);
      delete node;
      --num_nodes;
    }
  } BDDHasherArray_End_ForEach;

}

void
BDDmgr::RGC(BDDnode *i) {

  if (i == one || i == zero)
    return;

  i->Mark();

  BDDnode *t = (BDDnode*) i->Then();
  BDDnode *e = (BDDnode*) i->Else();

  if( t->NumReferences() == 0 ) cerr << "#";
  if( e->NumReferences() == 0 ) cerr << "#";
  t->UnReference();
  if (t->NumReferences() == 0)
    RGC(t);

  e->UnReference();
  if (e->NumReferences() == 0)
    RGC(e);
}

