#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                        Andrew Seawright
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are not implemented in this version.
// This was done to trade-off efficiency vs. ease of implemetation and ease
// of modification.
//
//	Revised (7/19/93) F. Brewer
//	* changed the inconsistent node creation procedure to prevent
//	some bugs leading to unknown numbers of inputs by the manager.
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// This routine collects the variable support of a function in a HasherTree.
//


void
BDDmgr::Support(BDDnode *f, HashExact *support) {

  HasherTree bdd_ht;
  SupportR(f, &bdd_ht, support);
}

void
BDDmgr::SupportR(BDDnode *f, HasherTree* bdd_ht, HashExact* support) {

  if (f == one || f == zero)
    return;

  if (bdd_ht->Put(*f,f)) {		// is this a new node?

    support->Put( Input(f->GetIndex()) );		// add it to the support

    SupportR( f->Then(), bdd_ht, support);
    SupportR( f->Else(), bdd_ht, support);
  }
}

////////////////////////////////////////////////////////////////////////////
//
// This routine returns the number of supporting variables contained in a
// BDD function.

integer
BDDmgr::NumSupport(BDDnode* f) {

  integer n;
  HasherTree *ht = new HasherTree;
  Support(f, ht);
  n = ht->Count();
  delete ht;
  return n;
}

////////////////////////////////////////////////////////////////////////////
//
// this routine returns the number of nodes in the ROBDD structure representing
// the function f by explicitly counting unique nodes. Complexity |f|.
//

integer 
BDDmgr::NumNodes(BDDnode* f, HashBase* ht) {

  integer n;
  boolean new_ht;
  if( (new_ht = !ht) )
    ht = new HasherTree;
  NumNodesR(f, ht);
  n = ht->Count();
  if( new_ht )
    delete ht;
  return n;
}

void
BDDmgr::NumNodesR(BDDnode *f, HashBase *ht) {

  if (f == one || f == zero) return;

  if (ht->Put(*f,f)) {
    NumNodesR(f->Then(), ht);
    NumNodesR(f->Else(), ht);
  }
}

////////////////////////////////////////////////////////////////////////////
//
// this routine returns the number of instances of each variable in a bdd
// (i.e. variable distribution)
//

integer*
BDDmgr::NodeDistribution(BDDnode* f, integer *instances ) {

  integer size = NumInputs(); 

  if( instances == NULL )
    instances = new integer[size];

  for(integer i=0; i<=size; i++) { instances[i]=0; }
  HasherTree *ht = new HasherTree;
  NodeDistributionR(f, ht,instances);
  delete ht;
  return	instances;
}

void
BDDmgr::NodeDistributionR(BDDnode *f, HasherTree *ht, integer* instances) {
  if (f == one || f == zero)
    return;
  if (ht->Put(*f,f)) {
    instances[f->GetIndex()]++;
    NodeDistributionR((BDDnode*)f->Then(), ht, instances);
    NodeDistributionR((BDDnode*)f->Else(), ht, instances);
  }
}

  
////////////////////////////////////////////////////////////////////////////
//
// This routine returns the number of paths to the terminal node 'one'
// in the ROBDD structure representing the function f.
// Routine has potentailly exponential complexity.

integer 
BDDmgr::NumCubes(BDDnode* f) {

  integer n;
  HasherTree *ht = new HasherTree;
  n = NumCubesR(f, ht);
  delete ht;
  return n;
}

integer
BDDmgr::NumCubesR(BDDnode *f, HasherTree *ht) {


  if (f == one)
    return 1;

  if (f == zero)
    return 0;   

  Object *o;
  if (ht->Get(&o, f))
    return (integer) o;

  else {

    integer n;
    n = NumCubesR(f->Then(), ht) +
        NumCubesR(f->Else(), ht);
    
    ht->Put((Object*)n, f);
    return n;
  }
}

////////////////////////////////////////////////////////////////////////////
//
// this routine returns the depth of the BDD function f
//

integer 
BDDmgr::DepthOfFunction(BDDnode* f) {

  integer d = 0, m = 0;
  HasherTree *ht = new HasherTree;
  DepthOfFunctionR(f, d, m, ht);
  delete ht;
  return m;
}

void
BDDmgr::DepthOfFunctionR( BDDnode *f, 
                              integer  d,
                              integer& m, 
                              HasherTree* ht) {

  if (f == one || f == zero)
    return;

  if ( ++d > m ) 
    m = d;

  Object *o;

  if( !(ht->Get(&o, f)) || ((integer)o < d) ) {

    ht->ForcePut((Object*)d, f);

    DepthOfFunctionR(f->Then(), d, m, ht);
    DepthOfFunctionR(f->Else(), d, m, ht);
  }
}

////////////////////////////////////////////////////////////////////////////
//
// NumMinterms computed the number of minterms
//
//
integer
BDDmgr::NumMinterms(BDDnode *f, integer N, HashBase* hc)
  {
  boolean new_hs;

  if( (new_hs = !hc) )
    hc = new HasherTree(TRUE);

  integer _bdd_points = 0;
  integer _bdd_height = 0;

  NumMinterms(f, hc, _bdd_points, _bdd_height);

  if( N > _bdd_height )
    _bdd_points = _bdd_points << (N - _bdd_height);

  if( new_hs )
    delete hc;
  return _bdd_points;
}


/////////////////////////////////////////////////////////////////////////////
//
// NumMinterms must calculate the level up from the bottom for each node,
// and the on set size for that node. The return IntObj2 has two values:
// val1 is the on-set-size of the node, and val2 is the level.
//

void
BDDmgr::NumMinterms(BDDnode *f, HashBase* hc, integer& _bdd_points, integer &_bdd_height)
  {
  if( f == zero )
    {
    _bdd_height = 0;
    _bdd_points = 0;
    return;
    }

  if( f == one )
    {
    _bdd_height = 0;
    _bdd_points = 1;
    return;
    }
  
  List *l;

  if( hc->Get((Object**)&l, f) )
    {
    _bdd_height = (integer) l->GetElement(0);
    _bdd_points = (integer) l->GetElement(1);
    return;
    }

  NumMinterms(f->Then(), hc, _bdd_points, _bdd_height);
  integer points = _bdd_points;
  integer height = _bdd_height;

  NumMinterms(f->Else(), hc, _bdd_points, _bdd_height);
  if( height > _bdd_height )
    _bdd_points = points + (_bdd_points << (height - _bdd_height));
  else
    _bdd_points = _bdd_points + (points << (_bdd_height - height));

  _bdd_height = MAX(_bdd_height, height) + 1;

  if( f->NumReferences() > 1)		// will anybody else ever look for this node?
    {
    l = new List;
    l->Add( (Object*)( _bdd_points ) );
    l->Add( (Object*)( _bdd_height ) );
    hc->Put( (Object*) l, f );
    }

  return;
  }

