#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <limits.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                    Andrew Seawright
//                                                      Forrest Brewer
//                                                      Chuck Monahan
//                                                      Ivan R.
//                                                      Andy Crews
//                                                      Tony Stornetta
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are implemented in this version, with the
// intent to minimize the amount of necessary code rewriting in existing
// software.
//
// The basic BDD package was originally designed at UCSB, and called HomeBrew.
// Inverted edges were implemented with the help of the C BDD package from CMU.
//
////////////////////////////////////////////////////////////////////////////

BDDnode*
BDDmanager::AndRestrict(BDDnode* f, BDDnode* g, BDDnode* c) { 
  if (c == one)
    return and( f, g);

  ++num_ite_calls;

    //
   //  check for terminal cases
  //
  
  if (f == zero || g == zero)
    return zero;
  if (f == one || f == g)
    return Restrict( g, c);
  if (g == one)
    return Restrict( f, c);

    //
   //  have we already computed this function ?
  //

  if ( (integer)f > (integer)g ) {  // order for computed
    SWAP( BDDnode*, f, g);
  }

  BDDnode* o;
  tempcomputed->Set(ANDR_OP,f,g,c);

  if (computed->Get(&o, *tempcomputed)) {
    return o;
  }

     //
    // recursively compute ite()
   //

  integer top;
  if( GetLevel(f) > GetLevel(g) )
    top = (GetLevel(f) > GetLevel(c))? f->GetIndex() : c->GetIndex();
  else
    top = (GetLevel(g) > GetLevel(c)) ? g->GetIndex() : c->GetIndex();


  BDDnode *r;
  BDDnode* c1 = Shannon1(c, top);
  BDDnode* c0 = Shannon0(c, top);

  if ( c0 == Zero()) {
    r = AndRestrict(Shannon1(f, top), Shannon1(g, top), c1);
  } else if ( c1 == Zero()) {
    r = AndRestrict(Shannon0(f, top), Shannon0(g, top), c0);
  } else if( GetLevel(c) > GetLevel(f) && GetLevel(c) > GetLevel(g)) {
    r = AndRestrict( f, g ,or( c0, c1 )); 
  } else {
    BDDnode *t, *e;
    t = AndRestrict(Shannon1(f, top), Shannon1(g, top), c1);
    e = AndRestrict(Shannon0(f, top), Shannon0(g, top), c0);
    r = (t == e)? t : MakeNode(top, t, e);
  }
  
  if (r == c)
    return One();

  tempcomputed->Set(ANDR_OP,f,g,c);
  computed->Replace(r, *tempcomputed); // store computed ite() in cache

  return r;
}


BDDnode*
BDDmanager::AndSmoothRestrict(BDDnode* f, BDDnode* g, BDDnode* c, IndexArray* ia, HashBase* hc, HashBase* sm) { 
  if (c == one)
    return AndSmoothR( f, g, ia, hc, sm);

  integer v = (GetLevel(f) > GetLevel(g))? GetLevel(f): GetLevel(g);
  if ( v < ia->LowestLevel())
    return AndRestrict( f, g, c);

  ++num_ite_calls;

    //
   //  check for terminal cases
  //
  
  if (f == zero || g == zero) {
    BDDnode* r = zero;
    return r;
  }
  if (f == one || f == g)
    return Restrict( SmoothR( g, ia, sm, TRUE), c);
  if (g == one)
    return Restrict( SmoothR( f, ia, sm, TRUE), c);

    //
   //  have we already computed this function ?
  //

  if ( (integer)f > (integer)g ) {  // order for computed
    SWAP( BDDnode*, f, g);
  }

  BDDnode*  o;
  BDDCacheKey* comp = NULL;
  if( f->NumReferences() > 1 && g->NumReferences() > 1 ) {
    comp = new BDDCacheKey(ANDR_OP,f,g,c);
    if ( hc->Get((Object**) &o, comp)) {
      delete comp;
      return o;
    }
  }

  BDDnode* r;
  tempcomputed->Set(ANDR_OP,f,g,c);
  if ( comp != NULL && computed->Get(&o, *tempcomputed)) {  // if "and" portion is precomputed
    r = SmoothR( o, ia, sm, TRUE);
  } else {
    tempcomputed->Set(ITE_OP,f,g,Zero());
    if (computed->Get(&o, *tempcomputed)) {
      r = Restrict( SmoothR( o, ia, sm, TRUE), c);
    } else {

      //
      // recursively compute ite()
      //
      
      integer top;
      if( GetLevel(f) > GetLevel(g) )
	top = (GetLevel(f) > GetLevel(c))? f->GetIndex() : c->GetIndex();
      else
	top = (GetLevel(g) > GetLevel(c))? g->GetIndex() : c->GetIndex();
      
      BDDnode* c1 = Shannon1(c, top);
      BDDnode* c0 = Shannon0(c, top);
      
      if ( c0 == Zero()) {
	r = AndSmoothRestrict(Shannon1(f, top), Shannon1(g, top), c1, ia, hc, sm);
      } else if ( c1 == Zero()) {
	r = AndSmoothRestrict(Shannon0(f, top), Shannon0(g, top), c0, ia, hc, sm);
      } else if( GetLevel(c) > GetLevel(f) && GetLevel(c) > GetLevel(g)) {
	r = AndSmoothRestrict( f, g ,or( c0, c1 ), ia, hc, sm); 
      } else {
	BDDnode* t = AndSmoothRestrict(Shannon1(f, top), Shannon1(g, top), c1, ia, hc, sm);
	BDDnode* e = AndSmoothRestrict(Shannon0(f, top), Shannon0(g, top), c0, ia, hc, sm);
	if ( t == e) {
	  r = t;
	} else {    // Is this a level that we want to smooth?
	  r = (ia->Find( v))? or( t, e) : MakeNode(top, t, e);
	}
      }
    }
  }
  
  if (r == c)        r = One();
  if (comp != NULL ) hc->ForcePut((Object*)r, comp);
  
  return r;
}


BDDnode *
BDDmanager::XRestrict(BDDnode *f, BDDnode *c, boolean greedy) {

  ASSERT(c != zero, "Restrict(): assertion that c!=0 failed");

  if (f == c)
    return One();
  if (c == one || f == one || f == zero)
                                { return f; }

  integer v = ( GetLevel(f) > GetLevel(c) ) ? f->GetIndex() : c->GetIndex();

  BDDnode* o, *c0, *c1;

  tempcomputed->Set(XRESTRICT_OP, f, c);
  if (restrict_hc->Get(&o, *tempcomputed)==0)
    {
    c0 = Shannon0(c,v);
    c1 = Shannon1(c,v);
    if (c0 == zero)
      o = XRestrict(Shannon1(f, v), c1, greedy);

    else if (c1 == zero)
      o = XRestrict(Shannon0(f, v), c0, greedy);

    else
      {
      BDDnode *f0 = Shannon0(f,v);
      BDDnode *f1 = Shannon1(f,v);

      ///////////////////////////////////////////////////////////////
      // if variable v is not in f, then the smallest function which
      // distinguishes f from all other cubes in c cannot contain
      // variable v either, so smooth this variable in c and continue

      BDDnode *fchild, *cchild = or( c0,c1 );

      if( GetLevel(c) > GetLevel(f) )
        o = XRestrict( f , cchild, greedy);

      ///////////////////////////////////////////////////////////////
      // Variable v exists in both f and c, and we need to determine
      // if it is possible to distinguish cube f from other cubes in
      // c if we ignore variable v.

      // if it is possible to ignore variable v, then lets do it.
      else if( greedy && and( f0, c ) == and( f, c) )
	o = XRestrict( f0 , cchild, greedy );

      else if( greedy && and( f1, c ) == and( f, c) )
	o = XRestrict( f1 , cchild, greedy );

      else if( !greedy && XRestrictIgnoreVar( (fchild=or(f0,f1)), f, c, cchild ) )
        o = XRestrict( fchild, cchild, greedy );

      else {

	f1 = XRestrict(f1, c1, greedy);
	f0 = XRestrict(f0, c0, greedy);
	o = (f0 == f1 ) ? f0 : MakeNode(v, f1, f0);
	}
      if( f->NumReferences() > 1 && c->NumReferences() > 1)
	{
	tempcomputed->Set( XRESTRICT_OP, f, c );
	restrict_hc->Replace(o,*tempcomputed);
	}
      }
  }
  return o;
}


boolean
BDDmanager::XRestrictIgnoreVar( BDDnode *fchild, BDDnode *f, BDDnode *c, BDDnode *cchild ) {
  
  integer c1=0, c2=0;

  f = XRestrictIgnoreVarR( f, c, c1);
  if( f != and(fchild, c) )
    return FALSE;		// can't ignore variable!!
  XRestrictIgnoreVarR( fchild, cchild, c2);
  return (c2 <= c1);		// is it a good idea to ignore the variable?
  }


BDDnode*
BDDmanager::XRestrictIgnoreVarR( BDDnode *f, BDDnode *g, integer &count) {
		// same as "and", except for count
  ++num_ite_calls;

  if (f == one) return g;
  if (g == one) return f;
  if (f == g) return f;
  if (f == zero || g == zero) { if( f==zero) count++; return zero;}

//  BDDnode* o;
//  tempcomputed->Set(ITE_OP,f,g,Zero());
//  if (computed->Get(&o, *tempcomputed)) return o;

  integer top = (GetLevel(f) > GetLevel(g))? f->GetIndex() : g->GetIndex();
  BDDnode *t, *e, *r;
  t = XRestrictIgnoreVarR(Shannon1(f, top), Shannon1(g, top),count);
  e = XRestrictIgnoreVarR(Shannon0(f, top), Shannon0(g, top),count);
  r =  (t == e)  ? t : MakeNode(top, t, e);

  if ( (integer)f > (integer)g )
    tempcomputed->Set(ITE_OP,g,f,Zero());
  else
    tempcomputed->Set(ITE_OP,f,g,Zero());
  computed->Replace(r, *tempcomputed); // store computed ite() in cache

  return r;
}

/*
boolean
BDDmanager::XRestrictRedCheck(  BDDnode *f, BDDnode *c,
				BDDnode *f1, BDDnode *c1,
				integer &v1, integer &v2 ) {

  if( (f==f1 && c==c1) || (f==c1 && c=f1) ) return TRUE;
  BDDnode *r1 = NULL, *r2 = NULL;
  if (f == one) r1 = c;
  if (c == one) r1 = f;
  if (f == c) r1 = f;
  if (f == zero || c == zero) r1 = zero;

  if (f1 == one) r2 = c1;
  if (c1 == one) r2 = f1;
  if (f1 == c1) r2 = f1;
  if (f1 == zero || c1 == zero) r2 = zero;
  if( r1 && r2 ) return (r1 == r2 );

//  BDDnode* o;
//  tempcomputed->Set(ITE_OP,f,g,Zero());
//  if (computed->Get(&o, *tempcomputed)) return o;

  integer top = (GetLevel(f) > GetLevel(c))? f->GetIndex() : c->GetIndex();
  integer top2 = (GetLevel(f1) > GetLevel(c1))? f1->GetIndex() : c1->GetIndex();
  top = MAX( top, top2 );
  BDDnode *t, *e;
  t = XRestrictRedCheck(Shannon1(f, top), Shannon1(c, top),count);
  e = XRestrictRedCheck(Shannon0(f, top), Shannon0(c, top),count);
  r1 =  (t == e)  ? t : MakeNode(top, t, e);

  t = XRestrictRedCheck(Shannon1(f1, top), Shannon1(c1, top),count);
  e = XRestrictRedCheck(Shannon0(f1, top), Shannon0(c1, top),count);
  r2 =  (t == e)  ? t : MakeNode(top, t, e);

//  if ( (integer)f > (integer)g )
//    tempcomputed->Set(ITE_OP,g,f,Zero());
//  else
//    tempcomputed->Set(ITE_OP,f,g,Zero());
//  computed->Replace(r, *tempcomputed); // store computed ite() in cache

  return (r1 == r2);
}
*/


////////////////////////////////////////////////////////////////////////////
//
//  This is like a superset computation (is f a superset of s?) for
//  variables in "ia", and an "AND" for variables not in "ia"
//

BDDnode*
BDDmanager::SupersetAnd( BDDnode *f, BDDnode *s, IndexArray *ia ) {
  HasherTree ht(FALSE,TRUE);

  OrderIndexes(ia);
  return SupersetAndR( f, s, ia, &ht );
  }


BDDnode *
BDDmanager::SupersetAndR( BDDnode *f, BDDnode *s, IndexArray *ia, HasherTree *ht ) {
  
  int fl = GetLevel(f), sl = GetLevel(s);
  int top = MAX( fl, sl );

  if( ia->Find(top) || top == -1 ) { 		// do superset
    if( f==s || f == One() || s == Zero() )
      return One();
    if( f==Zero() || s == One() )
      return Zero();
    }
  else {				// do and
    if( f == One() || f == s ) return s;
    if( s == One() ) return f;
    if( f == Zero() || s == Zero() ) return Zero();
    }

  if ( f->NumReferences() > 1 && s->NumReferences() > 1) {
    tempcomputed->Set(ITE_OP,f,s,Zero());
    BDDnode* o;
    if ( ht->Get( (Object**)&o, tempcomputed))
      return o;
    }
  

  BDDnode *l, *r;
  l = SupersetAndR( Shannon1(f, top), Shannon1(s, top), ia, ht );
  r = SupersetAndR( Shannon0(f, top), Shannon0(s, top), ia, ht );

  BDDnode *result;
  if( ia->Find(top) )			// do superset
    result = ( l==One() && r==One() ) ? One() : Zero();
  else					// do and
    result = ( l==r ) ? l : MakeNode(top, l, r);

  if ( f->NumReferences() > 1 && s->NumReferences() > 1) {
    tempcomputed->Set(SUPERAND_OP, f, s, Zero());
    ht->Put( result, new BDDCacheKey( *tempcomputed ) );
    }

  return result;
  }


////////////////////////////////////////////////////////////////////////////
//
// this routine evaluates a BDD function based on the state of the
// variable inputs contained in bv which is a BitVector Class (Ivan's)
// 

boolean
BDDmanager::Evaluate(BDDnode *f, BitVector *bv) {

  integer v;

  for (;;) {

    if (f == one)
      return TRUE;
    else if ( f == zero)
      return FALSE;
    else {
      v = f->GetIndex();
      ASSERT(v < bv->Size() && v>=0, "evaluation: variable out of range");
      f = ((*bv)[v]) ? f->Then():f->Else();
    }
  }
}


BDDnode*
BDDmanager::AndSmooth(BDDnode* f, BDDnode* g, IndexArray* ia, HashBase* hc) { 
  if ( ia->LowestLevel() > ((GetLevel(f) > GetLevel(g))? GetLevel(f): GetLevel(g)))
    return and( f, g);

  HasherTree* smooth = new HasherTree;
  BDDnode*    r      = AndSmoothR( f, g, ia, hc, smooth);

  delete smooth;
  return r;
}

BDDnode*
BDDmanager::AndSmoothR(BDDnode* f, BDDnode* g, IndexArray* ia, HashBase* hc, HashBase* sm) { 
  integer v = (GetLevel(f) > GetLevel(g))? GetLevel(f): GetLevel(g);
  if ( v < ia->LowestLevel())
    return and( f, g);

  ++num_ite_calls;

    //
   //  check for terminal cases
  //
  
  if (f == zero || g == zero)
    return zero;
  if (f == one || f == g)
    return SmoothR( g, ia, sm, TRUE);
  if (g == one)
    return SmoothR( f, ia, sm, TRUE);

    //
   //  have we already computed this function ?
  //

  if ( (integer)f > (integer)g ) {   // order for computed
    SWAP( BDDnode*, f, g);
  }

  BDDnode*  o;
  BDDCacheKey* comp = NULL;
  if( f->NumReferences() > 1 && g->NumReferences() > 1 ) {
    comp = new BDDCacheKey(ITE_OP,f,g,Zero());
    if ( hc->Get((Object**) &o, comp)) {
      delete comp;
      return o;
    }
  }

  BDDnode* r;
  if ( comp != NULL && computed->Get(&o, *comp)) {  // if "and" portion is precomputed
    r = SmoothR( o, ia, sm, TRUE);

  } else {                                 // else recursively compute AndSmooth()
    integer top = (GetLevel(f) > GetLevel(g))? f->GetIndex() : g->GetIndex();
    
    BDDnode* t = AndSmoothR(Shannon1(f, top), Shannon1(g, top), ia, hc, sm);
    BDDnode* e = AndSmoothR(Shannon0(f, top), Shannon0(g, top), ia, hc, sm);
    
    if ( t == e) {
      r = t;
    } else {    // Is this a level that we want to smooth?
      r = (ia->Find( v))? or( t, e) : MakeNode(top, t, e);
    }
  }

  if( comp != NULL ) hc->ForcePut((Object*)r, comp);

  return r;
}


////////////////////////////////////////////////////////////////////////////
//
// This routine returns the function g corrisponding to the cube encoding of
// the list of input functions by the integer w. least significant functions
// are first in the list. The routine only encodes a cube of the number of
// bits equal to the number of functions in the list even though the integer w 
// may contain more information. e.g. only the functions->Count() number of
// lsb bits of w are used to form the returned cube g
// Routine will work for codes up to 31 bits.
//

BDDnode *
BDDmanager::Code2Cube(List* functions, integer w) {

  BDDnode *g = one;
  List_ForEach_Type(*functions, BDDnode, f) {
    if ( 0x01 & w)
      g = and(g, f);
    else
      g = and(g, not(f));
    w >>= 1;
  } List_End_ForEach;
  return g;
}

///////////////////////////////////////////////////////////////////
//
// This routine takes a function and replaces don't care elements
// with zero terms.  It replaces only those elements whose indexes 
// are listred in the IndexArray

BDDnode*
BDDmanager::DC2Zero(BDDnode* f, integer index, HashBase* ht) {
  boolean new_ht = (ht == NULL);
  if (new_ht)
    ht = new HasherTree;

  integer  level = unique->LookupLevel(index);
  BDDnode* res   = DC2ZeroR( f, level, ht);

  if ( GetLevel(res) < level && res != Zero()) {
    res = MakeNode( index, Zero(), res);
  }

  if (new_ht)
    delete ht;
  return res;
}


BDDnode*
BDDmanager::DC2ZeroR(BDDnode* f, integer level, HashBase* ht) {
  integer lev = GetLevel(f);

  if( lev <= level )
    return f;

  Object* res = ht->Get( f);
  if ( res!=NULL)
    return (BDDnode*) res;

  BDDnode* tru = DC2ZeroR( f->Then(), level, ht);
  if ( lev > level && GetLevel(tru) < level && tru != Zero()) {
    integer index = unique->LookupIndex(level);
    tru = MakeNode( index, Zero(), tru);
  }
  
  BDDnode* fal = DC2ZeroR( f->Else(), level, ht);
  if ( lev > level && GetLevel(fal) < level && fal != Zero()) {
    integer index = unique->LookupIndex(level);
    fal = MakeNode( index, Zero(), fal);
  }
  
  // tru can not equal fal.  When in doubt, try to build an example
  BDDnode* r = MakeNode( f->GetIndex(), tru, fal);
  if( f->NumReferences() > 1)
    ht->Put( r, f);
  return r;
}

///////////////////////////////////////////////////////////////////
//
// This routine takes a function and replaces don't care elements
// with 'true' terms.  It replaces only those elements whose indexes 
// are listred in the IndexArray

BDDnode*
BDDmanager::DC2One(BDDnode* f, integer index, HashBase* ht) {
  boolean new_ht = (ht == NULL);
  if (new_ht)
    ht = new HasherTree;

  integer  level = unique->LookupLevel(index);
  BDDnode* res   = DC2OneR( f, level, ht);

  if ( GetLevel(res) < level && res != Zero()) {
    res = MakeNode( index, res, Zero());
  }

  if (new_ht)
    delete ht;
  return res;
}


BDDnode*
BDDmanager::DC2OneR(BDDnode* f, integer level, HashBase* ht) {
  integer lev = GetLevel(f);

  if( lev <= level )
    return f;

  Object* res = ht->Get( f);
  if ( res!=NULL)
    return (BDDnode*) res;

  BDDnode* tru = DC2OneR( f->Then(), level, ht);
  if ( GetLevel(tru) < level && level < lev && tru != Zero()) {
    integer index = unique->LookupIndex(level);
    tru = MakeNode( index, tru, Zero());
  }
  
  BDDnode* fal = DC2OneR( f->Else(), level, ht);
  if ( GetLevel(fal) < level && level < lev && fal != Zero()) {
    integer index = unique->LookupIndex(level);
    fal = MakeNode( index, fal, Zero());
  }

  // tru can not equal fal.  When in doubt, try to build an example
  BDDnode* r = MakeNode( f->GetIndex(), tru, fal);
  if( f->NumReferences() > 1)
    ht->Put( r, f);
  return r;
}

/////////////////////////////////////////////////////////////////////
//
// This function takes a minterm, and construct the set of all
// codes which are hamming distance "d" from the minterm.
//

BDDnode*
BDDmanager::BuildHamming(BDDnode *minterm, BDDnode *constraint, integer d )
  {
  // dont set minterm == Zero or One. The result is not really defined
  // (other than the fact that the program will crash)
  if( d == 0 )
    return and(minterm, constraint);
  if( constraint == Zero() )
    return Zero();
  
  integer top = MAX( minterm->GetIndex(), constraint->GetIndex() );
  BDDnode *g = Shannon1(minterm,    top), *h = Shannon0(minterm,    top);
  BDDnode *i = Shannon1(constraint, top), *j = Shannon0(constraint, top);

  if( g == One() || h == One() )	// then it is assumed that the other is Zero
    return MakeNode(minterm->GetIndex(), h , g);

  if( h == Zero() )
    {
    h = BuildHamming(g, j, d-1);
    g = BuildHamming(g, i, d);
    }
  else	// g is not zero
    {
    g = BuildHamming(h, i, d-1);
    h = BuildHamming(h, j, d);
    }

  return ( (g==h) ? g : MakeNode(minterm->GetIndex(), g, h) );
  }


//////////////////////////////////////////////////////////////////
//  MakeMin2N creates a BDD with minterms from 0 to N.
//  For example, if N = 3, and there are 4 indices in the IndexArray,
//  then MakeMin2N will create a BDD with minterms 0000,0001 & 0010.
//  WARNING! WARNING! This funciton is order dependent.  

BDDnode*
BDDmanager::MakeMin2N( int n, IndexArray* ia) {
  if (ia==NULL) ia=defaultIA;

  int place = 1;
  place = place<<(ia->Size());

  if (n>place)
  return Zero();  // Not enough bits in the IndexArray to encode n things

  if (n==place)
  return One();   // Every encoding of bits in the IndexArray is used

  OrderIndexes(ia);

  int level = ia->GoHighestLevel();      

  return MakeMin2NR(n,place,ia);

}


BDDnode*
BDDmanager::MakeMin2NR( int n, int place, IndexArray* ia) {

  int cindex = ia->CurrIndex();
  ia->GoNextLevel();
  place=place>>1;

  if (n<place) {            // Need a zero term
    BDDnode* r1 = MakeMin2NR (n,place,ia);
    return MakeNode(cindex,Zero(),r1);
  }

  if (n==place) {          // We're done!
    return MakeNode(cindex,Zero(),One());
  }

  if (n>place) {           // Need a one term
    n = n%place;
    BDDnode* r0 = MakeMin2NR (n,place,ia);
    return MakeNode(cindex,r0,One());
  }
}


BDDnode*
BDDmanager::Int2BDD( int n, IndexArray *ia ) {
  if( n>= 1<<ia->Size() ) return NULL;	// not big enough!

  OrderIndexes(ia);
  ia->GoLowestLevel();
  BDDnode *minterm = One();

  int p=0;
  while( TRUE ) {
    if( n & (1<<p) )
      minterm = MakeNode( ia->CurrIndex(), minterm, Zero() );
    else
      minterm = MakeNode( ia->CurrIndex(), Zero(), minterm );

    if( ia->AtHighestLevel() ) break;
    ia->GoPrevLevel();
    p++;
    }
  return minterm;
  }

///////////////////////////////////////////////////////////////////////
//
// This performs a set union for a part of the bdd: example:
// f = { 1001, 1100, 0001 }, SetUnion(f, all vars) = { 1101 }
// i.e. if any position has a '1', then the result has a '1'
// over a portion of the tree:
// f = { 1000, 1100, 0010, 0001 } then
// SetUnion(f, last three vars) = { 1100, 0011 }
//
// A quick method of the function SetUnion from Brayton, et. al. :
// "A Fully Implicit Algorithm for Exact State Minimization" DAC 94
//

BDDnode*
BDDmanager::SetUnion( BDDnode* f, IndexArray* unionme, HashBase* ht) {
  if (unionme==NULL) unionme=defaultIA;
  boolean new_ht = (ht==NULL);
  if( new_ht )
    ht = new HasherTree;
  
  OrderIndexes( unionme);
  BDDnode* r = SetUnionR(f, unionme, ht);
  if( new_ht )
    delete ht;
  return r;
  }
    
BDDnode*
BDDmanager::SetUnionR( BDDnode* f, IndexArray* unionme, HashBase* ht ) {
  if( f==Zero() || f==One() ){ return f;}

  integer level = GetLevel(f);
  if( level < unionme->LowestLevel() )
    { return f;}

  BDDnode* r;
  if( f->NumReferences() > 1 && ht->Get((Object**)&r, f) )
    { return r;}

  BDDnode *f1, *f0;

  f1 = SetUnionR( f->Then(), unionme, ht );
  f0 = SetUnionR( f->Else(), unionme, ht );

  if( f1 != Zero() && f0 != Zero() && unionme->Find(GetLevel(f)) ) {
    f1 = or(f1,f0);
    f0 = Zero();
    }
  // otherwise, leave f1 and f0 as is (after filling in DC's below)

  integer i;
  for ( i=MAX(GetLevel(f1)+1,unionme->LowestLevel()) ; i<MIN( unionme->HighestLevel(), level) ; i++) {
    if  (unionme->Find(i) && f1 != Zero()) {
      integer index = unique->LookupIndex(i);
      f1 = MakeNode( index, f1, Zero());
      }
    }
  
  for ( i=MAX(GetLevel(f0)+1,unionme->LowestLevel()) ; i< MIN( unionme->HighestLevel(), level) ; i++) {
    if  (unionme->Find(i) && f0 != Zero()) {
      integer index = unique->LookupIndex(i);
      f0 = MakeNode( index, f0, Zero());
      }
    }
  r = ((f1==f0)?f1:MakeNode( f->GetIndex(), f1, f0));
  if( f->NumReferences() > 1 )
    ht->Put(r, f);
  return r;
  }


////////////////////////////////////////////////////////////////////////////
//
// UnateInfo**  FormUnateArray(BDDnode*, integer, UnateInfo**);
//

UnateInfo**
BDDmanager::FormUnateArray(BDDnode* f, integer size, UnateInfo** unate_array) {

  //UnateInfo**     unate_array = new UnateInfo*[size+1];
  //for(integer i=0; i<=size; i++) {unate_array[i] = new UnateInfo;}
  // I hate this size warning:
  size++;		// this means nothing
  //
  HasherTree *ht = new HasherTree;
  FormUnateArrayR(f,unate_array,ht);
  delete ht;
  return(unate_array);
}

// strange, maybe (certainly) confusing, but...
//
// UnateInfo** is an array of unate_info objects -- in a current versionit is a
// programmer's responsibility to create it and to make it of an appropriate size
// (i.e. neglect 'integer size' parameter).
//
// how to understand it (if it is possible)?
//
// element of an array corresponds to the variable of the same index -- value
// of Pos(),Neg() tell whether it is positive/negative unate -- if both are
// FALSE -> variable is either binate or does not exist at all in bdd (i.e.
// f does not depend on it -- since it is a true 'don't_care' it is not fixed
// in all true solutions) -- if both are TRUE => variable is unate ...
//
void
BDDmanager::FormUnateArrayR(BDDnode* f, UnateInfo** unate_array, HashBase* ht) {

if (f == one || f == zero)
    return;
if(ht->Put(*f,f)) {
        integer father_index = f->GetIndex();
        integer left_index = ((BDDnode*) f->Left())->GetIndex();
        integer right_index = ((BDDnode*) f->Right())->GetIndex();
        if (((BDDnode*) f->Left()) != zero) {
                unate_array[f->GetIndex()]->Neg(FALSE);
                if((father_index-left_index)>1) {
                        for(integer counter=(left_index+1); counter<father_index; counter++) {
                                unate_array[counter]->Neg(FALSE);
                                unate_array[counter]->Pos(FALSE);
                        }
                }
                FormUnateArrayR(((BDDnode*) f->Left()),unate_array,ht);
        }
        if (((BDDnode*) f->Right()) != zero) {
                unate_array[f->GetIndex()]->Pos(FALSE);
                if((father_index-right_index)>1) {
                        for(integer counter=(right_index+1); counter<father_index; counter++) {
                                unate_array[counter]->Neg(FALSE);
                                unate_array[counter]->Pos(FALSE);
                        }
                }
                FormUnateArrayR(((BDDnode*) f->Right()),unate_array,ht);
        }
}
return;
}


//***
//***
//
// Finds one path in OBDD -- generates a BitVector of size
// 2*(f->index+1)+1. There are two consecutive bits per variable
// of index i:
//                 [2i] = 0/1 (false/true)
//                 [2i+1] = 0/1 (don't care/care)
//
//                 [size-1] = 0/1 (false/solution exists)
//
// FindOnePath1 is the same -- just the max_index is provided...
//***

BitVector
BDDmanager::FindOnePath(BDDnode* f) {

  integer   size = 2*(f->GetIndex()+1)+1;
  BitVector b(size);
  if(FindOnePathR(f,b))
    b.Set(size-1);
  else
    b.Clear(size-1);
  return b;

}

BitVector
BDDmanager::FindOnePath1(BDDnode* f, integer max_index) {

  integer   size = 2*(max_index+1) + 1;
  BitVector b(size);
  if(FindOnePathR(f,b))
    b.Set(size-1);
  else
    b.Clear(size-1);
  return b;

}

boolean
BDDmanager::FindOnePathR(BDDnode* f, BitVector& b) {

  if(f==One())
    return TRUE;
  if(f==Zero())
    return FALSE;
  integer i = f->GetIndex();
  if( ((integer) (lrand48()%2)) == 0) {
    b.Clear(2*i); b.Set(2*i+1);
    if(FindOnePathR(f->Else(),b))
      return TRUE;
    b.Set(2*i); b.Set(2*i+1);
    return(FindOnePathR(f->Then(),b));
  } else {
    b.Set(2*i); b.Set(2*i+1);
    if(FindOnePathR(f->Then(),b))
      return TRUE;
    b.Clear(2*i); b.Set(2*i+1);
    return(FindOnePathR(f->Then(),b));
  }
}

////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////
//
// This first version of Distinguish assumes that the two sets of variables
// are interleaved. More specifically, one set of variables is
// held in IndexArray ia, and each variable in the second set
// must be 1 level (and 1 index) below the variable in ia.
// 
// IndexArray must hold set of higher numbered indexes
// and the routine returns results using this higher numbered
// set.
//
// max holds the maximum size of the set in which we are interested
// (for speed purposes only).
//

BDDnode*
BDDmanager::Distinguish(BDDnode* q, integer max, IndexArray *ia, HashBase *ht)
  {
  boolean new_ht;
  if( (new_ht=!ht) )
    ht = new HasherTree(FALSE,FALSE);
 
  OrderIndexes(ia);
  int max_width = ia->Count()-max;
  BDDnode *d;

  if( max_width > 0 && max != 0 ) {	// max == 0 turns off tuples
    BDDnode **temp = new BDDnode*[max_width];

    int i,j;

    for( i=0 ; i<max_width ; i++ )
      temp[i] = Zero();

    for( i=0 ; i<=max ; i++ ) {
      temp[0] = MakeNode( ia->Index(ia->Count()-i-1), temp[0], One());
      for( j=1 ; j<max_width ; j++ )
        temp[j] = MakeNode( ia->Index(ia->Count()-i-j-1), temp[j], temp[j-1]);
      }
    d = DistinguishR( q, max, ia, ht, temp );
    d = and(d,temp[max_width-1]);
    delete temp;
    }
  else
    d = DistinguishR( q, max, ia, ht, NULL );

  if( new_ht )
    delete ht;
 
  return d;
  }

////////////////////////////////////////////////////////////////////////////

BDDnode*
BDDmanager::DistinguishR(BDDnode *q, integer max, IndexArray *ia, HashBase *ht, BDDnode **t)
  {
  if( q==Zero() ) return One();
  if( q==One() ) return Zero();

  BDDnode *d;
  if( ht->Get((Object**)&d, q) )
    return d;

  integer level = GetLevel(q);
  integer I = q->GetIndex();
  integer nextI = I-1;

  BDDnode *d11, *d10, *d01, *d00, *d1, *d0;

  if( !(ia->Find(level)) ) {             // this is in set 1
    return and( DistinguishR( q->Then(),max,ia,ht,t ),
                DistinguishR( q->Else(),max,ia,ht,t ) );
    }
  d10 = Shannon0( q->Then(), nextI );
  d01 = Shannon1( q->Else(), nextI );
  d11 = Shannon1( q->Then(), nextI );
  d00 = Shannon0( q->Else(), nextI );

  d = DistinguishR( d11,max,ia,ht,t );
  if( d!=Zero() ) {
    d1 = DistinguishR( d00, max, ia, ht, t );
    d1 = and(d1,d);
    if( d1 == Zero() ) d = Zero();
    else {
      d0 = DistinguishR(d10,max,ia,ht,t);
      d0 = and( d0, DistinguishR(d01,max,ia,ht,t) );
      d0 = and( d0, d1 );
      if( d0 == d1 ) d= d0;
      else d = MakeNode( I, d1, d0 );
      }
    }
  int h = ia->GetHeight(level)-max;
  if( h > 0 && max != 0 && h%max == 0 )
    d = and(d, t[h-1]);

  ht->Put( d, q );

  return d;
  }

////////////////////////////////////////////////////////////////////////////
//
// The second version of Distinguish is identical, except that the two
// sets appear in two bdd's s1 and s2, rather than a single bdd.
//

BDDnode*
BDDmanager::Distinguish(BDDnode *s1, BDDnode *s2, int max, IndexArray *vars, HashBase *ht ) {

  if (vars==NULL) vars=defaultIA;
  boolean new_ht;
  if( (new_ht=!ht) )
    ht = new HasherTree(FALSE,TRUE);

  int max_width = vars->Count()-max;
  BDDnode *d;

  OrderIndexes(vars);

  if( max_width > 0 && max != 0 ) {	// max == 0 turns off tuples
    BDDnode **temp = new BDDnode*[max_width];

    int i,j;
    for( i=0 ; i<max_width ; i++ )
      temp[i] = Zero();

    for( i=0 ; i<=max ; i++ ) {
      temp[0] = MakeNode( vars->Index(vars->Count()-i-1), temp[0], One());
      for( j=1 ; j<max_width ; j++ )
        temp[j] = MakeNode( vars->Index(vars->Count()-i-j-1), temp[j], temp[j-1]);
      }
    d = DistinguishR( s1, s2, max, ht, vars, temp );
    d = and(d,temp[max_width-1]);
    delete temp;
    }
  else
    d = DistinguishR( s1, s2, max, ht, vars, NULL );

  if( new_ht )
    delete ht;

  return d;
  }

////////////////////////////////////////////////////////////////////////////

BDDnode*
BDDmanager::DistinguishR(BDDnode *s1, BDDnode *s2, int max, HashBase *ht, IndexArray *vars, BDDnode **t)
  {
  if( s1==Zero() || s2==Zero() ) return One();  // any set dist.'s Zero
  if( s1 ==One() || s2==One() ) return Zero();  // no set dist.'s One
  if( s1==s2 ) return Zero();                   // idents cant be dist.

  BDDnode *d;
  tempcomputed->Set( DISTINGUISH_OP, s1, s2, (BDDnode*)max );
  if( ht->Get((Object**)&d, tempcomputed) )
    return d;

  int index = (GetLevel(s1) > GetLevel(s2)) ? s1->GetIndex() : s2->GetIndex();

  BDDnode *d10, *d01, *d1, *d0;

  d = DistinguishR(Shannon1(s1, index), Shannon1(s2, index), max, ht, vars,t);
  if( d!=Zero() ) {
    d1 = DistinguishR(Shannon0( s1, index ), Shannon0( s2, index ), max, ht, vars,t);
    d1 = and(d,d1);
    if( d1==Zero() ) d=Zero();
    else {
      d10 = DistinguishR(Shannon1( s1, index ), Shannon0( s2, index ), max, ht, vars,t);
      d01 = DistinguishR(Shannon0( s1, index ), Shannon1( s2, index ), max, ht, vars,t);
      d0 = and(d01,d10);
      d0 = and(d1,d0);
      if( d0 == d1 ) d = d0;
      else d = MakeNode( index, d1, d0 );
      }
    }

  int h = vars->GetHeight(unique->LookupLevel(index))-max;
  if( h > 0 && max != 0 && h%max == 0 ) {
    d = and(d, t[h-1]);
    }

  ht->Put(d,new BDDCacheKey(DISTINGUISH_OP, s1, s2, (BDDnode*)max));

  return d;
  }

//////////////////////////////////////////////////////////////////////
//
// Distinguish Cubes (like above function but faster for cubes)
//
BDDnode*
BDDmanager::DistinguishCubes( BDDnode *c1, BDDnode *c2, IndexArray *vars ) {
  OrderIndexes(vars);

  List dist;
  int l;
  while( c1 != One() || c2!= One() ) {
    l = MAX( GetLevel(c1), GetLevel(c2) );
    if( vars->Find( l ) && c1->GetIndex() == c2->GetIndex() ) {
      if( (c1->Then() == Zero() && c2->Else() == Zero()) ||
	  (c1->Else() == Zero() && c2->Then() == Zero())   )
	dist.Add((Object*)c1->GetIndex());
      if( GetLevel(c1) == l )
	c1 = (c1->Then()==Zero()) ? c1->Else() : c1->Then();
      if( GetLevel(c2) == l )
	c2 = (c2->Then()==Zero()) ? c2->Else() : c2->Then();
      }
    }
  BDDnode *result = Zero();
  List_ForEach_Object_Remove( dist, I ) {
    result = MakeNode( (int)I, One(), result );
    }List_End_ForEach
  
  return result;
  }


////////////////////////////////////////////////////////////////////////
//
// Find the minimum number of variables set to '1' to satisfy the
// function
//

int
BDDmanager::MinCube( BDDnode *s, HashBase *ht ) {
  boolean new_ht;
  if( new_ht=(ht==0) )
    ht =  new HasherTree;
  
  int mincube = MinCubeR( s, ht );

  if( new_ht )
    delete ht;
  
  return mincube;
  }


int
BDDmanager::MinCubeR( BDDnode *s, HashBase *ht ) {

  if( s == One() ) return 0;
  if( s == Zero() ) return (INT_MAX-1);

  int left, right, top;
  if( ht->Get( (Object**)&top, s ) )
    return top;

  left = MinCubeR( s->Then(), ht );
  right = MinCubeR( s->Else(), ht );

  left += 1;
  top = MIN(left, right);

  ht->Put( (Object*)top, s );

  return top;
  }

////////////////////////////////////////////////////////////////////////
//
// Find the maximum number of variables set to '1' to satisfy the
// function
// This function assumes that there are no variables in f that are
// not in "vars"
//

int
BDDmanager::MaxCube( BDDnode *s, IndexArray *vars, HashBase *ht ) {
  boolean new_ht;
  if( new_ht=(ht==0) ) ht =  new HasherTree;
  
  int maxcube = MaxCubeR( s, vars, ht );

  if( new_ht ) delete ht;
  
  return maxcube;
  }


int
BDDmanager::MaxCubeR( BDDnode *s, IndexArray *vars, HashBase *ht ) {

  if( s == One() ) return 0;
  if( s == Zero() ) return INT_MIN;

  int left, right, top;
  if( ht->Get( (Object**)&top, s ) )
    return top;

  left = MaxCubeR( s->Then(), vars, ht );
  right = MaxCubeR( s->Else(), vars, ht );

  int ldist = vars->GetHeight( GetLevel(s) );
  int rdist = ldist;
  ldist -= vars->GetHeight( GetLevel( s->Then() ) );
  rdist -= vars->GetHeight( GetLevel( s->Else() ) );
  left += ldist;	// ldist is don't cares, 1 for this level
  right += rdist-1;	// rdist is don't cares
  top = MAX(left, right);

  ht->Put( (Object*)top, s );

  return top;
  }

///////////////////////////////////////////////////////////////////////////
//
// VectorAnd for large number of and's--allows for garbage collection
// during the operation
//
// MAKE SURE YOU HAVE KEPT ALL F's IN THE LIST, and note that the list
// is cleared, but not deleted

BDDnode*
BDDmanager::VectorAnd( List *f, int max_nodes, int flags ) {

				if( flags & VECTOR_AND_DEBUG )
				  cerr << "-"<<f->Count()<<"-";
  List next, *l1, *l2;
  l1 = f; l2 = &next;
  while( l1->Count() + l2->Count() > 1 ) {

    if( l1->Count() == 0 ) { l2->Reverse(); SWAP( List*, l1, l2 ); }
    BDDnode * f1 = ( BDDnode * )l1->Pop();
    if( l1->Count() == 0 ) { l2->Reverse(); SWAP( List*, l1, l2 ); }
    BDDnode * f2 = ( BDDnode * )l1->Pop();
    UnKeep( f1 );
    UnKeep( f2 );
    if( flags & VECTOR_AND_DO_COF )
      f1 = Cofactor( f1, f2 );
    f1 = and( f1, f2);

    Keep(f1);
				if( flags & VECTOR_AND_DEBUG )
				  cerr << ".";
    l2->Add(f1);
    if( f1 == Zero() ) break;		// abort!

    if( NumNodes() > max_nodes ) {
				if( flags & VECTOR_AND_DEBUG )
				  cerr << "(" << NumNodes() << ")g";
      GC();
				if( flags & VECTOR_AND_DEBUG )
				  cerr << "c("<<NumNodes() << ")";
      }
    }

  if( l1->Count() == 0 ) SWAP( List*, l1, l2 );
  switch (l1->Count()) {
    case 0: return One();	// only if list was initially empty
    case 1: return (BDDnode*)l1->Pop();
    default:
      while( l1->Count() )
	UnKeep( (BDDnode*)l1->Pop() );
    }
  // both lists should be empty now
  return Zero();
  }

////////////////////////////////////////////////////////////////////
//
// Similar to VectorAnd
//

BDDnode*
BDDmanager::VectorOr( List *f, int max_nodes, int flags ) {

				if( flags & VECTOR_AND_DEBUG )
				  cerr << "-"<<f->Count()<<"-";
  List next, *l1, *l2;
  l1 = f; l2 = &next;
  while( l1->Count() + l2->Count() > 1 ) {

    if( l1->Count() == 0 ) { l2->Reverse(); SWAP( List*, l1, l2 ); }
    BDDnode * f1 = ( BDDnode * )l1->Pop();
    if( l1->Count() == 0 ) { l2->Reverse(); SWAP( List*, l1, l2 ); }
    BDDnode * f2 = ( BDDnode * )l1->Pop();
    UnKeep( f1 );
    UnKeep( f2 );
    f1 = or( f1, f2);

    Keep(f1);
				if( flags & VECTOR_AND_DEBUG )
				  cerr << ".";
    l2->Add(f1);
    if( f1 == One() ) break;	// abort!

    if( NumNodes() > max_nodes ) {
				if( flags & VECTOR_AND_DEBUG )
				  cerr << "(" << NumNodes() << ")g";
      GC();
				if( flags & VECTOR_AND_DEBUG )
				  cerr << "c("<<NumNodes() << ")";
      }
    }

  if( l1->Count() == 0 ) SWAP( List*, l1, l2 );

  switch (l1->Count()) {
    case 0: return Zero();	// only if list was initially empty
    case 1: return (BDDnode*)l1->Pop();
    default:
      while( l1->Count() )
	UnKeep( (BDDnode*)l1->Pop() );
    }
  // both lists should be empty now

  return One();
  }



BDDnode*
BDDmanager::ComplementBits( BDDnode* f, IndexArray *ia ) {
  if (ia!=NULL) OrderIndexes(ia);
  if( f==One() || f==Zero() ) return f;

  if( ia && (GetLevel(f) < ia->LowestLevel()) )
    return f;
  
  BDDnode *t = ComplementBits( f->Then(), ia );
  BDDnode *e = ComplementBits( f->Else(), ia );

  if( !ia || ia->Find(GetLevel(f)) )
    return ite( Input(f->GetIndex()), e, t );
  return ite( Input(f->GetIndex()), t, e );
  }

///////////////////////////////////////////////////////////////////
// Returns FALSE (0) or 1 for same or 2 for different.
// Works for cubes, not sums

int
BDDmanager::AllButOneSameOrDiff( BDDnode *f1, BDDnode *f2 ) {
  int f1pol=2, f2pol=2;
  int match=2;

  while ( f1!=One() || f2!=One() ) {
      
    int i1 = GetLevel(f1);
    int i2 = GetLevel(f2);
    int p1 = f1->Else()==Zero();	// these vars contain the polarity
    int p2 = f2->Else()==Zero();
    if( i1>i2 ) {
      if( f1pol == p1 ) return 0;	// no match
      f1pol = !p1;
      if( f2pol != 2 && match != 2 && match !=(f1pol!=f2pol) ) return 0;
      if( f2pol != 2 && match == 2 ) match = (f1pol!=f2pol);
      f1 = (p1) ? f1->Then() : f1->Else();
      continue;
      }
    if( i1<i2 ) {
      if( f2pol == p2 ) return 0;	// no match
      f2pol = !p2;
      if( f1pol != 2 && match != 2 && match !=(f1pol!=f2pol) ) return 0;
      if( f1pol != 2 && match == 2 ) match = (f1pol!=f2pol);
      f2 = (p2) ? f2->Then() : f2->Else();
      continue;
      }
    // i1 == i2
    if( match == 2 ) match = (p1!=p2);
    if( match != (p1!=p2) ) return 0;
    if( f1==f2 ) break;

    f1 = (p1) ? f1->Then() : f1->Else();
    f2 = (p2) ? f2->Then() : f2->Else();
    }
  if( match == 2 ) return (f1pol != f2pol)+1;
  return (match)+1;
  }

BDDnode*
BDDmanager::CommonZeros( BDDnode *f1, BDDnode *f2, int &n1, int &n0 ) {
  int i, max = MAX(f1->GetIndex(),f2->GetIndex());
  char cube[4096];
  ASSERT( max < 4094, "Index too high for static character array (>4094)" );
  for( i=0 ; i<=max ; cube[i++] = '-' );
  cube[max] = '\0';

  while ( f1!=One() && f2!=One() ) {
    int p1 = f1->Else()==Zero();	// these vars contain the polarity
    int p2 = f2->Else()==Zero();
    int l1 = GetLevel(f1);
    int l2 = GetLevel(f2);
    max = MAX(l1,l2);
    if( l2 == l1 && p1==p2 ) {
      cube[f1->GetIndex()] = (char)(p1==1)+'0';
      (p1) ? n1++ : n0++;
      }

    if( l1==max ) f1 = (p1) ? f1->Then() : f1->Else();
    if( l2==max ) f2 = (p2) ? f2->Then() : f2->Else();
    }
  return ReadCube( cube );
  }

//////////////////////////////////////////////////////////
// Tuple-like functions for patterns.  
// The pattern in duplicated over the index arrays mod
// the highest level in the pattern
// These are overloaded functions
// Steve Haynal 3/2/98

BDDnode*
BDDmanager::AtLeastN( integer n, IndexArray* vars, BDDnode* pattern) {

  if (vars==NULL) vars=defaultIA;

  integer psize = pattern->GetIndex() + 1;

  ASSERT(vars->Count()%psize == 0, "Index array is not divisable by BDD pattern levels.");

  integer max_width = (vars->Count()/psize) - (--n);
  if ( max_width <= 0 ) return Zero();

  OrderIndexes(vars);

  BDDnode** temp = new BDDnode*[max_width];
  int i,j;
  for ( i=0; i< max_width; i++)
    temp[i] = One();

  for (i=0; i<=n; i++) {
    temp[0] = MakePatternR( vars, vars->Count()-((i+1)*psize), pattern, temp[0], Zero());
    for( j=1; j < max_width; j++)
      temp[j] = MakePatternR( vars, vars->Count()-((i+j+1)*psize), pattern, temp[j], temp[j-1]);
  }
  BDDnode* result = temp[max_width-1];
  delete temp;

  return result;
}

BDDnode*
BDDmanager::Tuple( integer n, IndexArray* vars, BDDnode* pattern) {
  if (vars==NULL) vars=defaultIA;
  // initial checks

  integer psize = pattern->GetIndex() + 1;

  ASSERT(vars->Count()%psize == 0, "Index array is not divisable by BDD pattern levels.");

  integer max_width = (vars->Count()/psize) - (--n);
  if ( max_width <= 0 ) return Zero();

  OrderIndexes(vars);

  BDDnode** temp = new BDDnode*[max_width];

  temp[0] = One();
  BDDnode* notpattern = not(pattern);
  
  int i, j;
  for ( i=1; i< max_width; i++)
    temp[i] = MakePatternR( vars, vars->Count()-(i*psize), pattern, Zero(), temp[i-1]);
  
  for (i=0; i<=n; i++) {
    temp[0] = MakePatternR( vars, vars->Count()-((i+1)*psize), pattern, temp[0], Zero());
    for( j=1; j < max_width; j++)
      temp[j] = MakePatternR( vars, vars->Count()-((i+j+1)*psize), pattern, temp[j], temp[j-1]);
  }

  BDDnode* result = temp[max_width-1];
  delete temp;

  return result;
}

BDDnode*
BDDmanager::UpToN( integer n, IndexArray* vars, BDDnode* pattern) {

  if (vars==NULL) vars=defaultIA;

  integer psize = pattern->GetIndex() + 1;

  ASSERT(vars->Count()%psize == 0, "Index array is not divisable by BDD pattern levels.");

  integer max_width = (vars->Count()/psize) - n;
  if ( max_width <= 0 ) return One();

  OrderIndexes(vars);

  BDDnode** temp = new BDDnode*[max_width+1];
  int i,j;
  for ( i=0; i< max_width; i++)
    temp[i] = Zero();
  temp[max_width] = One();

  for ( i=n; i>=0; i--) {
    for( j=max_width-1 ; j >= 0; j--) {
      temp[j] = MakePatternR( vars, ((i+j)*psize), pattern, temp[j], temp[j+1]);
    }
  }


  BDDnode* result = temp[0];
  delete temp;

  return result;
}

BDDnode* 
BDDmanager::MakePatternR( IndexArray* vars, integer ipos, BDDnode* pattern, BDDnode* troot, BDDnode* froot) {

  if (pattern == one) return troot;
  if (pattern == zero) return froot;

  integer ipost = ipos + (pattern->GetIndex() - pattern->Then()->GetIndex());
  integer iposf = ipos + (pattern->GetIndex() - pattern->Else()->GetIndex());

  return MakeNode(vars->Index(ipos), MakePatternR(vars, ipost, pattern->Then(), troot, froot), MakePatternR(vars, iposf, pattern->Else(), troot, froot));
}

//////////////////////////////////////////////////////////////////////////
//
// calculate reachable states given a start state, next state relation,
// list of present state variables, and list of next state variables
//

BDDnode*
BDDmanager::Reach1( BDDnode *ss, BDDnode *nsr, IndexArray *psia, IndexArray *nsia ){
  BDDnode *rs = ss;
  HasherTree smooth;

  // create a way of transforming next state vars to present state vars
  HasherTree ns2ps;
  psia->GoHighestLevel();
  nsia->GoHighestLevel();

  while(1) {
    ns2ps.Put( Input( psia->CurrIndex() ), Input( nsia->CurrIndex() ) );
    if( psia->AtLowestLevel() ) break;
    psia->GoNextLevel();
    nsia->GoNextLevel();
    }

  while( ss != Zero() ) {
    BDDnode *ns = and(ss, nsr);
    ns = Smooth( ns, nsia, &smooth, FALSE );
    ns = Translate( ns, &ns2ps );
    ss = and_not( ns, rs );
    rs = or( rs, ns );
    }
  
  return rs;
  }

//////////////////////////////////////////////////////////////////////////
//
// calculate the set of steady states that continually exist
// list of present state variables, and list of next state variables
//

BDDnode*
BDDmanager::SteadyState1( BDDnode *ss, BDDnode *nsr, IndexArray *psia, IndexArray *nsia ){
  HasherTree smooth;

  // create a way of transforming next state vars to present state vars
  HasherTree ns2ps;
  psia->GoHighestLevel();
  nsia->GoHighestLevel();

  while(1) {
    ns2ps.Put( Input( psia->CurrIndex() ), Input( nsia->CurrIndex() ) );
    if( psia->AtLowestLevel() ) break;
    psia->GoNextLevel();
    nsia->GoNextLevel();
    }

  BDDnode *lasts = Zero();
  while( ss != lasts ) {
    lasts = ss;
    ss = and(ss, nsr);
    ss = Smooth( ss, nsia, &smooth, FALSE );
    ss = Translate( ss, &ns2ps );
    }
  
  return ss;
  }
