#ifndef BLIFDATA_H
#define BLIFDATA_H

#include "list.h"
#include "strobj.h"
#include "bddm.h"
#include "hash.h"
#include "hasher.h"


class BLIFnode : public Object {
  StringObj *name;
  HasherTree inputs;
  int globalIndex;
  BDDnode *func;
  HasherTree *translate;

public:
  BLIFnode( StringObj *n, List *ins, BDDnode *f, int gI, HasherTree *tr,
	    BDDmgr *bdd ) {
    inputs.SetDelete( FALSE, TRUE );
    name = n;
    int i=0;
    List_ForEach_Type( ins, StringObj, n ) {
      inputs.Put( bdd->Input(i++), new StringObj( n ) );
      }List_End_ForEach
    func = f;
    globalIndex = gI;
    translate = tr;
    }
  BLIFnode( BLIFnode &bn );
 ~BLIFnode() {
    inputs.Clear();
    translate->Clear();
    delete name, translate;
    }

  int Index() { return globalIndex; }
  BDDnode* Function() { return func; }
  StringObj *Name() { return name; }
  HasherTree *Translate() { return translate; }
  HasherTree &Inputs() { return inputs; }

  };


class BLIFdata : public Object {
  StringObj model;

  // names of Primary outputs and next state functions
  List po, ffin;	// contain their own copies of names

  // names of primary inputs, present state vars,
  // and intermediate nodes
  List pi, ffout;	// contain their own copies of names
 
  HasherTree name2bddvar;	// these two share stringobj's,
  HasherTree bddvar2name;	// so we only delete names once
  HasherTree name2node;		// contains own copy of name

  BDDnode *reset;
  BDDmgr *bddm;
  int keepGlobalFunctions;

public:
  BLIFdata( BDDmgr *b, int kgf = FALSE ) { bddm = b;
			      name2bddvar.SetDelete(FALSE,FALSE);
			      bddvar2name.SetDelete(TRUE,FALSE);
			      name2node.SetDelete(TRUE,TRUE);
			      reset = bddm->One();
			      keepGlobalFunctions = kgf; }

  BLIFdata( BLIFdata &bf );
 ~BLIFdata();

  virtual Object *Copy() { return new BLIFdata(*this); }
  void SetModel( char *m ) { model = m; }

  void AddOutput( char *o ) { po.Append( new StringObj( o ) ); }

  void AddInput( char *o ) {
    BDDnode *v = AssignNextIndex( o );
    pi.Append( new StringObj( o ) );
    }
  void AddLatch( char *in, char *out, char *r ) {
    BDDnode *v = AssignNextIndex( out );
    ffout.Append( new StringObj( out ) );
    ffin.Append( new StringObj( in ) );

    bddm->UnKeep( reset );
    if( r[0] == '0' )
      reset = bddm->and_not( reset, v );
    if( r[0] == '1' )
      reset = bddm->and( reset, v );
    bddm->Keep( reset );
    }

  BDDnode *AssignNextIndex( char *n ) {
    StringObj name(n);
    return AssignIndex( &name, name2bddvar.Count() );
    }

  BDDnode *AssignIndex( StringObj *n , int index ) {

    BDDnode *var;
    if( !(name2bddvar.Get((Object**)&var, n )) ) {
      var = bddm->Input(index);
      StringObj *name = new StringObj(n);
      name2bddvar.ForcePut( var, name );
      bddvar2name.ForcePut( name, var );
      bddm->Keep(var);
      }
    return var;
    }


  void AddFunctionGlobalVars( char *out, BDDnode *f ) {
    HasherTree *gl = new HasherTree(FALSE, FALSE);
    HasherTree *lg = new HasherTree(FALSE, FALSE);
    HasherTree *tr;
    List in;
    HasherTree sup;
    bddm->Support(f, &sup);
    int localindex = 0;
    Hasher_ForEach_Type( &sup, BDDnode, k, index ) {
      in.Add( GetVar( index ) );
      gl->Put( bddm->Input( localindex ), index );
      if( keepGlobalFunctions==FALSE )
	lg->Put( index, bddm->Input( localindex ) );
      localindex++;
      }Hasher_End_ForEach
    in.Reverse();

    if( keepGlobalFunctions ) {
      tr = gl;
      delete lg;
      }
    else {
      f = bddm->Translate(f, gl );
      tr = lg;
      delete gl;
      }

    BDDnode* i = AssignNextIndex( out );
    StringObj *n = new StringObj( out );
    bddm->Keep(f);
    BLIFnode *bn = new BLIFnode( n, &in, f, i->GetIndex(), tr, bddm );
    name2node.ForcePut( bn, n );
    }

  void AddFunctionLocalVars( List *in, BDDnode *f, char *out ) {

    bddm->Keep(f);
    HasherTree *gl = new HasherTree(FALSE,FALSE);
    HasherTree *lg = new HasherTree(FALSE,FALSE);
    HasherTree *tr;

    int var=0;
    List_ForEach_Type( in, StringObj, v ) {

      BDDnode *t = AssignNextIndex( v->String() );
      lg->Put( t, bddm->Input(var) );
      if( keepGlobalFunctions )
	gl->Put( bddm->Input(var), t );
      var++;
      }List_End_ForEach

    if( keepGlobalFunctions ) {
      f = bddm->Translate( f, lg );
      delete lg;
      tr = gl;
      }
    else {
      delete gl;
      tr = lg;
      }
    BDDnode* i = AssignNextIndex( out );
    StringObj *n = new StringObj( out );
    BLIFnode *bn = new BLIFnode( n, in, f, i->GetIndex(), tr, bddm );
    name2node.ForcePut( bn, n );
    }

  void ModifyFunctionGlobalVars( char *out, BDDnode *f ) {
    StringObj name(out);

    BLIFnode *bn = (BLIFnode *) name2node.Get(name);
    if( bn ) bddm->UnKeep(bn->Function());

    if( !(name2node.Remove( &name )) )
      cerr << "Warning, cant find blifnode " << out << " to modify.\n";

    bddm->Keep( f );
    AddFunctionGlobalVars( out, f );
    }

  void ModifyFunctionLocalVars( List *in, BDDnode *f, char *out ) {
    StringObj name(out);

    BLIFnode *bn = (BLIFnode *) name2node.Get(name);
    if( bn ) bddm->UnKeep(bn->Function());

    if( !(name2node.Remove( &name )) )
      cerr << "Warning, cant find blifnode " << out << " to modify.\n";
    bddm->Keep( f );
    AddFunctionLocalVars( in, f, out );
    }

  // user is required to make sure that this function is no longer
  // in the support set of any nodes
  void RemoveFunction( StringObj *out ) {
    BDDnode *var = (BDDnode*) name2bddvar.Get( out );

    BLIFnode *bn = (BLIFnode *) name2node.Get( out );
    if( bn ) bddm->UnKeep(bn->Function());

    name2bddvar.Remove( out );
    bddvar2name.Remove( var );
    name2node.Remove( out );
    }


  // Brute Force
  BDDnode *FuncOfPI( StringObj *name );
  BDDnode *FuncOfAll( StringObj *name );
    
  // after use, delete lists but not data
  List *GetPINamesL();            // Returns primary inputs in a List
  List *GetPONamesL();            // Returns primary outputs in a List
  List *GetPSNames();
  List *GetNSNames();
  // after use, delete tree and data, not key
  HasherTree *GetPINamesH(void);  // Returns primary inputs in a HashTree
  HasherTree *GetPONamesH(void);  // Returns primary outputs in a HashTree
  List *GetFuncNames();

  BDDnode *GetGlobalFunc( StringObj *name ) {
    BLIFnode *bn = (BLIFnode *) name2node.Get(name);
    BDDnode *temp = bn->Function();
    if( temp==NULL )
      return GetVar(name);

    if( keepGlobalFunctions ) return temp;
    else return bddm->Translate( temp, bn->Translate() );
    }
  
  BDDnode *GetLocalFunc( StringObj *name ) {
    BLIFnode *bn = (BLIFnode *) name2node.Get(name);
    BDDnode *temp = bn->Function();
    if( temp==NULL )
      return GetVar(name);

    if( keepGlobalFunctions )  return bddm->Translate( temp, bn->Translate() );
    else return temp;
    }
  
  BDDnode *GetVar( StringObj *name ) {
    return (BDDnode*) name2bddvar.Get(name);
    }

  StringObj* GetVar( BDDnode *var ) {
    return (StringObj*) bddvar2name.Get(var);
    }
  
  HasherTree *BDDvar2Name() {
    HasherTree *b2n = new HasherTree( &bddvar2name, TRUE );
    return b2n;
    }

  List *GetFuncInputs( StringObj *fname ); // delete list but not keys after use
  List *GetFuncInputs( BDDnode *fname );   // delete list but not keys after use
  HasherTree *GetFuncInputsH( StringObj *fn );	// don't delete these after use
  HasherTree *GetFuncInputsH( BDDnode *fn );	// don't delete these after use
  HasherTree *GetDrivenFuncs( StringObj *name, HasherTree *df=NULL );
  // just nodes
  int NumNodes() { return name2node.Count(); }
  // nodes + inputs
  int NumNames() { return name2bddvar.Count(); }
  virtual void Print( ostream &out = cout );


  };

#endif
