////////////////////////////////////////////////////////////////////////////
//
//  bitSparse.C : sparse bit matrix class         
//  Andrew Seawright
//
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>
#include "bsparse.h"

BitSparse::BitSparse(int xs, int ys) {

  int i;
  count = 0;
  xsize = xs;
  ysize = ys;
  x  =             new BitSparseCell*[xs];
  y  =             new BitSparseCell*[ys];
  num_in_row   =   new int[ys];
  num_in_col   =   new int[xs];

  for (i=0; i<xs; ++i) {
    x[i] = 0;
    num_in_col[i] = 0;
  }

  for (i=0; i<ys; ++i) {
    y[i] = 0;
    num_in_row[i] = 0;
  }
}

BitSparse::BitSparse(istream& in, boolean load) {

  int i;
  count = 0;
  in >> xsize >> ysize;
  x  =             new BitSparseCell*[xsize];
  y  =             new BitSparseCell*[ysize];
  num_in_row   =   new int[ysize];
  num_in_col   =   new int[xsize];

  for (i=0; i<xsize; ++i) {
    x[i] = 0;
    num_in_col[i] = 0;
  }

  for (i=0; i<ysize; ++i) {
    y[i] = 0;
    num_in_row[i] = 0;
  }
  if (load)
    Set( in);
}

BitSparse::~BitSparse() {

  BitSparseCell *p, *q;
  for (int i=0; i < xsize; ++i) {
    for (p=x[i]; p!=0; p=q) {
      q = p->nexty;
      delete p;
    }
  }
  delete x;
  delete y;
  delete num_in_row;
  delete num_in_col;
}

void
BitSparse::Resizex(int newx) {
  ASSERT( (count == 0), "BitSparse not all clear, can't resize");

  delete x;
  xsize = newx;
  x = new BitSparseCell*[newx];
  for (int i=0; i<newx; ++i) 
    x[i] = 0;
}

void
BitSparse::Set(int i, int j) {
  ASSERT( (i < xsize && j <ysize), "array index out of range");
  BitSparseCell *p, *px, *py;

  for (py=0,p=x[i]; p!=0; py=p,p=p->nexty) {
    if (p->y > j)
      break;
    if (p->y == j)
      return;
  }
  for (px=0,p=y[j]; p!=0; px=p,p=p->nextx) {
    if (p->x > i)
      break;
    if (p->x == i)
      return;
  }

  BitSparseCell *q = new BitSparseCell;
  q->x = i;
  q->y = j;

  if (px == 0) {
    q->nextx = y[j];
    y[j] = q;
  } else {
    q->nextx = px->nextx;
    px->nextx = q;
  }
  if (py == 0) {
    q->nexty = x[i];
    x[i] = q;
  } else {
    q->nexty = py->nexty; 
    py->nexty = q;
  }
  ++count;
  ++num_in_col[i];
  ++num_in_row[j];
}

void
BitSparse::Clear(int i, int j) {
  ASSERT( (i < xsize && j <ysize), "array index out of range");
  BitSparseCell *p, *px, *py;
  for (py=0,p=x[i];; py=p,p=p->nexty) {
    if (p == 0 || p->y > j)
      return;
    if (p->y == j)
      break;
  }
  for (px=0,p=y[j];; px=p,p=p->nextx) {
    if (p == 0 || p->x > i)
      return;
    if (p->x == i)
      break;
  }
  if (px == 0)
    y[j] = p->nextx;
  else
    px->nextx = p->nextx;
  if (py == 0)
    x[i] = p->nexty;
  else
    py->nexty = p->nexty;
  delete p;
  --count; 
  --num_in_col[i];
  --num_in_row[j];
}

boolean
BitSparse::Test(int i, int j) {
  ASSERT( (i < xsize && j <ysize), "array index out of range");
  BitSparseCell *p;
  for (p=x[i]; p!=0; p=p->nexty) {
    if (p->y > j)
      return FALSE;
    if (p->y == j)
      return TRUE;
  }
  return FALSE;
}

boolean
BitSparse::TestRow(int j) {
  ASSERT( (j < ysize), "array index out of range");
  return (y[j]!=0);
}

boolean
BitSparse::TestCol(int i) {
  ASSERT( i < xsize, "array index out of range");
  return (x[i]!=0);
}

////////////////////////////////////////////////////////////////////////////

void
BitSparse::RowMoveI(int a, int b, int c, int d) {   
  boolean val1 = Test(a, d);
  boolean val2 = Test(c, d);
  if (val1)        Set(c, d);
  else if (val2) Clear(c, d);  
  if (val2)        Set(a, d);
  else if (val1) Clear(a, d);

  if ( b==d) return;
  val1 = Test(a, b);
  val2 = Test(c, b);
  if (val1)        Set(c, b);
  else if (val2) Clear(c, b);
  if (val2)        Set(a, b);
  else if (val1) Clear(a, b);
}

////////////////////////////////////////////////////////////////////////////
void
BitSparse::Print( ostream& out) {                   // better printer function

  BitSparseCell *p;
  int i, j, n;

  out << "   ";
  for (j=0; j<xsize; ++j)
    out << "--";
  out << "-\n";
  for (i=0; i<ysize; ++i) {
    out.width(3);
    out << i;
    for (n=0, p=y[i]; p!=0; n = p->x + 1, p=p->nextx) {
      for (j=0; j < (p->x - n); ++j)
	out << "| ";
      out << "|@";
    }
    for (j=0; j < xsize-n; ++j)
      out << "| ";
    out << "|\n   ";
    for (j=0; j<xsize; ++j)
      out << "--";
    out << "-\n";
  }
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

void
BitSparse::Set( istream&  in) {
  integer r;
  integer c;
  integer t_count;

  in >> t_count;
  for (int i=0; i < t_count; i++) {
    in >> c >> r;
    Set( c, r);
  }
}

void
BitSparse::Clear( istream&  in) {
  integer r;
  integer c;
  integer t_count;

  in >> t_count;
  for (int i=0; i < t_count; i++) {
    in >> c >> r;
    Clear( c, r);
  }
}


