//
//
// BitVector packs bits in conescutive "integers" 
// of memory. The first element of a BitVector is
// stored in the first bit [0] of the first word
// (bit_vector[0]). That bit is treated as the
// most significant bit (MSB) of the BitVector 
// (it is important to be aware of that fact since
// it is used in Compare and Hash functions).i
//
//

#include <iostream.h>
#include "bv.h"

//
// This function constructs a BitVector of a size 'n'
// and initializes it to all zeros (including the 
// unused bits in the most significant word of a
// bit_vector[]).
// 
BitVector::BitVector(integer n) {
  integer i;
  num_bits = n;
  num_words = (n-1)/BIT_VEC_WORD_SIZE + 1;
  bit_vector = new bit_vector_word[num_words];
  for (i=0; i<num_words; i++) {
    bit_vector[i] = 0;
  }
}

//
// This function constructs a BitVector from a BitVector
// reference.
//
BitVector::BitVector(const BitVector& x) {
  integer i = x.num_words;
  num_bits = x.num_bits;
  bit_vector = new bit_vector_word[num_words=i];
  bit_vector_word* p = bit_vector;
  bit_vector_word* q = x.bit_vector;
  while (i--) *p++ = *q++; 
}

//
// This function constructs a BitVector from a pointer
// to a BitVector.
//
BitVector::BitVector(const BitVector *x) {
  integer i = x->num_words;
  num_bits = x->num_bits;
  bit_vector = new bit_vector_word[num_words=i];
  bit_vector_word* p = bit_vector;
  bit_vector_word* q = x->bit_vector;
  while (i--) *p++ = *q++;
}

//
// function resets bit vector to all zeros
//

void
BitVector::ClearAll() {

  integer i;
  for (i=0; i<num_words; ++i)
    bit_vector[i] = 0;
}

//
// This function returns the length of the BitVector.
//
integer  BitVector::VectorLength() {
   return(num_bits);
}

// 
// This function sets the n-th bit of the BitVector.
//
void BitVector::Set(integer n) {
  if (n>=num_bits)    
    FATAL("ERROR:cannot set a bit out of range")
  bit_vector_word temp_word = n/BIT_VEC_WORD_SIZE;
  bit_vector_word temp_bit = n%BIT_VEC_WORD_SIZE;
  bit_vector_word temp = (BIT_VEC_ONE << temp_bit);
  bit_vector[temp_word] = bit_vector[temp_word] | temp;
}

// 
// This function clears the n-th bit of a BitVector.
//
void BitVector::Clear(integer n) {
  if (n>=num_bits)
    FATAL("ERROR:cannot clear a bit out of range")
  bit_vector_word temp_word = n/BIT_VEC_WORD_SIZE;
  bit_vector_word temp_bit = n%BIT_VEC_WORD_SIZE;
  bit_vector_word temp = ~(BIT_VEC_ONE << temp_bit);
  bit_vector[temp_word] = bit_vector[temp_word] & temp;
}

// 
// This is a private function. It is used to clear  all
// the unused bits of the BitVector (to keep them zero
// at all times. PClear is protected: a user cannot 
// access the unused bits (out-of-range error).
//
void BitVector::PClear(integer n) {
  bit_vector_word temp_word = n/BIT_VEC_WORD_SIZE;
  bit_vector_word temp_bit = n%BIT_VEC_WORD_SIZE;
  bit_vector_word temp = ~(BIT_VEC_ONE << temp_bit);
  bit_vector[temp_word] = bit_vector[temp_word] & temp;
}
 
//
// return the value of the n-th bit of a BitVector.
// 
boolean BitVector::GetElement(int n) {
  if (n>=num_bits)    
    FATAL("ERROR:cannot extract a bit out of range")
  bit_vector_word temp_word = n/BIT_VEC_WORD_SIZE;
  bit_vector_word temp_bit = n%BIT_VEC_WORD_SIZE;
  bit_vector_word temp = bit_vector[temp_word];
  temp = temp >> temp_bit;
  return((boolean) (temp = temp & BIT_VEC_ONE));
}

// 
// This function overloads ~ operator to perform a
// complementation of a BitVector (the unused bits
// are kept consistently 0). It generates a temporary
// storage on the stack (original BitVector is not
// modified.
//
BitVector BitVector::operator~() {
  integer i = num_words;
  BitVector temp = *this;
  for(i=0; i<num_words; i++) {
    temp.bit_vector[i] = ~(temp.bit_vector[i]);
  }
  for(i=num_bits; i<(num_words*(sizeof(bit_vector_word))*8); i++) {
    temp.PClear(i);
  }
  return(temp);
}

// 
// This function overloads = operator so that the
// assignments such as (a = b), where both a and b
// are BitVectors, performs an actual copying of 
// the structure.
//
BitVector BitVector::operator=(const BitVector& x) {
  integer i;
  num_bits = x.num_bits;
  num_words = x.num_words;  
  for(i=0; i<num_words; i++) {
    bit_vector[i] = x.bit_vector[i];
  }
  return(*this);
}

// 
// The following two functions perform a BitVector 
// shifting (left/right) by 'n'. Zeros are are shifted
// in.
// 
void BitVector::ShiftLeft(integer n) {
  if (n>=num_bits)  
    FATAL("ERROR:shift out of range")  
  integer i = n;
  for(i=(num_bits-1-n); i>=0; i--) {
    if ((*this)[i]==0) this->Clear(i+n);
      else this->Set(i+n);
  }
  for(i=(n-1); i>=0; i--) {
    this->Clear(i);
  }
}

void BitVector::ShiftRight(integer n) {
  if (n>=num_bits)
    FATAL("ERROR:shift out of range") 
  integer i = n;
  for(i=0; i<(num_bits-n); i++) {
    if ((*this)[i+n]==0) this->Clear(i);
      else this->Set(i);
  }
  for(i=(num_bits-n); i<num_bits; i++) {
    this->Clear(i);
  }
}

// 
// The following two functions perform a BitVector 
// rotation (left/right) by 'n'. 
//
void BitVector::RotateLeft(integer n) {
  if (n>=num_bits) 
    FATAL("ERROR:shift out of range")
  integer i = n;
  BitVector temp(num_bits);
  temp.num_bits = num_bits;
  temp.num_words = num_words;
  for(i=0; i<num_bits; i++) {
    if ((*this)[i]==0) temp.Clear((i+n)%num_bits);
      else temp.Set((i+n)%num_bits);
  }
  for(i=0; i<num_words; i++) {
    bit_vector[i] = temp.bit_vector[i];
  }
}

void BitVector::RotateRight(integer n) {
  if (n>=num_bits) 
    FATAL("ERROR:shift out of range")
  integer i = n;
  BitVector temp(num_bits);
  temp.num_bits = num_bits;
  temp.num_words = num_words;
  for(i=0; i<num_bits; i++) {
    if ((*this)[(i+n)%num_bits]==0) temp.Clear(i);
      else temp.Set(i);
  }
  for(i=0; i<num_words; i++) {
  bit_vector[i] = temp.bit_vector[i];
  }
}

boolean 
BitVector::AllClear() {
  integer test = 0;
  for ( integer i=0; i<num_words; i++) {
    test = test|bit_vector[i];
  }
  return test == 0;
}


integer BitVector::NumOnes()
  {
  integer count = 0;

  for( integer i=0 ; i<num_bits ; i++ )
    if( (*this)[i] & 1 )
      count++;
  return count;
  }

// 
// This is a BitVector Print function. It prints info
// on the number of bits and words of a BitVector, and
// all the bit_vector[] elements.
//
void BitVector::Print(ostream& out) {
  integer i = num_words;
  out << "num_bits=" << num_bits << "\n" ;
  out << "num_words=" << num_words << "\n" ;

  for(i=num_words-1; i>=0; i--) {
    out << "word " << i << " = [" << bit_vector[i] << "]\n";
  }

}

// 
// This is another BitVector Print function. 
// It prints all the bit_vector[] elements.
//

void BitVector::Print2(ostream& out) {
  integer i = num_words;

  out << "[";
  for(i=num_words-1; i>=0; i--) {
    PrintWordHex(out, bit_vector[i]);
  }
  out << "]\n";
}

//    
// The following functions perform logical operations
// on BitVectors. They work via accumulation.
// Usage: a|=b, a&=b, a^=b.
//
BitVector& BitVector::operator |=(BitVector& f) {
  if(num_bits!=f.num_bits) 
    FATAL("ERROR:operand sizes do not match")
  integer i;
  for(i=0; i<num_words; i++) {
    bit_vector[i] = bit_vector[i] | f.bit_vector[i];
  }
  return(*this);
}

BitVector& BitVector::operator &=(BitVector& f) {
  if(num_bits!=f.num_bits) 
    FATAL("ERROR:operand sizes do not match")
  integer i;
  for(i=0; i<num_words; i++) {
    bit_vector[i] = bit_vector[i] & f.bit_vector[i];
  }
  return(*this);
}

BitVector& BitVector::operator ^=(BitVector& f) {
  if(num_bits!=f.num_bits) 
    FATAL("ERROR:operand sizes do not match")
  integer i;
  for(i=0; i<num_words; i++) {
    bit_vector[i] = bit_vector[i] ^ f.bit_vector[i];
  }
  return(*this);
}

//     
// The following functions perform logical operations
// on BitVectors. They create a temporary storage on
// the stack. Usage: a|b, a&b, a^b.
//
BitVector BitVector::operator |(BitVector& f) {
  if(num_bits!=f.num_bits) 
    FATAL("ERROR:operand sizes do not match")
  integer i;
  BitVector temp(num_bits);
  temp.num_bits = num_bits;
  temp.num_words = num_words;
  for(i=0; i<num_words; i++) {
    temp.bit_vector[i] = bit_vector[i] | f.bit_vector[i];
  }
  return(temp);
}

BitVector BitVector::operator &(BitVector& f) {
  if(num_bits!=f.num_bits)
    FATAL("ERROR:operand sizes do not match") 
  integer i;
  BitVector temp(num_bits);
  temp.num_bits = num_bits;
  temp.num_words = num_words;
  for(i=0; i<num_words; i++) {
    temp.bit_vector[i] = bit_vector[i] & f.bit_vector[i];
  }
  return(temp);
}

BitVector BitVector::operator ^(BitVector& f) {
  if(num_bits!=f.num_bits) 
    FATAL("ERROR:operand sizes do not match")
  integer i;
  BitVector temp(num_bits);
  temp.num_bits = num_bits;
  temp.num_words = num_words;
  for(i=0; i<num_words; i++) {
    temp.bit_vector[i] = bit_vector[i] ^ f.bit_vector[i];
  }
  return(temp);
}

void
BitVector::Resize( int n ) {

  num_bits = n;

  int new_num_words = (n-1)/BIT_VEC_WORD_SIZE + 1;
  if( new_num_words == num_words )
    return;	// already set num_bits
  
  int i;
  bit_vector_word *new_bit_vector = new bit_vector_word[new_num_words];

  for (i=0; i<num_words; i++) {
    if( i>=new_num_words ) break;
    new_bit_vector[i] = bit_vector[i];
    }
  delete bit_vector;
  num_words = new_num_words;
  bit_vector = new_bit_vector;
  }

//
// The following functions perform BitVector expansion
// from 'num_bits' to 'n' by adding 0's/1's.
// 
BitVector BitVector::Expand0(integer n) {
  if(n<=num_bits)
    FATAL("ERROR:new vector shorter") 
  integer i;
  BitVector temp(n);
  temp.num_bits = n;
  temp.num_words = (n-1)/BIT_VEC_WORD_SIZE + 1;
  for (i=0; i<num_words; i++) {
    temp.bit_vector[i] = bit_vector[i];
  }
  for (i=num_bits; i<temp.num_bits; i++) {
    temp.Clear(i);
  }
  return(temp);
}

BitVector BitVector::Expand1(integer n) {
  if(n<=num_bits) 
    FATAL("ERROR:new vector shorter")
  integer i;
  BitVector temp(n);
  temp.num_bits = n;
  temp.num_words = (n-1)/BIT_VEC_WORD_SIZE + 1;
  for (i=0; i<num_words; i++) {
    temp.bit_vector[i] = bit_vector[i];
  }
  for (i=num_bits; i<temp.num_bits; i++) {
    temp.Set(i);
  }
  return(temp);
}

//     
// This function compares two BitVectors. Bit 0
// is treated as the MSB. Note that a Hash function
// (defined in 'bv.h') just returnes the most
// significant word (bit_vector[0]) of the stored
// BitVector.
//
integer BitVector::Compare(Comparable *p) {
  if(p->isa() != BIT_VECTOR) 
    FATAL("not a BIT_VECTOR")
  BitVector *x = (BitVector *)p; 
  if(num_bits!=this->num_bits) 
    FATAL("operand sizes do not match")
  integer i;
  for(i=0; i<num_bits; i++) {
    if((*x)[i] > (*this)[i]) return(-1); 
      else { if((*x)[i] < (*this)[i]) return(1); } 
  } 
  return(0);
}

//
// output a packed hex number representing the bit_vector_word
//

char bit_vector_hex_table[] = "0123456789ABCDEF***";

void
BitVector::PrintWordHex(ostream& out, bit_vector_word x) {

  integer i;
  bit_vector_word m;
  m = BIT_VEC_TOP_ONE 
    | (BIT_VEC_TOP_ONE >> 1)
    | (BIT_VEC_TOP_ONE >> 2)
    | (BIT_VEC_TOP_ONE >> 3);

  for (i=0; i<BIT_VEC_WORD_SIZE/4; ++i) {
    out << bit_vector_hex_table[(m&x) >> (BIT_VEC_WORD_SIZE - (i*4) - 4)];
    m >>= 4;
  }
}


