////////////////////////////////////////////////////////////////////////////
//
//  dg.C : (a directed graph package)  
//  Andrw Seawright 
//  1/30/91 modified 6/26/91, 7/28/91
//
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include "dg.h"

////////////////////////////////////////////////////////////////////////////

dg::dg() {   // constructor 
  data = 0;
  printNode = printEdge = FALSE;
}

dg::~dg() {  // destructor
  List_ForEach_Object(nodes, n) {  // delete all nodes
    delete n;
  } List_End_ForEach;
  List_ForEach_Object(edges, e) {  // delete all edges
    delete e;
  } List_End_ForEach;
}

dg_node * 
dg::AddNode(dg_node *p) {  // Add a node to the graph
  
  p->node_num = nodes.Count();  // number the node
  nodes.Add(*p);                // add node to graph's node list
  return p;                
}

dg_node *
dg::AppendNode(dg_node *p) {  // Add a node to the graph
 
  p->node_num = nodes.Count();  // number the node
  nodes.Append(*p);                // add node to graph's node list
  return p;
}

dg_edge * 
dg::AddEdge(dg_node& node1, dg_node& node2, dg_edge *e) {  
  e->from_node = &node1;
  e->to_node   = &node2;
  edges.Add(*e);             // add edge to graph's edge list
  node1.out_edges.Add(*e);   // add edge to from node's fan-out list
  node2.in_edges.Add(*e);    // add edge to to node's fan-in list
  return e;
}

dg_edge * 
dg::AddEdge(dg_node *node1, dg_node *node2, dg_edge *e) {  
  e->from_node = node1;
  e->to_node   = node2;
  edges.Add(*e);             // add edge to graph's edge list
  node1->out_edges.Add(*e);   // add edge to from node's fan-out list
  node2->in_edges.Add(*e);    // add edge to to node's fan-in list
  return e;
}

////
//
// remove all directed edges from node a to node b;
//
//  note: this is not the fastest implementation

boolean
dg::RemoveEdges(dg_node *a, dg_node *b) {
  List to_delete;

  DG_ForEach_Out_Edge(*a, oe) {
    if (b == oe->ToNode()) {
      to_delete.Add( *oe);
    }
  } DG_End_ForEach;

  if (to_delete.Count() == 0)
    return FALSE;

  List_ForEach_Type_Remove( to_delete, dg_edge, e) {
    a->out_edges.RemoveAllof((Object *)e);
    b->in_edges.RemoveAllof((Object *)e);
    edges.RemoveAllof((Object *)e);
    delete e;
  } List_End_ForEach;
  return TRUE;
}

  //
 // Remove a node from the graph;
//

void
dg::RemoveNode(dg_node *node) {
  dg_node *dgn;
  
  while (node->out_edges.Count()) {
    dgn = ((dg_edge *) node->out_edges.Top())->to_node;
    RemoveEdges(dgn, node);
  }

  while (node->in_edges.Count()) {
    dgn = ((dg_edge *) node->in_edges.Top())->from_node;
    RemoveEdges(dgn, node);
  }

  nodes.RemoveAllof((Object *)node);
  delete node;
  Renumber();
}

void
dg_node::RemoveInEdge(dg_edge *e) {
   if (!in_edges.RemoveAllof(e)) {
      cerr << "RemoveInEdge: edge not found\n";
   }
}

void
dg_node::RemoveOutEdge(dg_edge *e) {
   if (!out_edges.RemoveAllof(e)) {
      cerr << "RemoveOutEdge: edge not found\n";
   }
}


//
// This routine adjusts the graph such that the source node is the
// first node in the graph. The graph is not changed otherwise.
//

void
dg::AdjustSourceFirst() {
  if (source) {
    nodes.RemoveAllof((Object *)source);
    nodes.Add(*source);
    Renumber();
  }
}

//
// This routine prunes all nodes not connected to the node of interest.
//

void
dg::PruneUnconnectedNodes(dg_node *node) {

  List remove;

  DG_ForEach_Node(*this, n) {
    n->Touch(FALSE);
  } DG_End_ForEach;

  MarkConnectedNodes(node);

  DG_ForEach_Node(*this, n) {
    if (!n->Touched())
      remove.Add(*n);
  } DG_End_ForEach;

  List_ForEach_Type(remove, dg_node, n) {
    RemoveNode(n);
  } List_End_ForEach;
  
  AdjustSourceFirst();
}

//
// This recursive routine marks all nodes following to the node of interest
// via (D)epth (F)irst (S)earch.
// note : assumes that all nodes have had their touch flags reset.
//

void
dg::MarkConnectedNodes(dg_node *node) {

  if ( !node->Touched()) {
    node->Touch(TRUE);
    DG_ForEach_Out_Node(*node, n) {
      if (!n->Touched()) {
	MarkConnectedNodes(n);
      }
    } DG_End_ForEach;
  }
}

//
// This recursive routine marks all preceeding nodes connected to the node of interest
// via (D)epth (F)irst (S)earch.
// note : assumes that all nodes have had their touch flags reset.
//

void
dg::BackMarkConnectedNodes(dg_node *node) {

  if ( !node->Touched()) {
    node->Touch(TRUE);
    DG_ForEach_In_Node(*node, n) {
      if (!n->Touched()) {
	BackMarkConnectedNodes(n);
      }
    } DG_End_ForEach;
  }
}

boolean
dg::isAcyclic() {

  // speeds it up
  List *sorted = TopologicalSortNodes();

  TouchNodes( 0 );
  List_ForEach_Type( sorted, dg_node, node ) {
    if( node->Touched() == 2 ) continue;
    if( !(isAcyclic( node )) ) {
      delete sorted;
      return FALSE;
      }
    }List_End_ForEach
  
  delete sorted;
  return TRUE;
  }

boolean
dg::isAcyclic( dg_node *node ) {
  node->Touch( 1 );

  DG_ForEach_Out_Node( *node, n ) {
    if( n->Touched() == 1 ) return FALSE;
    if( !(isAcyclic( n )) ) return FALSE;
    }DG_End_ForEach;

  node->Touch( 2 );		// don't start with this node
  return TRUE;
  }

List*
dg::TopologicalSortNodes() {
  TouchNodes(FALSE);
  List *order = new List;

  DG_ForEach_Node(*this,node) {
    TopSort( node, order );
    }DG_End_ForEach

  return order;
  }


List*
dg::TopologicalSortData() {
  List *ord = TopologicalSortNodes();

  List *ord_data = new List;

  List_ForEach_Type( ord, dg_node, node ) {
    ord_data->Add( node->Get() );
    }List_End_ForEach

  ord_data->Reverse();

  delete ord;
  return ord_data;
  }

//
// private topological sort routine
//

void
dg::TopSort( dg_node *node, List *o ) {
  if( node->Touched() ) return;

  node->Touch(TRUE);

  DG_ForEach_Out_Node( *node, n ) {
    TopSort(n, o);
    }DG_End_ForEach
  
  o->Add( node );
  }

//
// Simple input routine
//

void
dg::Read( istream& in) {
  
  int i, N, E, n1, n2;
  
  in >> N;  in >> E;
  dg_node **tmp_nodes = new dg_node*[N];
  
  for (i=0; i<N; ++i)
    tmp_nodes[i] = AddNode(new dg_node);
  
  for (i=0; i<E; ++i) {
    in >> n1; in >> n2;;
    AddEdge( tmp_nodes[n1], tmp_nodes[n2], new dg_edge);
  }
  Putsource(tmp_nodes[0]);
  delete tmp_nodes;
}

//
// Print routine
//

void
dg::Print( ostream& out) {
  out << "\ndg graph: ";
  if (data != 0)
    data->Print( out);
  else
    out << "\n";
  out << "number of nodes: " << nodes.Count() << "\n";
  out << "number of edges: " << edges.Count() << "\n";
  
  DG_ForEach_Node(*this, n) {
    out << "[" << n->node_num << "]";
    n->PrintII( out, printNode);  
    out << endl;
    
    out << " to: ";
    DG_ForEach_Out_Edge(*n, oe) {
      dg_node *on = oe->ToNode();
      out << on->node_num << ":";
      oe->PrintII( out, printEdge);
    } DG_End_ForEach; 
    out << endl;
    
    out << " from: ";
    DG_ForEach_In_Edge(*n, ie) {
      dg_node *in = ie->FromNode();
      out << in->node_num << ":";
      ie->PrintII( out, printEdge);
    } DG_End_ForEach; 
    out << "\n";

  } DG_End_ForEach;
}

//
// functions to add or transfer one graph to another
//

void
dg::AddGraph(dg *graph2add) {
  
  nodes.Add(&(graph2add->nodes));
  edges.Add(&(graph2add->edges));
  Renumber();
}

void
dg::TransferGraph(dg *graph2xfer) {
  
  nodes.Transfer(&(graph2xfer->nodes));
  edges.Transfer(&(graph2xfer->edges)); 
  Renumber();
}

//
// Reindex nodes in graph, used if a graph is added.
//

void
dg::Renumber() {
  int i = 0;
  DG_ForEach_Node(*this, node) {
    node->node_num = i++;
  } DG_End_ForEach;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

void
dg_node::Print(ostream &out) {
  out << "dg_node: " << node_num << " ";
}

void
dg_node::PrintII(ostream &out, boolean printData) {
  out << node_num << " ";
  if ( printData) {
    if (data != 0)
      data->Print(out);
  }
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

void
dg_edge::Print( ostream& out) {
  out << "dg_edge: ";
}

void
dg_edge::PrintII( ostream& out, boolean printData) {
  if (printData)
    if (data != 0)
      data->Print( out);
    else
      out << " ";
}

////////////////////////////////////////////////////////////////////////////

