////////////////////////////////////////////////////////////////////////////
//
//  hgraph.C : (a directed graph package)  
//  Chuck Monahan  
//  borrowed heavily from dg.C by Andrew Seawright
//
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include "hgraph.h"
#include "hasher.h"

////////////////////////////////////////////////////////////////////////////

Hgraph::Hgraph() {   // constructor 
  data = NULL;
  printNode = printEdge = FALSE;
}

Hgraph::~Hgraph() {  // destructor
  List_ForEach_Object(nodes, n) {  // delete all nodes
    delete n;
  } List_End_ForEach;
  List_ForEach_Object(edges, e) {  // delete all edges
    delete e;
  } List_End_ForEach;
}

HgraphNode* 
Hgraph::NewNode( boolean append, Object* o) {  // Add a node to the graph
  HgraphNode *p = new HgraphNode( o, nodes.Count());
  if ( append)                   // add node to graph's node list
    nodes.Append( p);
  else
    nodes.Add( p);
  return p;                
}

HgraphEdge* 
Hgraph::NewEdge( boolean append, Comparable* o) {  // Add a node to the graph
  HgraphEdge *p = new HgraphEdge( o);
  if ( append)                   // add node to graph's node list
    edges.Append( p);
  else
    edges.Add( p);
  return p;                
}

void
Hgraph::Connect(HgraphNode* head, HgraphNode* tail, HgraphEdge *e) {  
  e->from_nodes.Add( head);
  e->to_nodes.Add( tail);

  head->out_edges.Add( e);   // add edge to from node's fan-out list
  tail->in_edges.Add( e);    // add edge to to node's fan-in list
}

void
Hgraph::ConnectToNet(HgraphNode* head, HgraphEdge *e) {  
  e->from_nodes.Add( head);
  head->out_edges.Add( e);   // add edge to from node's fan-out list
}

void
Hgraph::ConnectFromNet(HgraphNode* tail, HgraphEdge *e) {  
  e->to_nodes.Add( tail);
  tail->in_edges.Add( e);    // add edge to to node's fan-in list
}

void
Hgraph::ConnectUnique(HgraphNode* head, HgraphNode* tail, HgraphEdge *e) {  
  e->from_nodes.AppendUnique( head);
  e->to_nodes.AppendUnique( tail);

  head->out_edges.AppendUnique( e);   // add edge to from node's fan-out list
  tail->in_edges.AppendUnique( e);    // add edge to to node's fan-in list
}

void
Hgraph::ConnectUniqueToNet(HgraphNode* head, HgraphEdge *e) {  
  e->from_nodes.AppendUnique( head);
  head->out_edges.AppendUnique( e);   // add edge to from node's fan-out list
}

void
Hgraph::ConnectUniqueFromNet(HgraphNode* tail, HgraphEdge *e) {  
  e->to_nodes.AppendUnique( tail);
  tail->in_edges.AppendUnique( e);    // add edge to to node's fan-in list
}

boolean
Hgraph::Disconnect(HgraphNode *head, HgraphNode *tail, HgraphEdge *e) {  
  if ( e->from_nodes.Member( head) && e->to_nodes.Member( tail)) {
    e->from_nodes.Remove( head);
    e->to_nodes.Remove( tail);

    head->out_edges.Remove( e);
    tail->in_edges.Remove( e);
    return TRUE;
  } else
    return FALSE;
}

  //
 // Remove a node from the graph;
//

boolean
Hgraph::RemoveEdge( HgraphEdge* e) {

  if (!edges.Member((Object*) e))
    return FALSE;

  HgraphNode *n;

  while ( e->from_nodes.Count()) {
    n = (HgraphNode*) e->from_nodes.Pop();
    n->out_edges.Remove( e);
  }

  while ( e->to_nodes.Count()) {
    n = (HgraphNode*) e->to_nodes.Pop();
    n->in_edges.Remove( e);
  }

  edges.Remove(e);
  delete e;
  return TRUE;  
}

  //
 // Remove a node from the graph;
//

boolean
Hgraph::RemoveNode( HgraphNode* n) {

  if (!nodes.Member( n))
    return FALSE;

  HgraphEdge *e;

  while ( n->out_edges.Count()) {
    e = (HgraphEdge *) n->out_edges.Pop();
    e->from_nodes.Remove( n);
  }

  while ( n->in_edges.Count()) {
    e = (HgraphEdge *) n->in_edges.Pop();
    e->to_nodes.Remove( n);
  }

  nodes.Remove(n);
  delete n;
  return TRUE;  
}

  //
 // Remove a node from the graph;
//

boolean
Hgraph::BypassNode( HgraphNode* n) {

 if (!nodes.Member( n))
    return FALSE;

  HgraphEdge *e;

  while ( n->in_edges.Count()) {
    e = (HgraphEdge *) n->in_edges.Pop();
    e->to_nodes.Remove( n);
    
    // Each node pointing at n, should point to n's children
    Hgraph_ForEach_Out_Edge( n, out) {
      Hgraph_ForEach_In_Node( e, src) {
	src->out_edges.AppendUnique( out);
	out->from_nodes.AppendUnique( src);
      } Hgraph_End_ForEach;
    } Hgraph_End_ForEach;
  }

  while ( n->out_edges.Count()) {
    e = (HgraphEdge *) n->out_edges.Pop();
    e->from_nodes.Remove( n);
  }

  nodes.Remove(n);
  delete n;
  return TRUE;  
}

void
Hgraph::RemoveFloatingEdges( ) {
  List deletion_list;

  Hgraph_ForEach_Edge( this, e) {
    if ( e->to_nodes.Count() == 0 || e->from_nodes.Count() == 0)
      deletion_list.Add( e);
  } Hgraph_End_ForEach;

  List_ForEach_Type( deletion_list, HgraphEdge, e) {
    RemoveEdge( e);
  } List_End_ForEach;
}

void
Hgraph::RemoveFloatingNodes( ) {
  List deletion_list;

  Hgraph_ForEach_Node( this, n) {
    if ( n->in_edges.Count() == 0 && n->out_edges.Count() == 0)
      deletion_list.Add( n);
  } Hgraph_End_ForEach;

  List_ForEach_Type( deletion_list, HgraphNode, n) {
    RemoveNode( n);
  } List_End_ForEach;
}

//
// This routine adjusts the graph such that the source node is the
// first node in the graph. The graph is not changed otherwise.
//

void
Hgraph::AdjustSourceFirst() {
  if (source) {
    nodes.RemoveAllof((Object *)source);
    nodes.Add(*source);
    Renumber();
  }
}

//
// This routine prunes all nodes not connected to the node of interest.
//

void
Hgraph::PruneUnconnectedNodes(HgraphNode *node) {

  List remove;

  Hgraph_ForEach_Node( this, n) {
    n->Touch(FALSE);
  } Hgraph_End_ForEach;

  MarkConnectedNodes(node);

  Hgraph_ForEach_Node( this, n) {
    if (!n->Touched())
      remove.Add(*n);
  } Hgraph_End_ForEach;

  List_ForEach_Type(remove, HgraphNode, n) {
    RemoveNode(n);
  } List_End_ForEach;
  
  AdjustSourceFirst();
}

void
Hgraph::OrderNodeList() {
  TouchNodes( TRUE);
  List old;
  old.Transfer( &nodes);

  //identify source nodes
  List_ForEach_Type_Remove( old, HgraphNode, node) {
    if ( node->SourceNode()) {
      node->ListDependants( &nodes);
    }
  } List_End_ForEach;
  
}

void
Hgraph::ReverseOrderNodeList() {
  TouchNodes( TRUE);
  List old;
  old.Transfer( &nodes);

  //identify source nodes
  List_ForEach_Type_Remove( old, HgraphNode, node) {
    if ( node->SinkNode()) {
      node->ListDependancies( &nodes);
    }
  } List_End_ForEach;
  
}

//
// This recursive routine marks all nodes following to the node of interest
// via (D)epth (F)irst (S)earch.
// note : assumes that all nodes have had their touch flags reset.
//

void
Hgraph::MarkConnectedNodes(HgraphNode *node) {

  if ( !node->Touched()) {
    node->Touch(TRUE);
    Hgraph_ForEach_Out_Edge( node, e) {
      Hgraph_ForEach_Out_Node( e, n) {
	if (!n->Touched()) {
	  MarkConnectedNodes(n);
	}
      } Hgraph_End_ForEach;
    } Hgraph_End_ForEach;
  }
}

//
// This recursive routine marks all preceeding nodes connected to the node of interest
// via (D)epth (F)irst (S)earch.
// note : assumes that all nodes have had their touch flags reset.
//

void
Hgraph::BackMarkConnectedNodes(HgraphNode *node) {

  if ( !node->Touched()) {
    node->Touch(TRUE);
    Hgraph_ForEach_In_Edge( node, e) {
      Hgraph_ForEach_In_Node( e, n) {
	if (!n->Touched()) {
	  BackMarkConnectedNodes(n);
	}
      } Hgraph_End_ForEach;
    } Hgraph_End_ForEach;
  }
}

//
// Print routine
//

void
Hgraph::Print( ostream& out) {
  out << "\nHgraph graph: ";
  if (data != 0)
    data->Print( out);
  else
    out << "\n";
  out << "number of nodes: " << nodes.Count() << "\n";
  out << "number of edges: " << edges.Count() << "\n";
  
  Hgraph_ForEach_Node( this, n) {
    n->PrintII( out, printNode);  
    out << endl;
    
    out << " to: ";
    Hgraph_ForEach_Out_Edge( n, oe) {
      Hgraph_ForEach_Out_Node( oe, on) {
	out << on->node_num << ":";
      } Hgraph_End_ForEach; 
      oe->PrintII( out, printEdge);
      out << ";";
    } Hgraph_End_ForEach; 
    out << endl;
    
    out << " from: ";
    Hgraph_ForEach_In_Edge( n, ie) {
      Hgraph_ForEach_In_Node( ie, in) {
	out << in->node_num << ":";
      } Hgraph_End_ForEach; 
      ie->PrintII( out, printEdge);
      out << ";";
    } Hgraph_End_ForEach; 
    out << "\n";

  } Hgraph_End_ForEach;
}

//
// functions to add or transfer one graph to another
//

void
Hgraph::Addgraph(Hgraph *graph2add) {
  
  nodes.Add(&(graph2add->nodes));
  edges.Add(&(graph2add->edges));
  Renumber();
}

void
Hgraph::TransferGraph(Hgraph *graph2xfer) {
  
  nodes.Transfer(&(graph2xfer->nodes));
  edges.Transfer(&(graph2xfer->edges)); 
  Renumber();
}

//
// Reindex nodes in graph, used if a graph is added.
//

void
Hgraph::Renumber() {
  int i = 0;
  Hgraph_ForEach_Node( this, node) {
    node->node_num = i++;
  } Hgraph_End_ForEach;
}

//
// new node get all of the edges of the old node
//

void
Hgraph::ReplaceNode( HgraphNode* oldn, HgraphNode* newn) {
  List temp;

  List_ForEach_Type_Remove(oldn->out_edges, HgraphEdge, oedge) {
    temp.Transfer( &oedge->from_nodes);
    List_ForEach_Type_Remove(temp, HgraphNode, fn) {
      if ( fn == oldn)
	oedge->from_nodes.Add( newn);
      else
	oedge->from_nodes.Add( fn);
    } List_End_ForEach;
    oedge->from_nodes.Reverse();
    newn->out_edges.Add( oedge);
  } List_End_ForEach

  List_ForEach_Type_Remove(oldn->in_edges, HgraphEdge, iedge) {
    temp.Transfer( &iedge->to_nodes);
    List_ForEach_Type_Remove(temp, HgraphNode, tn) {
      if ( tn == oldn)
	iedge->to_nodes.Add( newn);
      else
	iedge->to_nodes.Add( tn);
    } List_End_ForEach
    iedge->to_nodes.Reverse();
    newn->in_edges.Add( iedge);
  } List_End_ForEach
}

//
// new edge connects to all of the nodes of the old edge
//

void
Hgraph::ReplaceEdge( HgraphEdge* olde, HgraphEdge* newe) {
  List temp;

  List_ForEach_Type_Remove(olde->from_nodes, HgraphNode, fnode) {
    temp.Transfer( &fnode->out_edges);
    List_ForEach_Type_Remove(temp, HgraphEdge, oe) {
      if ( oe == olde)
	fnode->out_edges.Add( newe);
      else
	fnode->out_edges.Add( oe);
    } List_End_ForEach
    newe->from_nodes.Add( fnode);
  } List_End_ForEach

  List_ForEach_Type_Remove(olde->to_nodes, HgraphNode, tnode) {
    temp.Transfer( &tnode->in_edges);
    List_ForEach_Type_Remove(temp, HgraphEdge, ie) {
      if ( ie == olde)
	tnode->in_edges.Add( newe);
      else
	tnode->in_edges.Add( ie);
    } List_End_ForEach
    newe->to_nodes.Add( tnode);
  } List_End_ForEach
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

// Duplicates the node, connecting to the same edges than n
// connects to
HgraphNode::HgraphNode( HgraphNode &n ) {
  node_num = -1;
  data = n.data;

  List_ForEach_Type( n.out_edges, HgraphEdge, edge ) {
    edge->from_nodes.Add( this );
    out_edges.Add( edge );
    }List_End_ForEach

  List_ForEach_Type( n.in_edges, HgraphEdge, edge ) {
    edge->to_nodes.Add( this );
    in_edges.Add( edge );
    }List_End_ForEach
  }

void
HgraphNode::ListDependants( List* nodes) {
  Touch( FALSE);
  List_ForEach_Type( out_edges, HgraphEdge, edge) {
    List_ForEach_Type( edge->to_nodes, HgraphNode, node) {
      if ( node->Touched())
        node->ListDependants( nodes);
    } List_End_ForEach;
  } List_End_ForEach;

  nodes->Add( this); 
}

void
HgraphNode::ListDependancies( List* nodes) {
  Touch( FALSE);
  List_ForEach_Type( in_edges, HgraphEdge, edge) {
    List_ForEach_Type( edge->from_nodes, HgraphNode, node) {
      if ( node->Touched())
        node->ListDependancies( nodes);
    } List_End_ForEach;
  } List_End_ForEach;

  nodes->Add( this); 
}

void
HgraphNode::Print(ostream &out) {
  out << "HgraphNode: " << node_num << " ";
}

void
HgraphNode::PrintII(ostream &out, boolean printData) {
  out << node_num << ": ";
  if ( printData) {
    if (data != 0)
      data->Print(out);
  }
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

void
HgraphEdge::Print( ostream& out) {
  out << "HgraphEdge: ";
}

void
HgraphEdge::PrintII( ostream& out, boolean printData) {
  if (printData)
    if (data != 0)
      data->Print( out);
    else
      out << " ";
}

unsigned 
HgraphEdge::Hash(unsigned h) {  
  integer ret = (data == NULL)? 0:data->Hash( h);
  ret = ret ^ (h << 11);
  integer i = 0;
  Hgraph_ForEach_In_Node( this, inode) {
    ret = ret ^ ((integer)inode << i++);
  }Hgraph_End_ForEach;
  return ret;
}

integer
HgraphEdge::Compare(Comparable* c) {  
  ASSERT (c->isa() == HGRAPH_EDGE, "");
  HgraphEdge* edge = (HgraphEdge*) c;
  integer     dif  = Fanin() - edge->Fanin();
  if ( dif != 0) return dif;

  if ( data != NULL && edge->data != NULL) {
    integer dif = data->Compare( edge->data);
    if ( dif != 0) return dif;
  } else {
    dif = data - edge->data;
    if ( dif != 0) return dif;
  }

  List_iterator li((edge)->from_nodes);
  Hgraph_ForEach_In_Node( this, inode) {
    dif = inode - ((HgraphNode*)li.item());
    if ( dif != 0) return dif;
    ++li;
  }Hgraph_End_ForEach;

  return 0;
}

////////////////////////////////////////////////////////////////////////////

