////////////////////////////////////////////////////////////////////////////
//
//  memhash.C : hash tree for keeping track of allocated memory
//  Andrew Crews
//  
////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include "memhash.h"

int MemHashTree::freecount = 0;
MemHashNode *MemHashTree::freenodes = NULL;
MemHashNode *MemHashTree::freednodes = NULL;
MemHashNode::MemHashNode(int hv, int k, int o) {   // HashNode constructor

  right   =  left  =  0;
  hashval =  hv;
  key     =  k;
  data    =  o;
}

void*
MemHashNode::operator new( size_t n ) { return MemHashTree::AllocNode(); }

void
MemHashNode::operator delete( void *p ) { MemHashTree::FreeNode( p ); }

////////////////////////////////////////////////////////////////////////////
// MemHashTree constructor

MemHashTree::MemHashTree( ) { 
  
//  freesuc = freefail = allocsuc = allocfail = collisions = count = bytes = 0;
//  root        =  NULL;
}

void
MemHashTree::Clear(MemHashNode *p) {    // recursive destruction of tree 

  if (p->Right() != NULL) {
     Clear(p->Right());
     p->Right(NULL);
  }
  if (p->Left() != NULL) {
     Clear(p->Left());
     p->Left(NULL);
  }

  delete p;
}
////////////////////////////////////////////////////////////////////////////

boolean
MemHashTree::Put(int o, int key)
{
  MemHashNode *q = NULL, *p = root;
  int hashval = GetHashVal( key );

  for (;;) {
    if (p == NULL) { 
      p = new MemHashNode(hashval, key, o);         // build a new node
      if (root == NULL)
	root = p;
      else {
	if (q->Hashval() > hashval)
	  q->Right(p);
	else
	  q->Left(p);
      }
      bytes+=o;
      count++;
      allocsuc++;
      return TRUE;
    } 
    else if (p->Hashval() == hashval) { 
      if ( key == p->Key() )  {
	allocfail++;
	return FALSE;                    // key already in tree
	}
    }
    q = p;
    
    if (p->Hashval() > hashval)  
      p=p->Right();
    else
      p=p->Left();
  }
}

////////////////////////////////////////////////////////////////////////////

void
MemHashTree::ForcePut(int o, int key) {

  MemHashNode *q = NULL, *p = root;
  int hashval = GetHashVal( key );

  for (;;) {
    if (p == NULL) { 
      p = new MemHashNode(hashval, key, o);         // build a new node
      if (root == NULL)
	root = p;
      else {
	if (q->Hashval() > hashval)
	  q->Right(p);
	else
	  q->Left(p);
      }
      count++;
      bytes+=o;
      return;
    } 
    else if (p->Hashval() == hashval) { 
      if ( key == p->Key() ) {

	p->Data(o);                    // Jam it! 

        collisions++;
	bytes+=o;
	return;                         
      } 
    }
    q = p;
    if (p->Hashval() > hashval) 
      p=p->Right();
    else
      p=p->Left();
  }
}

////////////////////////////////////////////////////////////////////////////

int
MemHashTree::Get(int key)
{
  int hashval = GetHashVal( key );
  MemHashNode *p = root;

  for (;;) {
    if (p == NULL) 
      return 0;
    
    else if (p->Hashval() == hashval) {   // hashval match
       if ( key == p->Key() ) {   // match!
	 return p->Data();
	 }
     }
    p = (p->Hashval() > hashval) ? p->Right() : p->Left();
  }
}

////////////////////////////////////////////////////////////////////////////
//
//      o               o      |        o               o
//     / \             / \     |       / \             / \
//    p   x0          q   x0   |      p   x0          q   x0
//   / \             / \       |     / \        =>   / \
//  x1 x2           x1  x2     |    q   x1          x2  x1
//   \               \         |   / \
//   ...        =>   ...       |  x2 NULL
//     \               \       |________________________________________
//      r               r      |         o              o
//     / \             / \     |        / \    =>      / \
//    x3  q           x3  x4   |       p   x0         x1  x0
//       / \                   |      / \
//      x4 NULL                     NULL x1
//
// We are removing node p in the picture above, and replacing it with the
// node in position q--the first node with no right pointer, which is just
// less than the node p.
// o, p, q, and r are as shown above. x#'s represent any tree from that point
// down (anything from a single node to a full tree)
//
boolean
MemHashTree::Remove( int key ) {
  int hashval = GetHashVal(key);
  MemHashNode *p = root;
  MemHashNode *o = NULL;

  for (;;) {
    if( p == NULL ) {
      freefail++;
      return FALSE;
      }

    else if (p->Key() == key) { // hashval match

      // now search for a node with no "right" pointer to put in its place
      MemHashNode *r = NULL;
      MemHashNode *q= (MemHashNode *)p->Left() ;
      for(  ;
                 q != NULL && q->Right() ;
                 r = q, q = (MemHashNode *)q->Right() );

      if( r )            // => q exists, as well
        {
        r->Right(q->Left());
        q->Left(p->Left());
        }
      if( q )
        q->Right(p->Right());
      else                      // case 3, set, but ->Right() can still be NULL
        q = (MemHashNode *)p->Right();

      // Now q is the node which is replacing p, or is NULL, so fix the parent (o)
      if( o )
        ( o->Hashval() > hashval ) ? o->Right(q) : o->Left(q);
      else
	root = q;
      count--;
      bytes-=p->Data();
      delete p;
      freesuc++;
      return TRUE;
      }
    o = p;
    if (p->Hashval() > hashval)
      p=(MemHashNode *)p->Right();
    else
      p=(MemHashNode *)p->Left();
    } // end for

  }

////////////////////////////////////////////////////////////////////////////

void
MemHashTree::PrintStats( ostream& out)
{ 
  int max_depth = 0;
  out << "Memory verification stats:\n";
  out << bytes << " bytes allocated in " << count << " blocks\n";
  out << allocsuc << " successful calls to malloc, "<<allocfail<<" errors\n";
  out << freesuc << " successful calls to free, " << freefail << " errors\n";
  out << "----------------------------------------------\n";
  out << "collisions=  " << collisions << endl;

}

////////////////////////////////////////////////////////////////////////////

void*
MemHashTree::AllocNode() {
  MemHashNode *p;
  if( freecount ) {
    p = freenodes;
    freenodes++;
    freecount--;
    return (void*)p;
    }
  else if( freednodes ) {
    p = MemHashTree::freednodes;
    freednodes = *(MemHashNode**)freednodes;
    return (void*)p;
    }
  else {
    freenodes = (MemHashNode*)::malloc(65536*sizeof(MemHashNode));
    freecount = 65535;
    p = freenodes;
    freenodes++;
    return (void*)p;
    }
  }

void
MemHashTree::FreeNode( void *n ) {
  *(MemHashNode**)n = freednodes;
  freednodes = (MemHashNode*)n;
  }
