//
// rbtree.C : red black tree with hashing  
// Tom Weinstein
//

#include "rbtree.h"

/**********************************************************************/

RBNode::operator == (RBNode n)
{
  return (hashval == n.hashval && !strcmp(key, n.key));
}

RBNode::operator != (RBNode n)
{
  return (hashval != n.hashval || strcmp(key, n.key));
}

RBNode::operator < (RBNode n)
{
  return (hashval < n.hashval ||
	  ((hashval == n.hashval) && strcmp(key, n.key) < 0));
}

RBNode::operator <= (RBNode n)
{
  return (hashval < n.hashval ||
	  ((hashval == n.hashval) && strcmp(key, n.key) <= 0));
}

RBNode::operator > (RBNode n)
{
  return (hashval > n.hashval ||
	  ((hashval == n.hashval) && strcmp(key, n.key) > 0));
}

RBNode::operator >= (RBNode n)
{
  return (hashval > n.hashval ||
	  ((hashval == n.hashval) && strcmp(key, n.key) >= 0));
}

/**********************************************************************/

RBNode *RBNode::rbnull = NULL;

RBNode::RBNode(int HashVal, char *Key, Object *Data)
{
  right = left = RBNode::rbnull;
  color = RED;
  hashval = HashVal;
  key = Key;
  data = Data;
}

RBNode::RBNode(RBNode *orig, boolean copy)
{
  color = orig->color;
  key = orig->key;
  hashval = orig->hashval;

  data = copy ? orig->data->Copy() : orig->data;

  right = (orig->right == RBNode::rbnull) 
    ? RBNode::rbnull : new RBNode(orig->right, copy);
  left = (orig->left == RBNode::rbnull) 
    ? RBNode::rbnull : new RBNode(orig->right, copy);
}

void
RBTree::RBDestroy(RBNode *pp) {    // recursive destruction of tree 

  if (pp->right != RBNode::rbnull) {
     RBDestroy(pp->right);
     pp->right = NULL;
  }
  if (pp->left != RBNode::rbnull) {
     RBDestroy(pp->left);
     pp->left = NULL;
  }
  delete pp;
}

/**********************************************************************/

RBTree::RBTree()
{
  if(!RBNode::rbnull) {
    RBNode::rbnull = new RBNode(0, "", NULL);
    RBNode::rbnull->color = BLACK;
    RBNode::rbnull->right = RBNode::rbnull->left = RBNode::rbnull;
  }

  root = new RBNode(0, "", NULL);
  root->color = BLACK;
  count = 0;
}

RBTree::RBTree(RBTree *orig, boolean copy)
{
  root = (orig->root == RBNode::rbnull)
    ? RBNode::rbnull : new RBNode(orig->root, copy);
}


integer RBTree::hashf (char *s)
{  
  int h = 0;

  for ( ; *s != '\0'; ++s) {
    h += (int)*s << 11;
    h >>= 3;
    if(h&0x01)
      h ^= 0x6e80000;
  }
  return (integer) h&0x7fffffff;
}


boolean RBTree::Put(Object &o, char *key)
{
  integer hashval = hashf(key);
  RBNode *node = new RBNode(hashval, key, &o);

  gg = g = p = x = root;

  if(*node == *x) {
    delete node;
    return FALSE;
  }
  if(x->right->red() && x->left->red())
    split(node);

  while(x != RBNode::rbnull) {
    gg = g; g = p; p = x;
    x = (*node < *x) ? x->left : x->right;

    if(*node == *x) {
      delete node;
      return FALSE;
    }
    if(x->right->red() && x->left->red())
      split(node);
  }

  x = node;
  if (*node < *p)
    p->left = x;
  else
    p->right = x;

  split(node);

  count++;
  return TRUE;
}

void RBTree::ForcePut(Object &o, char *key)
{
  integer hashval = hashf(key);
  static RBNode node(0, NULL, NULL);

  node.hashval = hashval;
  node.key = key;

  gg = g = p = x = root;

  if(node == *x) {
    x->data = &o;
    return;
  }
  if(x->right->red() && x->left->red())
    split(&node);

  while(x != RBNode::rbnull) {
    gg = g; g = p; p = x;
    x = (node < *x) ? x->left : x->right;

    if(node == *x) {
      x->data = &o;
      return;
    }
    if(x->right->red() && x->left->red())
      split(&node);
  }

  x = new RBNode(hashval, key, &o);

  if (node < *p)
    p->left = x;
  else
    p->right = x;

  split(&node);
  count++;
}


Object* RBTree::Get(char *key)
{
  register integer hashval = hashf(key);
  register RBNode *xx = root->right;

  for (;;) {
    if (xx == RBNode::rbnull) 
      return 0;
    
    else if (xx->hashval == hashval) {   // hashval match
       if (strcmp(key, xx->key) == 0)   // match!
	 return xx->data;
       else 
	 xx = (strcmp(key, xx->key) < 0) ? xx->left : xx->right;
     } else
       xx = (hashval < xx->hashval) ? xx->left : xx->right;
  }
}


RBNode *RBTree::rotate(RBNode *y, RBNode *node)
{
  RBNode *c, *gc;

  c = (*node < *y) ? y->left: y->right;

  if (*node < *c) {
    gc = c->left;
    c->left = gc->right;
    gc->right = c;
  } else {
    gc = c->right;
    c->right = gc->left;
    gc->left = c;
  }

  if (*node < *y)
    y->left = gc;
  else
    y->right = gc;

  return gc;
}

void RBTree::split(RBNode *node)
{
  x->color = RED;
  x->left->color = BLACK;
  x->right->color = BLACK;

  if (p->red()) {
    g->color = RED;
    if (*node < *g != *node < *p) p = rotate(g, node);
    x = rotate(gg, node);
    x->color = BLACK;
  }
  root->right->color = BLACK;
}


void RBTree::Print(ostream& out)
{
  if (root ->right!= RBNode::rbnull) {
    out << "RB Tree : " << count << " item(s).\n";
    PrintNode(root->right, 0, out);
  } else
    out << "RB Tree : empty\n";
}

void RBTree::PrintNode(RBNode *pp, int depth, ostream& out)
{
  out << "Depth=" << depth << " Color=" << (pp->color ? "Red" : "Black")
    << " Key=" << pp->key << " ->";
  if (pp->data)
    pp->data->Print(out);

  if (pp->left != RBNode::rbnull) {
    cout << "left: ";
    PrintNode(pp->left, depth+1, out);
  }
  if (pp->right != RBNode::rbnull) {
    cout << "right: ";
    PrintNode(pp->right,  depth+1, out);
  }
}


/**********************************************************************/

RBIterator::RBIterator(RBTree& t) { // constructor
  if (t.root) {
    n = t.root;
    done = FALSE;
  } else {
    done = TRUE;
  }
}

RBIterator::~RBIterator() {     // destructor
}

void RBIterator::operator ++ () {    // iterator to walk through hash tree 

  if (done == TRUE)
    return;

    if (n->left != RBNode::rbnull) {  // try to go left
      stack.Add(*n);     // save current node
      n = n->left;
      return;
    } 

    while (n->right == RBNode::rbnull) {
      
      if (stack.Count() == 0) {
        done = TRUE;
        return;
      }
      n = (RBNode *) stack.Pop();
    }
   n = n->right;
   return;
}

