// rpn_bdd.C
// Tony Stornetta
//
// This small test program allows you to enter a boolean expression
// in reverse polish notation, to create a test BDD.
// A maximum of VARS variables can be used, and a stack size of
// STACKSIZE should provide plenty of space for a BDD that is to be
// viewed.  The expression shouldn't be longer than LINESIZE characters!
// Simple problems like under or overflowing yer stack are
// checked.
//
// The syntax:
//
// #	This symbol pops the previous function or variable off the stack.
// It is useful for entering the level order in which variables
// will be input (ie A#B#C#D#).  Otherwise, new variables are
// just assigned to the next level as they are encountered.
//
// @	This symbol does garbage collection, garbage collecting ALL
// extraneous functions AND input variables not used, to clean
// up the tree.  It does not pop anything off the stack
//
// ~	This is the logical not of the next variable.  It does nothing
// to the stack and assumes the next character will be an input
// variable.
//
// <	Pops two functions off the stack and cofactors the second by
// the first
//
// !	This pops one variable or function off the stack and negates
// it, pushing the result back onto the stack.
//
// &	This pops two variables or functions off the stack
// and performs a logical AND on them, pushing the result back
// onto the stack.
//
// |	This pops two variables or functions off the stack
// and performs a logical OR on them, pushing the result back
// onto the stack.
//
// ^	This pops two variables or functions off the stack
// and performs a logical XOR on them, pushing the result back
// onto the stack.
//
// %	This sets a flag such that the next letter becomes the top
// function on the stack.  The function is popped off the stack.
// Variables may not be used as functions, however, functions may
// be reassigned.  No result is pushed onto the stack.
//
// a-Z	Any letter serves as an input variable.  If it
// does not exist, it will create it as the next input variable.
// The variable gets pushed onto the stack.
//

#include <stdlib.h>
#include "bdd.h"

enum type
  {
  FUNCTION, VARIABLE
  };

#define STACKSIZE	100
#define VARS		100
#define LINESIZE	10240

int 
main(int argc, char *argv[])
  {

  BDDmanager     *bdd;
  char            line[LINESIZE];
  char           *ptr;
  int             i;

  bdd = new BDDmanager;

  cout << "Enter Boolean function, rpn: ";
  cin.getline(line, LINESIZE);

  // Parse boolean function
  ptr = line;

  char            variables[VARS];
  char		 *varnames[VARS];
  BDDnode        *nodes[VARS];
  enum type       types[VARS];
  BDDnode        *stack[STACKSIZE];
  int             stackPos = 0;
  int             varcnt = 0;
  int             neg = 0;
  enum type       Type = VARIABLE;
  int             levels = 0;

  while (*ptr != '\0')
    {
    switch (*ptr)
      {
      case '@':
	bdd->Keep(stack[stackPos - 1]);
	for (i = 0; i < varcnt; i++)
	  {
	  bdd->UnKeep(bdd->not(nodes[i]));
	  bdd->UnKeep(nodes[i]);
	  }
	bdd->GC();
	neg = 0;
	Type = VARIABLE;
	break;
      case '#':
	if (stackPos < 1)
	  {
	  cerr << "You're gonna crunch yer stack!" << endl;
	  exit(1);
	  }
	stackPos--;
	neg = 0;
	Type = VARIABLE;
	break;
      case '!':
	if (stackPos < 1)
	  {
	  cerr << "You're gonna crunch yer stack!" << endl;
	  exit(1);
	  }
	stack[stackPos - 1] = bdd->not(stack[stackPos - 1]);
	neg = 0;
	Type = VARIABLE;
	break;
      case '~':
	neg = 1;
	Type = VARIABLE;
	break;
      case '|':
	if (stackPos < 2)
	  {
	  cerr << "You're gonna crunch yer stack!" << endl;
	  exit(1);
	  }
	stackPos--;
	stack[stackPos - 1] = bdd->or(stack[stackPos - 1], stack[stackPos]);
	neg = 0;
	Type = VARIABLE;
	break;
      case '<':
	if (stackPos < 2)
	  {
	  cerr << "You're gonna crunch yer stack!" << endl;
	  exit(1);
	  }
	stackPos--;
	stack[stackPos - 1] = bdd->Cofactor(stack[stackPos - 1], stack[stackPos]);
	neg = 0;
	Type = VARIABLE;
	break;
      case '&':
	if (stackPos < 2)
	  {
	  cerr << "You're gonna crunch yer stack!" << endl;
	  exit(1);
	  }
	stackPos--;
	stack[stackPos - 1] = bdd->and(stack[stackPos - 1], stack[stackPos]);
	neg = 0;
	Type = VARIABLE;
	break;
      case '^':
	if (stackPos < 2)
	  {
	  cerr << "You're gonna crunch yer stack!" << endl;
	  exit(1);
	  }
	stackPos--;
	stack[stackPos - 1] = bdd->xor(stack[stackPos - 1], stack[stackPos]);
	neg = 0;
	Type = VARIABLE;
	break;
      case '%':
	Type = FUNCTION;
	neg = 0;
	break;
      case '1':
	if (stackPos == STACKSIZE)
	  {
	  cerr << "Ya blew yer stack!" << endl;
	  exit(1);
	  }
	if (neg == 0)
	  stack[stackPos++] = bdd->One();
	else
	  stack[stackPos++] = bdd->Zero();
	neg = 0;
	Type = VARIABLE;
	break;
      case '0':
	if (stackPos == STACKSIZE)
	  {
	  cerr << "Ya blew yer stack!" << endl;
	  exit(1);
	  }
	if (neg == 0)
	  stack[stackPos++] = bdd->Zero();
	else
	  stack[stackPos++] = bdd->One();
	neg = 0;
	Type = VARIABLE;
	break;
      case ' ':
	break;
      default:
	if ((*ptr >= 'A' && *ptr <= 'Z') || (*ptr >= 'a' && *ptr <= 'z'))
	  {
	  int             pos = 0;
	  // Search to see if variable already exists
	  for (i = 0; i < varcnt && variables[i] != *ptr; i++)
	    pos++;

	  if (pos == varcnt)
	    {
	    // Enter new variable/function in table
	    variables[varcnt] = *ptr;
	    varnames[varcnt]= new char[2];
	    varnames[varcnt][0] = *ptr;
	    varnames[varcnt][1] = '\0';
	    if (Type == FUNCTION)
	      {
	      // Assign and pop off function from stack!
	      nodes[varcnt] = stack[--stackPos];
	      types[varcnt] = FUNCTION;
	      }
	    else
	      {
	      // Assign new variable, pushing onto stack
	      nodes[varcnt] = bdd->Input(levels++);
	      types[varcnt] = VARIABLE;
	      if (stackPos == STACKSIZE)
		{
		cerr << "Ya blew yer stack!" << endl;
		exit(1);
		}
	      stack[stackPos++] = nodes[varcnt];
	      }
	    varcnt++;
	    if (varcnt == VARS)
	      {
	      cerr << "Too many vars!" << endl;
	      exit(1);
	      }
	    }
	  else
	    {
	    if (Type == FUNCTION)
	      {
	      if (types[i] == VARIABLE)
		{
		cerr << "Can't assign a function to a variable!" << endl;
		exit(1);
		}
	      // Assign and pop off function from stack!
	      nodes[i] = stack[--stackPos];
	      }
	    else
	      // Push existing variable on stack
	    if (stackPos == STACKSIZE)
	      {
	      cerr << "Ya blew yer stack!" << endl;
	      exit(1);
	      }
	    stack[stackPos++] = nodes[pos];
	    }
	  if (neg)
	    {
	    neg = 0;
	    stack[stackPos - 1] = bdd->not(stack[stackPos - 1]);
	    }
	  Type = VARIABLE;
	  }
	else
	  {
	  cerr << "Bogus character entered!" << endl;
	  exit(1);
	  }
	}
    if (ptr++ == line + LINESIZE)
      {
      cerr << "Your input line is just too dang long!" << endl;
      exit(1);
      }
    }

  if (stackPos != 1)
    cerr << "Warning: Your stack is not completely empty!" << endl;

  // cout << "sizeof (BDDnode) = " << sizeof (BDDnode) << endl;
  // cout << "sizeof (HasherNode) = " << sizeof (HasherNode) << endl;
  // bdd->FindRLE();
  if (argc > 1)			// debug mode

    bdd->XDump_nowait();
  bdd->XDump(stack[0], "Result", 0, varnames);
  bdd->PrintCubes(stack[0]);
  bdd->Print();

  return 0;
  }
