#ifndef BDDMGR_H
#define BDDMGR_H
#define cachprofile

////////////////////////////////////////////////////////////////////////////
//
// bdd.h : C++ BDD package based on [Brace, Rudell, and Bryant 27th DAC 1990]
//
// Andrew Sewawright
// 
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <strstream.h>
#include <limits.h>
#include "compable.h"
#include "hasher.h"
#include "hash.h"
#include "hashcach.h"
#include "tnode.h"
#include "intobj.h"
#include "list.h"
#include "indarray.h"
#include "hasharry.h"
#include "hashbase.h"

////////////////////////////////////////////////////////////////////////////

#define ITE_OP 1
#define COFACTOR_OP 2
#define RESTRICT_OP 3
// leave numbers 100 - 199 for BDDmanager (bdd.h)

#ifndef BDD_PRIME		// assign these in private code if desired

// Use the following BDD_PRIMES : they are 2^BDD_PRIME_BITS - 1, and all
// three are prime numbers--the % BDD_PRIME can be done much faster when
// we use a prime which is also 2^n - 1

#define COFACTOR_PRIME_BITS 17
#define COFACTOR_PRIME 131071

#define BDD_PRIME_BITS 19
#define BDD_PRIME 524287

#endif

// Forward class declarations:
class BDDHasherArray;
class BDDmanager;
class BDDmgr;
class BDDnode;
class CofactorHashKey;
class BLIFdata;
struct Tcl_Interp;

typedef BDDHasherArray BDDHasher;

////////////////////////////////////////////////////////////////////////////

class BDDnode : public Tnode {
 protected:
  // a BDDnode contains whatever Tnode contains plus:

  BDDnode* next;

  BDDnode() : Tnode() {;} // reference, mark inited by tnode
  BDDnode(integer v, BDDnode *a, BDDnode *b) { index = v; left = a; right = b; reference=0; }
 ~BDDnode() {;}

 public:
  integer  GetIndex()	{ return index; }
  integer  Level()	        { return GetIndex(); }   // backwards compatibility

  BDDnode* Then() { return (BDDnode *) left;}  // True side
  BDDnode* Else() { return (BDDnode *) right;} // False side
  BDDnode* Next() { return next;}              // For Hash list... do not use

  // Comparable virtual functions
  virtual unsigned Hash(unsigned h=0) 
    			{  
 			   integer ret = (   ((integer)Then()     << 5 )
 			                   ^ ((integer)Else()     << 7 )
 			                   ^ ((integer)GetIndex() << 11)
 			                   ^ (h << 8)			);
 			   return ret;
 			}

  virtual unsigned Hash2( )
			    { return (    ((integer)Then() << 5)
					^ ((integer)Else() << 7)
					  			); }

  virtual integer Compare(Comparable *);
   
  // Object virtual functions
  virtual boolean Flush() { return (NumReferences()==0); }
  virtual integer isa() { return BDDNODE; }
  virtual void Print(ostream& out = cout);
  void PrintWithManager(BDDmgr *,HasherTree* ht,ostream& out = cout);

  friend class BDDmgr;
  friend class BDDmanager;
  friend class BDDCache;
  friend class BDDCacheKey;
  friend class BDDHashTable;
};

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

#include "bddcache.h"
class BDDCacheKey;
class BDDCache;

class BDDmgr : public Object {
protected:
  integer                timestamp;

  BDDnode		*one;		// terminal 1
  BDDnode		*zero;		// terminal 0

  BDDnode		*tempnode;	// temporary used by MakeNode()
  BDDCacheKey		*tempcomputed;	// temporary used by ite()

  integer		 num_inputs;	// current number of variables
  integer		 num_nodes;	// total number of ALLOCATED bdd nodes
  unsigned      	 num_ite_calls;	// total number of ite() calls
  BDDCache		*cofactor_hc;	// cofactor cache
  BDDCache		*restrict_hc;	// restrict cache
  IndexArray		*defaultIA;	// all vars

  HasherTree		*keptNodes;	// Nodes we called Keep on
  integer		 num_keeps;
public:
  BDDHasher		*unique;  	// unique "hash table"
  BDDCache		*computed;	// computed cache

  BDDmgr( integer bits=19, integer cofbits=17 );

  // used to copy index order from 'bdd2'
  BDDmgr( BDDmgr* bdd2, integer bits=19, integer cofbits=17 );
  BDDnode* Copy2Manager( BDDnode *f, BDDmgr *bdd2, HashBase *ht = NULL);
  void     Dump2Manager( BDDmgr *bdd2 );

 ~BDDmgr();


  integer NumInputs() { return num_inputs; }
  integer NumNodes()  { return num_nodes; }
  integer GetLevel(BDDnode*);
  integer iteCalls()  { return num_ite_calls; }
  integer TimeStamp() { return timestamp;}
  boolean ReOrdered() { return timestamp != 0;}

  BDDnode* One()                { return one; }
  BDDnode* Zero()               { return zero; }

  BDDnode* ShannonT(BDDnode* n, integer v);	// "safe" version of Shannon for export
  BDDnode* ShannonF(BDDnode* n, integer v);	// "safe" version of Shannon for export

  // input and output of bdd's

  void    PrintCubesF( BDDnode *f, ostream &out=cout, char *lf="%s\n",
		       char *pf="{\n%s}\n", IndexArray *ia=NULL );

  void    PrintCubes(BDDnode *f, char eol, boolean braces=TRUE, ostream& out=cout);
  void    PrintCubes(BDDnode *f, ostream& out=cout) { PrintCubes(f,'\n',TRUE, out);}
  BDDnode* RealizeCube( string cube_descr);
  BDDnode* Read(istream& in = cin);
  BDDnode* ReadEq  (istream& in=cin, integer base=0, char eol='\n') ;
  BDDnode* ReadEq  (char* in, integer base=0, char eol='\n')
    { istrstream a(in); return ReadEq  (a, base, eol);}
  BDDnode* ReadCube(istream& in=cin, integer base=0, char eol='\n');
  BDDnode* ReadCube(char* cube, IndexArray* ia, char eol='\n');
  BDDnode* ReadCube(char* cube, integer base=0);

  void Dump(ostream& out=cout);
  void WriteBinary(List *, ostream& out);
  List *ReadBinary(istream &in);
  BLIFdata* ReadBLIF( istream &in, BLIFdata* bd=NULL );
  BDDnode *ReadEqn( HashTree *htname, HashTree *htindex, char *eqn );
  void ReadPLA( List *onset, List *offset, List *dcset, istream& in=cin );
  void ReadPLA( List *onset, List *offset, List *dcset, char *in )
    { istrstream a(in); ReadPLA( onset, offset, dcset, a ); }

  // XDump takes a BDDnode, a list of BDDnodes, or nothing (unique hasher)
	// defaults to entire unique hasher
  void XDump(Object *f = NULL, char* window_title="BDD Tree", boolean show_ref=FALSE,
	     char **indexNames=NULL);
	// defaults to entire unique hasher
  void XDump_nowait(Object *f = NULL, char* window_title="BDD Tree", boolean sr=FALSE,
	     char **indexNames=NULL);

  // bdd ordering functions
  void Keep( List *fl );
  void UnKeep( List *fl );
  BDDnode* Keep( BDDnode * );
  void UnKeep( BDDnode * );

  void     GC() { doGC(TRUE); }
  void     GC_Sift(int start = 0, int end = -1, int mod = 1);
  void     GC_Window3(int start = 0, int end = -1, int mod = 1);
  void     VariableSwap( int upper_index );
  void	   SwapMVar( int level, int mod );
  void     Swap2Var( int index1, int index2 );
  void	   Reorder_Sift(int downto = 0, int from = -1, int group_size = 1);
  void	   Reorder_Window3(int downto = 0, int from = -1, int group_size = 1);
  BDDnode* Input(integer n) { return (MakeNode(n, one, zero)); }
  BDDnode *CreateVarBefore( int index );
  BDDnode *CreateVarAfter( int index );
  void     OrderIndexes( IndexArray *ia );
  BDDnode* Translate(BDDnode *f, HasherTree *translate, HashBase *dup_check=NULL);

  // bdd core functions
  BDDnode* ite  (BDDnode*, BDDnode*, BDDnode*);
  BDDnode* itek (BDDnode*f, BDDnode*g, BDDnode*h) { return Keep( ite(f,g,h) ); }
  BDDnode* and  (BDDnode *f, BDDnode *g);
  BDDnode* andk (BDDnode *f, BDDnode *g)  { return Keep( and(f,g) );; }

  BDDnode* not  (BDDnode *f)              { return ite(f, zero, one);     }
  BDDnode* notk (BDDnode *f)              { return Keep(ite(f, zero, one));     }
  BDDnode* and_not(BDDnode *f, BDDnode *g){ return ite(g, zero, f);       }
  BDDnode* and_notk(BDDnode *f, BDDnode *g){ return Keep(ite(g, zero, f));       }
  BDDnode* or   (BDDnode *f, BDDnode *g)  { return ite(f, one,    g);     }
  BDDnode* ork  (BDDnode *f, BDDnode *g)  { return Keep(ite(f, one,    g));     }
  BDDnode* or_not(BDDnode *f, BDDnode *g) { return ite(g, f,    one);     }
  BDDnode* or_notk(BDDnode *f, BDDnode *g) { return Keep(ite(g, f,    one));     }
  BDDnode* nand (BDDnode *f, BDDnode *g)  { return ite(f, not(g), one);   }
  BDDnode* nandk(BDDnode *f, BDDnode *g)  { return Keep(ite(f, not(g), one));   }
  BDDnode* nor  (BDDnode *f, BDDnode *g)  { return ite(f, zero,   not(g));}
  BDDnode* nork (BDDnode *f, BDDnode *g)  { return Keep(ite(f, zero,   not(g)));}
  BDDnode* xor  (BDDnode *f, BDDnode *g)  { return ite(f, not(g), g);     }
  BDDnode* xork (BDDnode *f, BDDnode *g)  { return Keep(ite(f, not(g), g));     }
  BDDnode* xnor (BDDnode *f, BDDnode *g)  { return ite(f, g,      not(g));}
  BDDnode* xnork(BDDnode *f, BDDnode *g)  { return Keep(ite(f, g,      not(g)));}
  BDDnode* diff (BDDnode *f, BDDnode *g)  { return and_not(f, g);         }
  BDDnode* diffk(BDDnode *f, BDDnode *g)  { return Keep(and_not(f, g));         }

  // bdd core2 functions

  BDDnode* Cofactor (BDDnode *, BDDnode *); // Generalized Cofactoring Alg
  BDDnode* Restrict (BDDnode *, BDDnode *);
  boolean  Intersect( BDDnode* f, BDDnode* g, HashBase* ht=NULL);  // do f and g intersect?
  boolean  Subset   ( BDDnode* f, BDDnode* g, HashBase* ht=NULL);  // is g a subset of f?
  BDDnode* Compose (BDDnode *f, BDDnode *v, BDDnode *g);
  BDDnode* VariableShift(BDDnode*, integer shift);
  BDDnode* VariableShift(BDDnode*, integer shift, IndexArray *ia);

  BDDnode* ExistentialAbstraction (BDDnode*, List *);
  BDDnode* ExistentialAbstraction (BDDnode*, BDDnode*);
  BDDnode* Smooth(BDDnode*, IndexArray*ia=NULL, HashBase* ht=NULL, boolean InList=TRUE );
  BDDnode* UniversalAbstraction (BDDnode*, List *);
  BDDnode* USmooth(BDDnode*, IndexArray*ia=NULL, HashBase* ht=NULL, boolean InList=TRUE );

  List*    ListCubes(BDDnode* f, IndexArray* ia);
  BDDnode* DC2Zero(BDDnode*, IndexArray* ia=NULL, HashBase* ht=NULL);  
  BDDnode* DC2One (BDDnode*, IndexArray* ia=NULL, HashBase* ht=NULL);  
  BDDnode* One2DC (BDDnode*, IndexArray*ia=NULL, HashBase* ht=NULL);
  BDDnode* Zero2DC (BDDnode*, IndexArray*ia=NULL, HashBase* ht=NULL);
  BDDnode* SampleCube( BDDnode* f, boolean var_pref=TRUE);     

  BDDnode* UpToN   ( integer n, IndexArray* ia=NULL);
  BDDnode* AtLeastN( integer n, IndexArray* ia=NULL);
  BDDnode* Tuple   ( integer n, IndexArray* ia=NULL);

  // bdd statistics functions
  integer  NumNodes(BDDnode *, HashBase* = NULL);
  integer  NumCubes(BDDnode *);  
  integer  NumMinterms(BDDnode *, integer, HashBase* = NULL);
  integer  DepthOfFunction(BDDnode *);
  integer  NumSupport(BDDnode *);

  void       Support(BDDnode*, HashExact*); 
  BDDnode*   Range(List *);
  integer*   NodeDistribution(BDDnode*, integer* i=NULL);


  // Object functions

  void Print(ostream &out = cout);
  void Scan (istream &in  = cin) { Read(in);}
  integer isa () {return BDDMANAGER;}
  
friend class ZBDDmanager;

protected:
  BDDnode *MakeNode(integer, BDDnode *, BDDnode *);
  boolean  IntersectR( BDDnode*, BDDnode*, HashBase*);
  boolean  SubsetR( BDDnode*, BDDnode*, HashBase*);
  void     NumNodesR(BDDnode *, HashBase *);
  integer  NumCubesR(BDDnode *, HasherTree *);
  void     DepthOfFunctionR(BDDnode *, integer, integer&, HasherTree *);
  BDDnode *SmoothR(BDDnode*, IndexArray*, HashBase*, boolean );
  BDDnode *USmoothR(BDDnode*, IndexArray*, HashBase*, boolean );
  void     SupportR(BDDnode*, HasherTree*, HashExact*); 
  void     NumMinterms(BDDnode *, HashBase*, integer& p, integer& h);
  BDDnode *DC2ZeroR(BDDnode*, IndexArray* ia, HashBase* ht);  
  BDDnode *DC2OneR (BDDnode*, IndexArray* ia, HashBase* ht);
  BDDnode *One2DCR (BDDnode*, IndexArray* ia, HashBase* ht);
  BDDnode *Zero2DCR (BDDnode*, IndexArray* ia, HashBase* ht);
  void     PrintCubesR( BDDnode *, char*, ostream& out, char eol);
  void     PrintCubesFR( BDDnode *, char*, ostream& out, char *s, char *e,
			 IndexArray *ia );

  void     doGC(boolean clear_caches);
  void     RGC(BDDnode *);
  void     VSGSR( BDDnode* );


  BDDnode *VariableShift(BDDnode* , integer shift, HashBase* ht);
  BDDnode *VariableShift(BDDnode* , integer shift, HashBase* ht, IndexArray *ia);
  BDDnode *TranslateR(BDDnode *f, HasherTree *translate, HashBase *dup_check=NULL);
  BDDnode *Copy2ManagerR( BDDnode *f, BDDmgr *bdd2, HashBase *ht = NULL);
  BDDnode *FastCopy2ManagerR( BDDnode *f, BDDmgr *bdd2, HashBase *ht = NULL);
  void     WriteBinaryR(BDDnode *, ostream &out);
  BDDnode *CopyNode( BDDnode*, BDDnode *, BDDnode *, BDDnode *, HashBase& );
  void     UnMark( BDDnode *f );
  void     NodeDistributionR(BDDnode*, HasherTree*, integer*);
  List    *ReadEqnR( char *tok,List *rpn,HashTree *name,HashTree *index,int& ind );
  char    *ReadBLIFFunction( istream &in, BLIFdata *bd );
  void     ListCubesR(BDDnode* f, List* cubes, char* cube, IndexArray* ia);

  // assumes v >= n->GetIndex()
  BDDnode* Shannon1(BDDnode* n, integer v)
    { return (v == n->GetIndex()) ? n->Then() : n; }

  BDDnode* Shannon0(BDDnode* n, integer v)
    { return (v == n->GetIndex()) ? n->Else() : n; }


  friend Tcl_AppInit(Tcl_Interp *interp);
};

////////////////////////////////////////////////////////////////////////////

class BDDsubstitution : public Comparable {

  BDDnode *function;
  BDDnode *variable;

  friend class BDDmgr;

public:
  BDDsubstitution () {;}
  BDDsubstitution (BDDnode *f, BDDnode *v) { function = f; variable = v;}
 ~BDDsubstitution () {;}
};

////////////////////////////////////////////////////////////////////////////

#define BDD_ForEach_Cube( bdd, f, ia, cube ) { \
	BDDnode *_f; \
	BDDnode *cube; \
	for( _f=f ; _f != bdd->Zero() ; _f = bdd->and_not(_f,cube) ) { \
	  cube = bdd->SampleCube( _f, FALSE ); \

#define BDD_ForEach_Minterm( bdd, f, ia, mt ) { \
	BDDnode *_f; \
	BDDnode *mt; \
	for( _f=f ; _f != bdd->Zero() ; _f = bdd->and_not(_f,mt) ) { \
	  mt = bdd->SampleCube( _f, FALSE ); \
	  mt = bdd->DC2Zero( mt, ia );

#define BDD_ForEach_Node_In_Cube( bddm, c, n, dir ) { \
	BDDnode *n; \
	boolean dir; \
	for( n=c ; n!=bddm->One() ; n= (dir) ? n->Then() : n->Else() ) { \
	  dir = (n->Then()!=bddm->Zero());

#define BDD_End_ForEach }}

#endif

