#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                        Andrew Seawright
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are not implemented in this version.
// This was done to trade-off efficiency vs. ease of implemetation and ease
// of modification.
//
//	Revised (7/19/93) F. Brewer
//	* changed the inconsistent node creation procedure to prevent
//	some bugs leading to unknown numbers of inputs by the manager.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
//
// Correlation returns a value between 0 and 1 for the amount in common two
// function have. If f1 == f2 or f1 = NOT(f2) then return a 1.
// Ref: Exploiting Circuit Structure in Decomposing the State Space of Large
//      Finite State Machines for Approximate Traversal

float
BDDmanager::Correlation( BDDnode* f1, BDDnode* f2, HashBase* correlation_ht ) {
  boolean new_ht = !correlation_ht;
  
  if( new_ht )
    correlation_ht = new HasherTree(TRUE,TRUE);
  
  float unnormalized = CorrelationR(f1, f2, correlation_ht);
  
  if( new_ht )
    delete correlation_ht;
  
  return ((float) (1 - 4 * (unnormalized * (1 - unnormalized))));
}


float
BDDmanager::CorrelationR( BDDnode* f1, BDDnode* f2, HashBase* correlation_ht )
  {
  if( f1==f2)
    return 1.0;

  if ((f1==one && f2==zero ) || (f1==zero && f2==one))
    return 0.0;
  
  integer          v;
  Object*          o;

  tempcomputed->Set(CORRELATION_OP, f1,f2);

  v = ( GetLevel(f1) > GetLevel(f2) ) ? f1->GetIndex() : f2->GetIndex();

  if( correlation_ht->Get( &o, tempcomputed ) )
    return (float) ((FloatObj*)o)->val;

  /////////////////////////
  // need to compute it
  float f;

  f =  CorrelationR(Shannon1(f1,v), Shannon1(f2,v), correlation_ht);
  f += CorrelationR(Shannon0(f1,v), Shannon0(f2,v), correlation_ht);
  f /= 2.0;
  
  if( f1->NumReferences() > 1 || f2->NumReferences() > 1)
    {
    tempcomputed->Set(CORRELATION_OP, f1,f2);
    correlation_ht->Put( new FloatObj(f), new BDDCacheKey(CORRELATION_OP, f1, f2));
    }
  
  return f;
  }

////////////////////////////////////////////////////////////////////////////
//
// NumMintermsD computed the number of minterms
//


double
BDDmanager::NumMintermsD(BDDnode *f, integer N, HasherTree* hc)
  {
  boolean new_hs;

  if( (new_hs = !hc) )
    hc = new HasherTree(TRUE);

  double _bdd_dpoints = 0;
  integer _bdd_height = 0;

  NumMintermsDR(f,hc, _bdd_dpoints, _bdd_height);

  if( N > _bdd_height )
    _bdd_dpoints = _bdd_dpoints * pow(2.0,  (N - _bdd_height));

  if( new_hs )
    {
    Hasher_ForEach_Type( hc, List,k, l)
      {
      l->CleanOut();
      }Hasher_End_ForEach;
    delete hc;
    }
  return _bdd_dpoints;
}

void
BDDmanager::NumMintermsDR(BDDnode *f, HasherTree* hc, double& _bdd_dpoints, integer& _bdd_height)
  {
  if( f == zero )
    {
    _bdd_height = 0;
    _bdd_dpoints = 0;
    return;
    }

  if( f == one )
    {
    _bdd_height = 0;
    _bdd_dpoints = 1.0e-250;
    return;
    }
  
  List *l;

  if( hc->Get((Object**)&l, f) )
    {
    _bdd_height = ((IntObj*) l->GetElement(0))->val;
    _bdd_dpoints = ((DoubleObj*) l->GetElement(1))->val;
    return;
    }

  NumMintermsDR(f->Then(), hc, _bdd_dpoints, _bdd_height);
  integer height;
  double  dpoints;
  dpoints = _bdd_dpoints;
  height = _bdd_height;

  NumMintermsDR(f->Else(), hc, _bdd_dpoints, _bdd_height);
  if( height > _bdd_height )
    _bdd_dpoints = dpoints + _bdd_dpoints * pow( 2.0, (double) (height - _bdd_height));
  else
    _bdd_dpoints = _bdd_dpoints + dpoints * pow( 2.0, (double) (_bdd_height - height));

  _bdd_height = MAX(_bdd_height, height) + 1;

  if( f->NumReferences() > 1)		// will anybody else ever look for this node?
    {
    l = new List;
    l->Add( ::new DoubleObj( _bdd_dpoints ) );
    l->Add( new IntObj   ( _bdd_height  ) );
    hc->Put( (Object*) l, f );
    }
  return;
}

/////////////////////////////////////////////////////////////////
// 
// reports which indexes are DC for all the cubes of function f
//
/////////////////////////////////////////////////////////////////


BitVector* 
BDDmanager::DCterms( BDDnode* f, integer bot_index, integer top_index) {
  top_index = (top_index == -1)? num_inputs:top_index;
   
  HasherTree* ht = new HasherTree;
  BitVector*  bv = new BitVector( top_index-bot_index+1);
  bv = DCtermsR( f, bv, bot_index, top_index, ht);

  delete ht;
  return bv;
}

BitVector*
BDDmanager::DCtermsR( BDDnode* f, BitVector* bv, integer bot_index, integer top_index, HashBase*ht){
  // have we been here before?
  if ( f == one || f == zero || ht->Get( f) != NULL)
    return bv;

  integer index = f->GetIndex();
  if (index <= top_index && index>=bot_index)
    bv->Set( index-bot_index);

  bv = DCtermsR( f->Then(), bv, bot_index, top_index, ht);
  bv = DCtermsR( f->Else(), bv, bot_index, top_index, ht);

  ht->Put( f);
  return bv;
}

typedef struct mincostdata {
  HasherTree *ht_c, *ht_s;
  int *costs;
  int *c;
  IndexArray *ia;
  int max_bound;
  } mincostdata_t;

